#!/usr/bin/env python

from __future__ import print_function, division
import os, os.path, sys, subprocess, glob, re, time
import multiprocessing
import scipy, scipy.io

"""
  A script to set up a basic similarity graph over Caltech256 features.

  Will execute the following steps:
  1. Download Caltech 256
  2. Unpack the downloaded tar file.
  3. Write a list the files in the order they'll be indexed.
  4. Write a flat with a "ground truth" clustering.
  5. Generate a v1-like feature vector for each image.
  6. Concatenate all the v1-like features into one binary file.
  7. Construct a k-NN graph over the images by this feature, weighted by the
     distance.
  8. Reweight the graph according to a Gaussian RBF.
  9. Build the graph Laplacian.

  All steps are executed lazily, so if the script is interrupted and restarted
  redundant work will not be re-done.  The whole process may take several hours.
"""

# Options
nprocs = multiprocessing.cpu_count()-1

graph_degree = 16
vptree_max_nodes = 1024
rbf_sigma = 4e3

# Locations
caltech256_url = 'http://www.vision.caltech.edu/Image_Datasets/Caltech256/256_ObjectCategories.tar'

data_dir = os.path.dirname(__file__)
if data_dir == '':
    data_dir = '.'
data_dir = os.path.abspath(data_dir)

caltech256_dir = os.path.join(data_dir, '256_ObjectCategories')

root_dir = os.path.dirname(data_dir)
v1like_dir = os.path.join(root_dir, 'external', 'v1like')
v1like_exe = os.path.join(v1like_dir, 'v1like_extract.py')
v1like_cfg = os.path.join(v1like_dir, 'config', 'v1like_a.py')

v1like_path = os.path.join(data_dir, 'caltech256-v1like.dat')

v1graph_exe = os.path.join(root_dir, 'bin', 'v1-graph')
distgraph_path = os.path.join(data_dir, 'caltech256-v1like-distances.graph')

reweight_path = os.path.join(root_dir, 'graph', 'reweight_graph.py')
graph_path = os.path.join(data_dir, 'caltech256-v1like.graph')

laplacian_exe = os.path.join(root_dir, 'bin', 'laplacian')
laplacian_path = os.path.join(data_dir, 'caltech256-v1like.sparsemat')

# Download and unpack dataset
os.chdir(data_dir)
if not os.path.exists('256_ObjectCategories.tar'):
    cmd = ['wget', caltech256_url]
    print(' '.join(cmd))
    subprocess.call(cmd)

if not os.path.isdir(caltech256_dir):
    cmd = ['tar', '-xvf', '256_ObjectCategories.tar']
    print(' '.join(cmd))
    subprocess.call(cmd)
    assert os.path.isdir(caltech256_dir)

# Get a list of all caltech256 images
os.chdir(caltech256_dir)
images = glob.glob(os.path.join('*', '*.jpg'))
images.sort()

flist_path = os.path.join(data_dir, 'caltech256-list.txt')
if not os.path.isfile(flist_path):
    file_list = file(flist_path, 'w+')
    for I in images:
        file_list.write('%s\n' % I)
    file_list.close()

# Write a "ground truth" clustering using the directory names
lbl_path = os.path.join(data_dir, 'caltech256-truth.txt')
if not os.path.isfile(lbl_path):
    clustering = file(lbl_path, 'w+')
    for I in images:
        clustering.write('%d ' % int(I[:I.find('.')]))
    clustering.close()

# Generate v1like features for each image
jobs = []
v1like_files = []
for i, src in enumerate(images):
    feat_dir = os.path.splitext(src)[0] + '.feat'
    if not os.path.isdir(feat_dir):
        os.mkdir(feat_dir)

    base = os.path.splitext(os.path.basename(src))[0]
    tgt = os.path.join(feat_dir, base + '.v1like.mat')

    v1like_files.append(tgt)

    if os.path.isfile(tgt):
        continue

    cmd = ['python', v1like_exe, v1like_cfg, os.path.abspath(src), os.path.abspath(tgt)]
    jobs.append(cmd)

jobs = list(enumerate(jobs))
procs = []
while len(jobs) > 0:
    # Wait for an existing process to complete
    while len(procs) >= nprocs:
        for i, P in enumerate(procs):
            if P.poll() != None:
                procs.pop(i)
        
        if len(procs) >= nprocs:
            time.sleep(0.1)

    # Pop a job off the queue
    i, cmd = jobs[0]
    jobs.pop(0)
    print(' '.join(cmd) + ' (%d remain)' % (len(jobs)))

    procs.append(subprocess.Popen(cmd))


# Concatenate all the v1like features into one file
if not os.path.isfile(v1like_path):
    v1like_file = file(v1like_path, 'w+')
    for i, v1like_matfile in enumerate(v1like_files):
        mat_dict = scipy.io.loadmat(v1like_matfile)
        mat_dict['data'].tofile(v1like_file)

        if i % 512 == 0:
            print('%d of %d written' % (i, len(v1like_files)))
    v1like_file.close()

# Construct the k-NN graph
if not os.path.isfile(distgraph_path):
    cmd = [v1graph_exe, str(len(v1like_files)), str(graph_degree), str(vptree_max_nodes), v1like_path, '86400', distgraph_path]
    print(' '.join(cmd))
    subprocess.call(cmd)

# Reweight according to radial basis function
if not os.path.isfile(graph_path):
    cmd = [reweight_path, '%g' % rbf_sigma, distgraph_path, graph_path]
    print(' '.join(cmd))
    subprocess.call(cmd)

# Construct graph laplacian
if not os.path.isfile(laplacian_path):
    cmd = [laplacian_exe, graph_path, laplacian_path]
    print(' '.join(cmd))
    subprocess.call(cmd)
