#include <assert.h>

#include "adjacency_csr.h"
#include "graph_internal.h"

#include "util/sort.h"
#include "util/int.h"

#include "neighborhood.h"

typedef struct {
  graph base;

  bool sort;

  neighbor_t **ind_in, **ind_out;
  neighbor_t *in, *out;

} adjacency_csr_t;

static const graph_vtable_t vtable;

static inline adjacency_csr_t *cast(graph *G)
{
  assert(G->v == &vtable);

  return (adjacency_csr_t *)G;
}

static inline const adjacency_csr_t *cast_const(graph *G)
{
  assert(G->v == &vtable);

  return (const adjacency_csr_t *)G;
}

/**
 * Find the degree of each vertex from an edge set.
 */
static void degree(size_t V, size_t E, const edge_t *edges,
                   size_t *in, size_t *out)
{
  size_t *d, i;

  for(i = 0; i < V; i++) {
    in[i] = out[i] = 0;
  }
 
  for(i = 0; i < E; i++) {
    out[edges[i].u]++;
    in[edges[i].v]++;
  }
}

/**
 * Find the CSR indices from a degree list
 */
static void indices(size_t V, const size_t *degrees,
                    neighbor_t *base, neighbor_t **indices)
{
  size_t n, i;

  n = 0;
  for(i = 0; i < V; i++) {
    indices[i] = base + n;
    n += degrees[i];
  }
  indices[V] = base + n;
}

graph *graph_adjacency_csr_create(
  size_t V, size_t E,
  bool directed, bool sort,
  const edge_t *edges)
{
  adjacency_csr_t *G;

  size_t i;
  edge_t *e;

  size_t *degree_in, *degree_out, *added_in, *added_out;
  neighbor_t nbr;

  int u;
  
  e = graph_edges_dup(E, edges);

  if(!directed) {
    graph_undirected_edges(E, e);
  }

  // Remove duplicate edges
  graph_edges_dedup(&E, e, NULL);

  // Initialize base
  G = (adjacency_csr_t *)malloc(sizeof(adjacency_csr_t));
  if(G == NULL) {
    return NULL;
  }

  G->base.v = &vtable;
  G->base.V = V;
  G->base.E = E;
  G->base.directed = directed;

  G->sort = sort;

  // Allocate data arrays
  G->in = malloc(sizeof(neighbor_t) * 2 * E);
  G->out = G->in + E;

  // Find node degrees
  degree_in = malloc(sizeof(size_t) * 2 * V);
  degree_out = degree_in + V;
  degree(V, E, e, degree_in, degree_out);

  // Create index arrays
  G->ind_in = malloc(sizeof(neighbor_t *) * 2 * (V + 1));
  G->ind_out = G->ind_in + V + 1;

  indices(V, degree_in, G->in, G->ind_in);
  indices(V, degree_out, G->out, G->ind_out);

  // Copy data into G
  added_in = calloc(2 * V, sizeof(size_t));
  added_out = added_in + V;

  for(i = 0; i < E; i++) {
    // Outgoing edge
    u = e[i].u;
    nbr.v = e[i].v;
    nbr.w = e[i].w;

    assert(u < (int)V);
    assert(added_out[u] < degree_out[u]);
    G->ind_out[u][added_out[u]++] = nbr;

    // Incoming edge
    u = e[i].v;
    nbr.v = e[i].u;

    assert(u < (int)V);
    assert(added_in[u] < degree_in[u]);
    G->ind_in[u][added_in[u]++] = nbr;
  }

  if(sort) {
    for(i = 0; i < V; i++) {
      graph_neighbor_sort(G->ind_in[i+1] - G->ind_in[i], G->ind_in[i]);
      graph_neighbor_sort(G->ind_out[i+1] - G->ind_out[i], G->ind_out[i]);
    }
  }

  // Cleanup
  free(added_in);
  free(degree_in);
  free(e);

  return (graph *)G;
}


graph *graph_adjacency_csr_clone(graph *src, bool sort)
{
  edge_t *e;
  graph *G;

  // Get flat edge array
  e = (edge_t *)malloc(sizeof(edge_t) * src->E);
  graph_edges_all(src, e);

  // Create the graph
  G = graph_adjacency_csr_create(src->V, src->E, src->directed, sort, e);

  // Cleanup
  free(e);

  return G;
}


static void destroy(graph *base)
{
  adjacency_csr_t *G;

  G = cast(base);

  free(G->ind_in);
  free(G->in);

  free(G);
}

static void degrees(graph *base, size_t n, const int *v, int *in, int *out)
{
  adjacency_csr_t *G;
  size_t i;
  int u;

  G = cast(base);

  for(i = 0; i < n; i++) {
    u = v[i];

    assert(u >= 0 && u < G->base.V);

    if(in != NULL) {
      in[i] = G->ind_in[u+1] - G->ind_in[u];
    }
    if(out != NULL) {
      out[i] = G->ind_out[u+1] - G->ind_out[u];
    }
  }
}

static void add_edges(adjacency_csr_t *G, size_t n, edge_t *e)
{
  assert(0);
}

static void set_edges(graph *base, size_t n, const edge_t *e)
{
  adjacency_csr_t *G;
  bool sort;

  size_t i;
  neighbor_t *entry, *begin, *end;
  int v;

  size_t nadd;
  edge_t *add;

  G = cast(base);
  sort = G->sort;

  nadd = 0;
  add = (edge_t *)malloc(n * sizeof(edge_t));

  for(i = 0; i < n; i++) {
    // Find what we're looking for and in what range
    begin = G->ind_out[e[i].u];
    end = G->ind_out[e[i].u + 1];
    v = e[i].v;
    
    if(sort) {
      // Binary search
      entry = graph_neighbor_bsearch(end - begin, begin, v);
    }
    else {
      // Linear search
      entry = graph_neighbor_lsearch(end - begin, begin, v);
    }

    // Save result
    if(entry != NULL) {
      entry->w = e[i].w;
    }
    else {
      assert(nadd < n);
      add[nadd++] = e[i];
    }
  }

  if(nadd != 0) {
    add_edges(G, nadd, add);
  }

  free(add);
}

static void edges_access(graph *base, size_t n, const edge_t *e, bool *present, double *weights)
{
  adjacency_csr_t *G;
  size_t i;
  neighbor_t *begin, *end, search, *entry;
  int v;
  bool sort;

  if(n == 0) {
    return;
  }

  G = cast(base);
  sort = G->sort;

  for(i = 0; i < n; i++) {
    // Find what we're looking for and in what range
    begin = G->ind_out[e[i].u];
    end = G->ind_out[e[i].u + 1];
    v = e[i].v;

    if(sort) {
      // Binary search
      entry = graph_neighbor_bsearch(end - begin, begin, v);
    }
    else {
      // Linear search
      entry = graph_neighbor_lsearch(end - begin, begin, v);
    }

    // Save result
    if(entry != NULL) {
      present[i] = true;
      weights[i] = entry->w;
    }
    else {
      present[i] = false;
    }
  }
}

static void neighborhood(graph *base, int u, edge_t *incoming, edge_t *outgoing)
{
  adjacency_csr_t *G;
  neighbor_t *i, *begin, *end;

  G = cast(base);

  // Incoming edges
  begin = G->ind_in[u];
  end = G->ind_in[u+1];

  for(i = begin; i < end; i++) {
    incoming[i - begin].u = i->v;
    incoming[i - begin].v = u;
    incoming[i - begin].w = i->w;
  }

  // Outgoing edges
  begin = G->ind_out[u];
  end = G->ind_out[u+1];

  for(i = begin; i < end; i++) {
    outgoing[i - begin].u = u;
    outgoing[i - begin].v = i->v;
    outgoing[i - begin].w = i->w;
  }
}

//////////////////////////// Edge iteration //////////////////////////////////

typedef struct {
  int u;
  neighbor_t *n, *end;

} iter_t;

static graph_edge_iterator_t edge_iterate_begin(graph *base)
{
  iter_t *i;
  adjacency_csr_t *G;
  int u;

  i = (iter_t *)malloc(sizeof(iter_t));
  G = cast(base);

  // Get first vertex of nonzero degree
  for(u = 0; u < G->base.V && G->ind_out[u] == G->ind_out[u+1]; u++);

  i->u = u;
  i->n = G->ind_out[u];
  i->end = G->ind_out[u+1];

  return (graph_edge_iterator_t)i;
}

static edge_t edge_iterate_get(graph *base, graph_edge_iterator_t j)
{
  iter_t *i;
  adjacency_csr_t *G;
  edge_t e;

  i = (iter_t *)j;
  G = cast(base);

  e.u = i->u;
  e.v = i->n->v;
  e.w = i->n->w;

  return e;
}

static void edge_iterate_next(graph *base, graph_edge_iterator_t j)
{
  iter_t *i;
  adjacency_csr_t *G;
  int u;

  i = (iter_t *)j;
  G = cast(base);
  
  i->n++;
  if(i->n == i->end) {
    // Get next vertex of nonzero degree
    for(u = i->u + 1; u < G->base.V && G->ind_out[u] == G->ind_out[u+1]; u++);

    i->u = u;
    i->end = G->ind_out[u+1];
  }
}

static bool edge_iterate_end(graph *base, graph_edge_iterator_t j)
{
  iter_t *i;
  adjacency_csr_t *G;

  i = (iter_t *)j;
  G = cast(base);
  
  return i->u == G->base.V;
}

static void edges_visit(graph *base, void (*func)(void *data, const edge_t *e), void *data)
{
  adjacency_csr_t *G;
  edge_t e;
  size_t V;
  neighbor_t *n;
  
  G = cast(base);
  V = G->base.V;

  for(e.u = 0; e.u < (int)V; e.u++) {
    for(n = G->ind_out[e.u]; n < G->ind_out[e.u+1]; n++) {
      e.v = n->v;
      e.w = n->w;

      assert(base->directed || e.u <= e.v);

      func(data, (&e));
    }
  }
}

static void edge_iterate_cleanup(graph *base, graph_edge_iterator_t j)
{
  free(j);
}

static const graph_vtable_t vtable = {
  .destroy = &destroy,
  .degrees = &degrees,
  //.set_edges = set_edges,
  .edges_access = &edges_access,
  .neighborhood = &neighborhood,

  .edge_iterate_begin = &edge_iterate_begin,
  .edge_iterate_get = &edge_iterate_get,
  .edge_iterate_next = &edge_iterate_next,
  .edge_iterate_end = &edge_iterate_end,
  .edge_iterate_cleanup = &edge_iterate_cleanup,
  .edges_visit = &edges_visit
};
