#include <assert.h>

#include "adjacency_list.h"
#include "graph_internal.h"

#include "util/sort.h"
#include "util/int.h"

#include "util.h"
#include "neighborhood.h"

typedef struct {
  size_t d_in, d_out;
  neighbor_t *in;
  neighbor_t *out;

} neighborhood_t;

typedef struct {
  graph base;

  bool sort;
  neighborhood_t *neighborhoods;

} adjacency_list_t;

static const graph_vtable_t vtable;

static inline adjacency_list_t *cast(graph *G)
{
  assert(G->v == &vtable);

  return (adjacency_list_t *)G;
}

static inline const adjacency_list_t *cast_const(const graph *G)
{
  assert(G->v == &vtable);

  return (const adjacency_list_t *)G;
}

graph *graph_adjacency_list_create(int V, bool directed, bool sort)
{
  adjacency_list_t *G;
  neighborhood_t *nbr;
  int i;

  G = (adjacency_list_t *)malloc(sizeof(adjacency_list_t));
  G->base.v = &vtable;
  G->base.V = V;
  G->base.E = 0;
  G->base.directed = directed;

  G->sort = sort;

  nbr = (neighborhood_t *)malloc(sizeof(neighborhood_t) * V);
  for(i = 0; i < V; i++) {
    nbr[i].d_in = nbr[i].d_out = 0;
    nbr[i].in = nbr[i].out = NULL;
  }
  G->neighborhoods = nbr;

  return (graph *)G;
}

static void destroy(graph *base)
{
  adjacency_list_t *G;
  int i, V;
  neighborhood_t *nbr;

  assert(base != NULL); // should already be handled

  G = cast(base);

  // Free neighborhoods
  V = G->base.V;
  nbr = G->neighborhoods;

  for(i = 0; i < V; i++) {
    free(nbr[i].in);
    free(nbr[i].out);
  }

  // Top-level cleanup
  free(nbr);
  free(G);
}


static void degrees(graph *base, size_t n, const int *v, int *in, int *out)
{
  adjacency_list_t *G;
  size_t i;

  G = cast(base);

  for(i = 0; i < n; i++) {
    assert(v[i] >= 0 && v[i] < G->base.V);

    if(in != NULL) {
      in[i] = G->neighborhoods[v[i]].d_in;
    }
    if(out != NULL) {
      out[i] = G->neighborhoods[v[i]].d_out;
    }
  }
}

//////////////////////////// set edges ////////////////////////////

static void set_edges_neighborhood(adjacency_list_t *G, int u, size_t n, const edge_t *e, bool fwd)
{
  int old_d_out;
  neighborhood_t *nbr;

  nbr = G->neighborhoods + u;

  if(fwd) {
    old_d_out = nbr->d_out;
    graph_neighbor_list_set(&(nbr->d_out), &(nbr->out), G->sort, n, e, fwd, -1);
    G->base.E += nbr->d_out - old_d_out;
  }
  else {
    graph_neighbor_list_set(&(nbr->d_in), &(nbr->in), G->sort, n, e, fwd, u);
  }
}

static void set_edges_unidirectional(adjacency_list_t *G, size_t n, const edge_t *e, bool fwd)
{
  size_t i, begin, end;
  neighborhood_t *nbr;
  int j;

  // Iterate over edges, finding contiguous segments to go
  // to the same edge
  begin = 0;
  j = fwd ? e[0].u : e[0].v;
  for(i = 0; i < n; i++) {
    if((fwd && e[i].u != e[begin].u) || (!fwd && e[i].v != e[begin].v)) {
      end = i;

      set_edges_neighborhood(G, j, end - begin, e + begin, fwd);
      begin = end;
      j = fwd ? e[begin].u : e[begin].v;
    }
  }
  end = n;
  set_edges_neighborhood(G, j, end - begin, e + begin, fwd);
}

static void set_edges(graph *base, size_t n, const edge_t *in_e)
{
  adjacency_list_t *G;
  edge_t *e;

  G = cast(base);

  if(n <= 1) {
    // Common trivial case
    set_edges_unidirectional(G, n, in_e, true);
    set_edges_unidirectional(G, n, in_e, false);
  }
  else {
    e = graph_edges_dup(n, in_e);

    // Add to outgoing
    graph_sort_edges_u(n, e);
    set_edges_unidirectional(G, n, e, true);

    // Add to incoming
    graph_sort_edges_v(n, e);
    set_edges_unidirectional(G, n, e, false);

    // Cleanup
    free(e);
  }
}

//////////////////////////// access edges ////////////////////////////

static void edges_access_neighborhood(
  adjacency_list_t *G, neighborhood_t *nbr,
  size_t n, const numbered_edge_t *ne,
  bool *present, double *weights)
{
  size_t i, j, out_i, d;
  bool present_i;

  d = nbr->d_out;

  //if(G->sort) {
    // TODO
  //  assert(0);
  //}
  //else {
    for(i = 0; i < n; i++) {
      out_i = ne[i].i;

      if(present != NULL) {
        present[out_i] = false;
      }

      for(j = 0; j < d; j++) {
        if(ne[i].e.v == nbr->out[j].v) {
          // Mark as present
          if(present != NULL) {
            present[out_i] = true;
          }

          // Save weight
          if(weights != NULL) {
            weights[out_i] = nbr->out[j].w;
          }

          break;
        }
      }
    }
    //}
}

static void edges_access(graph *base, size_t n, const edge_t *e, bool *present, double *weights)
{
  adjacency_list_t *G;
  numbered_edge_t *ne;

  size_t begin, end, i;
  int u;

  G = cast(base);

  // TODO: n <= 1 case

  ne = graph_number_edges(n, e);
  graph_sort_numbered_edges_u(n, ne);

  begin = 0;
  u = ne[0].e.u;
  for(i = 1; i < n; i++) {
    if(ne[i].e.u != u) {
      end = i;
      edges_access_neighborhood(G, G->neighborhoods + u, 
                                end-begin, ne+begin,
                                present, weights);

      begin = end;
      u = ne[begin].e.u;
    }
  }
  end = n;
  edges_access_neighborhood(G, G->neighborhoods + u, 
                            end-begin, ne+begin,
                            present, weights);

  free(ne);
}

////////////////////////// Neighborhood ////////////////////////

static void neighborhood(graph *base, int u, edge_t *in, edge_t *out)
{
  adjacency_list_t *G;
  neighborhood_t nbr;
  size_t i;

  G = cast(base);

  assert(u >= 0 && u < G->base.V);

  // Just a simple copy
  nbr = G->neighborhoods[u];
  for(i = 0; i < nbr.d_in; i++) {
    in[i].u = nbr.in[i].v;
    in[i].v = u;
    in[i].w = nbr.in[i].w;
  }
  
  for(i = 0; i < nbr.d_out; i++) {
    out[i].u = u;
    out[i].v = nbr.out[i].v;
    out[i].w = nbr.out[i].w;
  }
}

////////////////////// Edge Iteration /////////////////////////
typedef struct {
  /**
   * Index of vertex.
   */
  int v;

  /**
   * Index of edge within @c v's neighborhood.
   */
  int e;

} edge_iterator_t;


static graph_edge_iterator_t edge_iterate_begin(graph *base)
{
  adjacency_list_t *G;
  edge_iterator_t *i;

  G = cast(base);
  i = (edge_iterator_t *)malloc(sizeof(edge_iterator_t));

  // Find a non-empty vertex
  for(i->v = 0; i->v < G->base.V && G->neighborhoods[i->v].d_out == 0; i->v++);
  i->e = 0;

  return (graph_edge_iterator_t)i;
}

static edge_t edge_iterate_get(graph *base, graph_edge_iterator_t pi)
{
  adjacency_list_t *G;
  edge_iterator_t *i;
  edge_t e;
  neighbor_t n;

  G = cast(base);
  i = (edge_iterator_t *)pi;

  // Check we're not past-the-end
  assert(i->v < G->base.V);
  assert(i->e < (int)G->neighborhoods[i->v].d_out);

  // Get neighborhood
  n = G->neighborhoods[i->v].out[i->e];

  // Fill edge
  e.u = i->v;
  e.v = n.v;
  e.w = n.w;

  return e;
}

static void edge_iterate_next(graph *base, graph_edge_iterator_t pi)
{
  adjacency_list_t *G;
  edge_iterator_t *i;

  G = cast(base);
  i = (edge_iterator_t *)pi;

  // Check we're not past-the-end
  assert(i->v < G->base.V);
  assert(i->e < (int)G->neighborhoods[i->v].d_out);

  i->e++;
  if(i->e >= (int)G->neighborhoods[i->v].d_out) {
    i->e = 0;
    i->v++;

    // Find a non-empty vertex
    for(; i->v < G->base.V && G->neighborhoods[i->v].d_out == 0; i->v++);
  }
}

static bool edge_iterate_end(graph *base, graph_edge_iterator_t pi)
{
  adjacency_list_t *G;
  edge_iterator_t *i;

  G = cast(base);
  i = (edge_iterator_t *)pi;

  return i->v >= G->base.V;
}

static void edge_iterate_cleanup(graph *base, graph_edge_iterator_t i)
{
  free(i);
}

static const graph_vtable_t vtable = {
  .destroy = &destroy,
  .degrees = &degrees,
  .set_edges = &set_edges,
  .edges_access = &edges_access,
  .neighborhood = &neighborhood,

  .edge_iterate_begin = edge_iterate_begin,
  .edge_iterate_get = edge_iterate_get,
  .edge_iterate_next = edge_iterate_next,
  .edge_iterate_end = edge_iterate_end,
  .edge_iterate_cleanup = edge_iterate_cleanup,
};

void graph_adjacency_list_dump(graph *base)
{
  adjacency_list_t *G;
  int i, j;
  neighborhood_t *nbr;

  G = cast(base);

  // Basic graph properties
  printf("V = %d, E = %d\n", G->base.V, G->base.E);
  printf("directed = %s, sort = %s\n",
         G->base.directed ? "true" : "false",
         G->sort ? "true" : "false");

  for(i = 0; i < G->base.V; i++) {
    printf("  %3d: ", i);

    nbr = G->neighborhoods + i;

    for(j = 0; j < (int)nbr->d_out; j++) {
      printf("(%d,%d) ", i, nbr->out[j].v);
    }

    printf("\n     : ");
    for(j = 0; j < (int)nbr->d_in; j++) {
      printf("(%d,%d) ", nbr->in[j].v, i);
    }
    printf("\n");
  }
}
