#include <assert.h>
#include <stdio.h>
#include <stdint.h>

#include "binary.h"
#include "graph_internal.h"

typedef struct {
  graph base;

  FILE *fp;
  int *inp, *outp;
  off_t edges_start;

} bingraph;

static const graph_vtable_t vtable;

static inline bingraph *cast(graph *G)
{
  assert(G->v == &vtable);

  return (bingraph *)G;
}

static inline const bingraph *cast_const(graph *G)
{
  assert(G->v == &vtable);

  return (const bingraph *)G;
}

static int readints(FILE *fp, int n, int *p)
{
  size_t r;

  r = fread(p, sizeof(int), (size_t)n, fp);
  if(r != (size_t)n) {
    return -1;
  }

  return 0;
}

graph *graph_binary_open(const char *path)
{
  FILE *fp;
  bingraph *G;
  int i, V, header[3];

  // Open file
  fp = fopen(path, "rb");
  if(fp == NULL) {
    return NULL;
  }

  if(readints(fp, 3, header) != 0) {
    fclose(fp);
    return NULL;
  }

  // Set up structure
  G = (bingraph *)malloc(sizeof(bingraph));
  if(G == NULL) {
    fclose(fp);
    return NULL;
  }

  G->base.v = &vtable;
  G->base.V = V = header[0];
  G->base.E = header[1];
  G->base.directed = (header[2] != 0);
  G->fp = fp;

  // Read index
  G->inp = (int *)malloc(sizeof(int) * 2 * (V+1));
  if(G->inp == NULL) {
    free(G);
    fclose(fp);
    return NULL;
  }
  G->outp = G->inp + V;

  for(i = 0; i < V+1; i++) {
    if(readints(fp, 1, &(G->inp[i])) != 0
       || readints(fp, 1, &(G->outp[i])) != 0) {
      free(G->inp);
      free(G);
      fclose(fp);
      return NULL;
    }
  }

  G->edges_start = ftello(fp);
  assert(G->edges_start = (off_t)(12 + (V+1)*8));

  return (graph *)G;
}

static void destroy(graph *base)
{
  bingraph *G;

  G = cast(base);

  fclose(G->fp);
  free(G->inp);
  free(G);
}

static void degrees(graph *base, size_t n, const int *v, int *in, int *out)
{
  bingraph *G;
  size_t i;
  int u;

  G = cast(base);

  for(i = 0; i < n; i++) {
    u = v[i];

    assert(u >= 0 && u < G->base.V);

    if(in != NULL) {
      in[i] = G->outp[u] - G->inp[u];
    }
    if(out != NULL) {
      out[i] = G->inp[u+1] - G->outp[u];
    }
  }
}

static const graph_vtable_t vtable = {
  .destroy = &destroy,
  .degrees = &degrees,
  //.edges_access = &edges_access,
  //.neighborhood = &neighborhood,

  //.edge_iterate_begin = &edge_iterate_begin,
  //.edge_iterate_get = &edge_iterate_get,
  //.edge_iterate_next = &edge_iterate_next,
  //.edge_iterate_end = &edge_iterate_end,
  //.edge_iterate_cleanup = &edge_iterate_cleanup,
  //.edges_visit = &edges_visit
};
