#include <stdio.h>
#include <assert.h>

#include "knn.h"
#include "util/vptree.h"
#include "util/progress.h"
#include "adjacency_csr.h"

typedef struct {
  int V;

  void *knn_user_data;
  double (*distance)(void *knn_user_data, int i, int j);
} context;

static double vpdistance(void *user_data, const void *p1, const void *p2)
{
  context *ctx;
  int i, j;

  ctx = (context *)user_data;
  i = *(const int *)p1;
  j = *(const int *)p2;

  return ctx->distance(ctx->knn_user_data, i, j);
}

static double oneweight(void *knn_user_data, int i, int j)
{
  return 1;
}

graph *graph_build_epsilon(
  int V, double epsilon, void *knn_user_data,
  double (*distance)(void *knn_user_data, int i, int j),
  double (*weight)(void *knn_user_data, int i, int j))
{
  int i, j;

  context ctx;
  vptree_options vpopts;
  vptree *vp;

  int *ints;
  const void **ptrs;

  const void **neighbors;
  edge_t *edges, e;
  int nneighbors, ealloc, E;

  graph *G;

  time_t timer;

  if(weight == NULL) {
    weight = oneweight;
  }

  fprintf(stderr, "Building VP tree for graph...");
  fflush(stdout);

  // Build VP-tree
  ctx.V = V;
  ctx.knn_user_data = knn_user_data;
  ctx.distance = distance;

  vpopts = vptree_default_options;
  vpopts.user_data = &ctx;
  vpopts.distance = vpdistance;

  vp = vptree_create(sizeof(vpopts), &vpopts);

  // Fill vp tree
  ints = (int *)malloc(sizeof(int) * V);
  ptrs = (const void **)malloc(sizeof(const void *) * V);
  for(i = 0; i < V; i++) {
    ints[i] = i;
    ptrs[i] = &(ints[i]);
  }

  vptree_add_many(vp, V, ptrs);

  fprintf(stderr, " done\n");
  fprintf(stderr, "Querying VP tree...\n");
  timer = time(NULL);

  // Perform nearest-neighbor queries
  // Build graph in process
  ealloc = 16;
  E = 0;
  edges = (edge_t *)malloc(sizeof(edge_t) * ealloc);

  for(i = 0; i < V; i++) {
    // Query
    neighbors = vptree_neighborhood(vp, &i, epsilon, &nneighbors);

    // Assure sufficient space
    if(E + nneighbors > ealloc) {
      ealloc = E + nneighbors;
    }
    edges = (edge_t *)realloc(edges, sizeof(edge_t) * ealloc);

    for(j = 0; j < nneighbors; j++) {
      if(neighbors[j] == ptrs[i]) {
        continue;
      }

      // Add to graph
      e.u = i;
      e.v = *(const int *)(neighbors[j]);
      e.w = weight(knn_user_data, i, e.v);

      assert(e.v >= 0 && e.v < V);

      edges[E++] = e;
    }

    free(neighbors);
    print_progress(stderr, timer, i, V, 1);
  }

  assert(ealloc >= E);

  // VP_tree cleanup
  vptree_destroy(vp);
  free(ptrs);
  free(ints);

  // Build graph structure
  G = graph_adjacency_csr_create(V, E, false, true, edges);

  // Final cleanup
  free(edges);

  return G;
}
