#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <memory.h>
#include <sysexits.h>

#include "util/geom.h"
#include "graph/clustering_graph.h"
#include "graph/io.h"
#include "util/progress.h"

const char * const USAGE = "\
v1-graph <n> <k> <max nodes> <v1 file> <v1 dim> <graph>\n\
\n\
           n: The total number of images.\n\
           k: Number of nearest neighbors to connect to each vertex.\n\
   max nodes: Maximum number of VP-tree nodes to visit for approximate k-NN.\n\
              Set to -1 to perform exact k-NN.\n\
     v1 file: File containing v1-like featurs.\n\
      v1 dim: Length of v1-like feature vectors.\n\
       graph: Output graph path.\n\
";

typedef struct {

  int nimages;

  /**
   * Length of descriptors.
   */
  size_t dim;

  /**
   * Length of v1-like file.
   */
  FILE *v1like_file;

} v1_graph_ctx;

static void v1_descriptor_load(v1_graph_ctx *p, int i, float *desc);
static double v1_distance(void *p, int i, int j);

int main(int argc, char **argv)
{
  int k, max_nodes, dim;
  const char *v1path, *graph_path;

  v1_graph_ctx ctx;
  int i, nimages, m;
  graph *G;

  time_t timer;

  // Interpret arguments
  if(argc != 7) {
    fprintf(stderr, "Usage: %s\n", USAGE);
    return EX_USAGE;
  }
  nimages = atoi(argv[1]);
  k = atoi(argv[2]);
  max_nodes = atoi(argv[3]);
  v1path = argv[4];
  dim = atoi(argv[5]);
  graph_path = argv[6];

  // Set up context
  ctx.nimages = nimages;
  ctx.dim = dim;
  ctx.v1like_file = fopen(v1path, "r");

  if(ctx.v1like_file == NULL) {
    fprintf(stderr, "Could not open v1-like file \"%s\"\n", v1path);
  }

  // Build graph
  G = graph_clustering(nimages, k, max_nodes, &ctx, v1_distance, v1_distance);

  printf("\nBuilt graph with %d edges\n", graph_edge_count(G));

  // Output graph
  graph_write_path(graph_path, G);
  
  // Cleanup
  fclose(ctx.v1like_file);
  graph_destroy(G);

  return 0;
}

static void v1_descriptor_load(v1_graph_ctx *ctx, int i, float *desc)
{
  FILE *fp;
  off_t off;
  int s;
  size_t nread;

  assert(i >= 0 && i < ctx->nimages);

  fp = ctx->v1like_file;
  off = ((off_t)i) * ((off_t)ctx->dim);
  
  // Seek to descriptor
  s = fseeko(fp, off, SEEK_SET);
  assert(s == 0);

  // Read this descriptor
  nread = fread(desc, sizeof(float), ctx->dim, fp);
  assert((size_t)nread == ctx->dim);
}

static double float_distance(size_t dim, float *p, float *q)
{
  double d, diff;
  size_t i;

  d = 0;
  for(i = 0; i < dim; i++) {
    diff = (double)(p[i] - q[i]);
    d += diff*diff;
  }

  return sqrt(d);
}

static double v1_distance(void *ctx_ptr, int i, int j)
{
  v1_graph_ctx *ctx;
  size_t dim;
  float *p, *q;
  double dist;

  ctx = (v1_graph_ctx *)ctx_ptr;

  dim = ctx->dim;
  
  // Load descriptors
  p = (float *)malloc(sizeof(float) * dim * 2);
  q = p + dim;
  v1_descriptor_load(ctx, i, p);
  v1_descriptor_load(ctx, j, q);

  // Calculate distance
  dist = float_distance(dim, p, q);

  free(p);
  return dist;
}

