#include <assert.h>

#include "graph.h"
#include "graph_internal.h"

void graph_destroy(graph *G)
{
  if(G == NULL) {
    return;
  }

  (*G->v->destroy)(G);
}

////////////////////////// Base class properties ////////////////////

int graph_vertex_count(graph *G)
{
  return G->V;
}

int graph_edge_count(graph *G)
{
  return G->E;
}

bool graph_directed(graph *G)
{
  return G->directed;
}

///////////////////////////// Degree //////////////////////////

int graph_vertex_degree(graph *G, int v)
{
  int d_in, d_out;

  if(G->directed) {
    graph_vertex_degrees(G, 1, &v, &d_in, &d_out);

    return d_in + d_out;
  }
  else {
    graph_vertex_degrees(G, 1, &v, &d_in, &d_out);

    return d_in;
  }
}

void graph_vertex_degrees(graph *G, size_t n, const int *v, int *in, int *out)
{
  size_t i;

  if(G->directed) {
    (*G->v->degrees)(G, n, v, in, out);
  }
  else {
    if(in == NULL) {
      if(out == NULL) {
        return;
      }

      in = (int *)malloc(n * sizeof(int));
      graph_vertex_degrees(G, n, v, in, out);
      free(in);
      return;
    }
    else if(out == NULL) {
      out = (int *)malloc(n * sizeof(int));
      graph_vertex_degrees(G, n, v, in, out);
      free(out);
      return;
    }
    
    (*G->v->degrees)(G, n, v, in, out);

    for(i = 0; i < n; i++) {
      in[i] = out[i] = in[i] + out[i];
    }
  }
}

///////////////////////// Neighborhood //////////////////////

void graph_vertex_neighborhood(graph *G, int u, edge_t *e)
{
  int d_in, d_out;

  // Get degrees
  (*G->v->degrees)(G, 1, &u, &d_in, &d_out);

  // Send to directed case, splitting array as needed
  graph_vertex_neighborhood_directed(G, u, e, e + d_in);
}

void graph_vertex_neighborhood_directed(graph *G, int u, edge_t *incoming, edge_t *outgoing)
{
  assert(u >= 0 && u < G->V);

  (*G->v->neighborhood)(G, u, incoming, outgoing);
}

///////////////////////// Get/Set edges //////////////////

void graph_set_edge(graph *G, const edge_t *e)
{
  graph_set_edges(G, 1, e);
}

void graph_set_edges(graph *G, size_t n, const edge_t *e)
{
  edge_t *dup;

  if(G->directed) {
    (*G->v->set_edges)(G, n, e);
  }
  else {
    dup = graph_undirected_edges_dup(n, e);
    
    (*G->v->set_edges)(G, n, dup);

    free(dup);
  }
}

bool graph_edge_present(graph *G, int u, int v)
{
  bool present;

  graph_edges_present(G, 1, &u, &v, &present);

  return present;
}

void graph_edges_present(graph *G, size_t n, const int *u, const int *v, bool *present)
{
  graph_retrieve_edges(G, n, u, v, present, NULL);
}

double graph_edge_weight(graph *G, int u, int v, double none)
{
  double w;

  w = none;
  graph_edge_weights(G, 1, &u, &v, none, &w);

  return w;
}

void graph_edge_weights(graph *G, size_t n, const int *u, const int *v, double none, double *w)
{
  size_t i;

  for(i = 0; i < n; i++) {
    w[i] = none;
  }

  graph_retrieve_edges(G, n, u, v, NULL, w);
}

void graph_retrieve_edge(graph *G, int u, int v, bool *present, double *w)
{
  graph_retrieve_edges(G, 1, &u, &v, present, w);
}

// graph_zip_edges(size_t n, edge_t *e, const int *u, const int *v, const double *w);

void graph_retrieve_edges(graph *G, size_t n, const int *u, const int *v, bool *present, double *w)
{
  edge_t *e;

  // Put inputs into an edge_t list
  e = malloc(sizeof(edge_t) * n);
  graph_zip_edges(n, e, u, v, NULL);

  if(!G->directed) {
    graph_undirected_edges(n, e);
  }

  // The actual call
  (*G->v->edges_access)(G, n, e, present, w);


  // Cleanup
  free(e);
}


////////////////////////// Edge iteration /////////////////////

graph_edge_iterator_t graph_edge_iterate_begin(graph *G)
{
  return (*G->v->edge_iterate_begin)(G);
}

edge_t graph_edge_iterate_get(graph *G, graph_edge_iterator_t i)
{
  return (*G->v->edge_iterate_get)(G, i);
}

void graph_edge_iterate_next(graph *G, graph_edge_iterator_t i)
{
  (*G->v->edge_iterate_next)(G, i);
}

bool graph_edge_iterate_end(graph *G, graph_edge_iterator_t i)
{
  return (*G->v->edge_iterate_end)(G, i);
}

void graph_edge_iterate_cleanup(graph *G, graph_edge_iterator_t i)
{
  (*G->v->edge_iterate_cleanup)(G, i);
}

void graph_edges_visit(graph *G, void (*func)(void *data, const edge_t *e), void *data)
{
  edge_t e;
  graph_edge_iterator_t i;

  if(G->v->edges_visit != NULL) {
    (*G->v->edges_visit)(G, func, data);
  }
  else {
    for(i = graph_edge_iterate_begin(G);
        !graph_edge_iterate_end(G, i);
        graph_edge_iterate_next(G, i)) {
    
      e = graph_edge_iterate_get(G, i);
      (*func)(data, &e);
    }

    graph_edge_iterate_cleanup(G, i);
  }
}

void graph_edges_all(graph *G, edge_t *e)
{
  size_t E, j;
  graph_edge_iterator_t i;

  if(G->v->edges_all != NULL) {
    (*G->v->edges_all)(G, e);
  }
  else {
    E = G->E;
    
    for(i = graph_edge_iterate_begin(G), j = 0;
        !graph_edge_iterate_end(G, i);
        graph_edge_iterate_next(G, i), j++) {

      assert(j < E);
      e[j] = graph_edge_iterate_get(G, i);
    }
    assert(j == E);

    graph_edge_iterate_cleanup(G, i);
  }

}

////////////////////// Zip/Unzip edges ////////////////////////

void graph_unzip_edges(size_t n, const edge_t *e, int *u, int *v, double *w)
{
  size_t i;

  for(i = 0; i < n; i++) {
    u[i] = e[i].u;
    v[i] = e[i].v;

    if(w != NULL) {
      w[i] = e[i].w;
    }
  }
}

void graph_zip_edges(size_t n, edge_t *e, const int *u, const int *v, const double *w)
{
  size_t i;

  for(i = 0; i < n; i++) {
    e[i].u = u[i];
    e[i].v = v[i];

    if(w != NULL) {
      e[i].w = w[i];
    }
    else {
      e[i].w = 1;
    }
  }
}

////////////////////// Internal Functions //////////////////////

static int compare_edges(const void *p1, const void *p2)
{
  edge_t e1, e2;

  e1 = *(const edge_t *)p1;
  e2 = *(const edge_t *)p2;

  if(e1.u != e2.u) {
    return e1.u - e2.u;
  }
  else {
    return e1.v - e2.v;
  }
}

static int compare_edges_weight(const void *p1, const void *p2)
{
  edge_t e1, e2;

  e1 = *(const edge_t *)p1;
  e2 = *(const edge_t *)p2;

  if(e1.w < e2.w) {
    return -1;
  }
  else if(e1.w > e2.w) {
    return 1;
  }
  else {
    return 0;
  }
}

static int compare_numbered_edges(const void *p1, const void *p2)
{
  numbered_edge_t ne1, ne2;

  ne1 = *(const numbered_edge_t *)p1;
  ne2 = *(const numbered_edge_t *)p2;

  if(ne1.e.u != ne2.e.u) {
    return ne1.e.u - ne2.e.u;
  }
  else {
    return ne1.e.v - ne2.e.v;
  }
}

numbered_edge_t *graph_number_edges(size_t n, const edge_t *e)
{
  numbered_edge_t *ne;
  size_t i;

  ne = (numbered_edge_t *)malloc(sizeof(numbered_edge_t) * n);
  for(i = 0; i < n; i++) {
    ne[i].e = e[i];
    ne[i].i = i;
  }

  return ne;
}

void graph_denumber_edges(size_t n, const numbered_edge_t *ne, edge_t *e, int *perm)
{
  size_t i;

  for(i = 0; i < n; i++) {
    e[i] = ne[i].e;
    perm[i] = ne[i].i;
  }
}

void graph_sort_edges(size_t n, edge_t *e, int *perm)
{
  numbered_edge_t *ne;

  if(perm == NULL) {
    qsort(e, n, sizeof(edge_t), &compare_edges);
  }
  else {
    // For code simplicitly, we frequently will call this in a trivial case
    if(n == 1) {
      perm[0] = 0;
      return;
    }

    // Sort
    ne = graph_number_edges(n, e);
    graph_sort_numbered_edges(n, ne);

    // Save results
    graph_denumber_edges(n, ne, e, perm);

    // Cleanup
    free(ne);
  }
}

void graph_sort_edges_weight(size_t n, edge_t *e)
{
  qsort(e, n, sizeof(edge_t), &compare_edges_weight);
}

void graph_sort_numbered_edges(size_t n, numbered_edge_t *ne)
{
  qsort(ne, n, sizeof(numbered_edge_t), &compare_numbered_edges);
}

void graph_undirected_edges(size_t n, edge_t *e)
{
  int swap;
  size_t i;

  for(i = 0; i < n; i++) {
    if(e[i].u > e[i].v) {
      swap = e[i].u;
      e[i].u = e[i].v;
      e[i].v = swap;
    }
  }
}

edge_t *graph_undirected_edges_dup(size_t n, const edge_t *src)
{
  edge_t *dst;

  dst = graph_edges_dup(n, src);
  graph_undirected_edges(n, dst);

  return dst;
}

void graph_edges_dedup(size_t *pn, edge_t *e, int *perm)
{
  size_t n, m, i, j;
  numbered_edge_t *ne;

  // For code simplicitly, we frequently will call this in a trivial case
  if(*pn == 0) {
    return;
  }
  if(*pn == 1) {
    if(perm != NULL) {
      perm[0] = 0;
    }
    return;
  }

  m = n = *pn;
  j = 0;

  // Non-trivial case
  if(perm == NULL) {
    // Sort
    qsort(e, n, sizeof(edge_t), &compare_edges);

    //for(i = 0; i < n; i++) {
    //  printf("i = %lu, (%d,%d)\n", i, e[i].u, e[i].v);
    //}

    // Find (now adjacent) duplicates
    for(i = 0; i < n-1; i++) {
      if(e[i].u == e[i+1].u && e[i].v == e[i+1].v) {
        // Remove from count
        m--;
      }
      else {
        // Add a unique edge to the output list
        e[j++] = e[i];
      }
    }
    e[j++] = e[n-1];
    
    assert(j == m);

    //for(i = 0; i < m; i++) {
    //  printf("i = %lu, (%d,%d)\n", i, e[i].u, e[i].v);
    //}
  }
  else {
    // Sort
    ne = graph_number_edges(n, e);
    qsort(ne, n, sizeof(numbered_edge_t), &compare_numbered_edges);

    // Find (now adjacent) duplicates
    for(i = 1; i < n; i++) {
      if(ne[i].e.u == ne[i-1].e.u && ne[i].e.v == ne[i-1].e.v) {
        // Remove from count
        m--;
      }
      else {
        // Add a unique edge to the output list
        e[j] = ne[i].e;
        perm[j] = ne[i].i;
        j++;
      }
    }

    // Cleanup
    free(ne);
  }

  *pn = m;
}

void graph_edges_reverse(size_t n, edge_t *e)
{
  size_t i;
  int swap;

  for(i = 0; i < n; i++) {
    swap = e[i].u;
    e[i].u = e[i].v;
    e[i].v = swap;
  }
}
