#ifndef __GRAPH_H__
#define __GRAPH_H__

#include <stdlib.h>
#include <stdbool.h>

typedef struct graph graph;

typedef struct {
	int u, v;
	double w;
} edge_t;

/* TODO: inferface for just setting weights for existing edges */

/**
 * General graph destructor
 */
void graph_destroy(graph *G);

/*************************** Accessors **************************/
int graph_vertex_count(graph *G);
int graph_edge_count(graph *G);
bool graph_directed(graph *G);

/**
 * Number of edges adjacent to vertex @c v
 *
 * @note In directed graph, returns sum of both incoming and outgoing degrees.
 */
int graph_vertex_degree(graph *G, int v);

/**
 * Get degrees of a list of vertices.
 *
 * @arg @c n Number of vertices queried.
 * @arg @c v Query vertices.
 * @arg @c incoming Number of incoming edges.  Ignored if NULL.
 * @arg @c outgoing Number of outgoing edges.  Ignored if NULL.
 *
 * @note @c incoming and @c outgoing will be the same in the undirected case.
 */
void graph_vertex_degrees(graph *G, size_t n, const int *v,
                          int *incoming, int *outgoing);

/**
 * Retrieve the neighborhood of a given vertex.
 *
 * @arg @c u Index of queried vertex.
 * @arg @c e Output argument in which a list of edges is saved.
 *           Must have space allocated for the total degree
 *           of the node.  Ignored if NULL.
 *
 * @todo Document reference to graph_vertex_degree for sizes.
 */
void graph_vertex_neighborhood(graph *G, int u, edge_t *e);

/**
 * Retrieve the incoming and outgoing neighborhood of a given vertex.
 *
 * @arg @c u Index of queried vertex.
 * @arg @c incoming Must have space allocated for the incoming degree
 *                  of the node.  Ignored if NULL.
 * @arg @c outgoing Must have space allocated for the going degree
 *                  of the node.  Ignored if NULL.
 *
 * @todo Only valid on directed graphs.  On undirected graphs will split
 *       edges between incoming and outgoing in undefined manner.
 * @todo Document reference to graph_vertex_degree for sizes.
 */
void graph_vertex_neighborhood_directed(graph *G, int u, edge_t *incoming, edge_t *outgoing);

/**
 * Add a single edge to the graph or set its weight if it already exists.
 */
void graph_set_edge(graph *G, const edge_t *e);

/**
 * Add multiple edges to the graph or set their weights if existing.
 */
void graph_set_edges(graph *G, size_t n, const edge_t *e);

/**
 * Whether an edge connects vertices @c u and @c v
 */
bool graph_edge_present(graph *G, int u, int v);

/**
 * Whether an edge connects each pair of vertices @c u[i] and @c v[i]
 */
void graph_edges_present(graph *G, size_t n, const int *u, const int *v, bool *present);

/**
 * The weight of an edge
 *
 * @arg @c none The weight to return if the edge does not exist.
 */
double graph_edge_weight(graph *G, int u, int v, double none);

/**
 * The weight of each edge pair @c u[i] and @c v[i]
 *
 * @arg @c none The weight to return if the edge does not exist.
 * @arg @c w Output argument
 */
void graph_edge_weights(graph *G, size_t n, const int *u, const int *v, double none, double *w);

/**
 * Retrieve the presence and weight of an edge.
 *
 * @arg @c present Output presence.  Ignored if NULL.
 * @arg @c w Output weight. Ignored if NULL.
 */
void graph_retrieve_edge(graph *G, int u, int v, bool *present, double *w);

/**
 * Retrieve presence and weights of edges.
 *
 * @arg @c present Output presence.  Ignored if NULL.
 * @arg @c w Output weight. Ignored if NULL.
 */
void graph_retrieve_edges(graph *G, size_t n, const int *u, const int *v, bool *present, double *w);

/**
 * Unpack the individual elements of an edge array.
 *
 * @arg @c w Ignored if NULL.
 */
void graph_unzip_edges(size_t n, const edge_t *e, int *u, int *v, double *w);

/**
 * Pack the individual elements of an edge array.
 *
 * @arg @c w The weights.  May be NULL, in which case the weights will be set
 *           to 1.
 */
void graph_zip_edges(size_t n, edge_t *e, const int *u, const int *v, const double *w);

/**************************** Edge iteration *************************/

typedef void *graph_edge_iterator_t;

/**
 * Begin iteration over edges.
 *
 * Iterator pointing to first edge is returned.
 *
 * @note Caller must call graph_edge_iterate_cleanup on result.
 */
graph_edge_iterator_t graph_edge_iterate_begin(graph *G);

/**
 * Get the edge the iterator is currently pointing to.
 */
edge_t graph_edge_iterate_get(graph *G, graph_edge_iterator_t i);

/**
 * Iterate to next edge.
 */
void graph_edge_iterate_next(graph *G, graph_edge_iterator_t i);

/**
 * Check if iteration is past-the-end.
 */
bool graph_edge_iterate_end(graph *G, graph_edge_iterator_t i);

/**
 * Clean up iteration.
 */
void graph_edge_iterate_cleanup(graph *G, graph_edge_iterator_t i);

/**
 * Call a closure on each edge.
 */
void graph_edges_visit(graph *G, void (*func)(void *data, const edge_t *e), void *data);

/**
 * Retrieve a list of all the edges of a graph.
 */
void graph_edges_all(graph *G, edge_t *e);

#endif // #ifndef __GRAPH_H__
