#ifndef __GRAPH_INTERNAL_H__
#define __GRAPH_INTERNAL_H__

#include <memory.h>
#include <stdio.h>

typedef struct graph_vtable graph_vtable_t;

/**
 * Base object for graphs.
 */
struct graph
{
  const graph_vtable_t *v;

  /**
   * Vertex count.
   */
  int V;

  /**
   * Edge count.
   */
  int E;

  /**
   * Whether this is a directed or undirected graph.
   */
  bool directed;
};

struct graph_vtable
{
  void (*destroy)(graph *base);

  void (*degrees)(graph *base, size_t n, const int *v, int *in, int *out);

  void (*set_edges)(graph *base, size_t n, const edge_t *e);
  void (*edges_access)(graph *base, size_t n, const edge_t *e, bool *present, double *weights);

  void (*neighborhood)(graph *base, int u, edge_t *incoming, edge_t *outgoing);

  // Edge iteration
  graph_edge_iterator_t (*edge_iterate_begin)(graph *base);
  edge_t (*edge_iterate_get)(graph *base, graph_edge_iterator_t i);
  void (*edge_iterate_next)(graph *base, graph_edge_iterator_t i);
  bool (*edge_iterate_end)(graph *base, graph_edge_iterator_t i);
  void (*edge_iterate_cleanup)(graph *base, graph_edge_iterator_t i);
  void (*edges_all)(graph *base, edge_t *e);
  void (*edges_visit)(graph *base, void (*func)(void *data, const edge_t *e), void *data);
};

/**
 * Duplicate a list of edges.
 *
 * @note Caller must free result.
 */
static inline edge_t *graph_edges_dup(size_t n, const edge_t *src)
{
  edge_t *dst;

  dst = (edge_t *)malloc(n * sizeof(edge_t));
  memcpy(dst, src, n * sizeof(edge_t));

  return dst;
}

/**
 * Sort a list of edges, ordered first by @c u, then @c v
 *
 * @arg @c perm Array of original positions of each edge in the resulting list.
 *              Ignored if NULL.
 */
void graph_sort_edges(size_t n, edge_t *e, int *perm);

/**
 * Sort a list of edges by weight.
 */
void graph_sort_edges_weight(size_t n, edge_t *e);

/**
 * Set a group of edges for use in an undirected graph.
 *
 * Simply enforces u <= v
 *
 * @todo Reference to graph_undirected_edges_dup
 */
void graph_undirected_edges(size_t n, edge_t *e);

/**
 * Return a duplicate list of edges for use in an undirected graph.
 *
 * @note Caller must free result.
 * @todo Reference to graph_undirected_edges
 */
edge_t *graph_undirected_edges_dup(size_t n, const edge_t *e);

/**
 * Remove duplicate from a list of edges.
 *
 * @arg @c n On input, the number of edges in the array.  On output,
 *           the number of unique edges in the final array.
 * @arc @c e On input, an array of edges possibly with duplicates.
 *           On output, an array of edges from the input list with
 *           no duplicates.  Elements past the output @c n will have
 *           undefined contents.
 * @arg @c perm The original indices of the edges.  Ignored if NULL.
 */
void graph_edges_dedup(size_t *n, edge_t *e, int *perm);

/**
 * Reverse a list of edges, swapping @c u and @c v for each.
 */
void graph_edges_reverse(size_t n, edge_t *e);

/**
 * Edges with a list index attached.
 *
 * Allows for sorting while saving the original permutation easily.
 */
typedef struct {
  edge_t e;
  int i;
} numbered_edge_t;

/**
 * Attach indices to each element of an edge list.
 *
 * @note Caller must free result.
 */
numbered_edge_t *graph_number_edges(size_t n, const edge_t *e);

/**
 * Extract edges and indices from a numbered edge list.
 */
void graph_denumber_edges(size_t n, const numbered_edge_t *ne, edge_t *e, int *perm);

/**
 * Sort a numbered edge list.
 */
void graph_sort_numbered_edges(size_t n, numbered_edge_t *ne);

#endif // #ifndef __GRAPH_INTERNAL_H__
