#include <stdio.h>

#include "knn.h"
#include "util/vptree.h"
#include "util/progress.h"
#include "adjacency_csr.h"

typedef struct {
  int V;

  void *knn_user_data;
  double (*distance)(void *knn_user_data, int i, int j);
} context;

static double vpdistance(void *user_data, const void *p1, const void *p2)
{
  context *ctx;
  int i, j;

  ctx = (context *)user_data;
  i = *(const int *)p1;
  j = *(const int *)p2;

  return ctx->distance(ctx->knn_user_data, i, j);
}

static double oneweight(void *knn_user_data, int i, int j)
{
  return 1;
}

static void progress_callback(void *user_data, int i, int n)
{
  time_t *timer;

  timer = (time_t *)user_data;

  if(i % 32*1024 == 0) {
    print_progress(stderr, *timer, i, n, 1);
  //  fprintf(stderr, "Built node %d of %d (%.2f%%)\n", i, n, (((float)i)*100)/n);
  }
}

graph *graph_build_knn(
  int V, int k, int max_nodes, int verbose,
  void *knn_user_data,
  double (*distance)(void *knn_user_data, int i, int j),
  double (*weight)(void *knn_user_data, int i, int j))
{
  int i, j;

  context ctx;
  vptree_options vpopts;
  vptree *vp;

  int *ints;
  const void **ptrs;

  const void **nn;
  edge_t *edges, e;

  graph *G;

  time_t timer;

  if(weight == NULL) {
    weight = oneweight;
  }

  if(verbose) {
    fprintf(stderr, "Building VP tree for graph...\n");
  }

  // Build VP-tree
  ctx.V = V;
  ctx.knn_user_data = knn_user_data;
  ctx.distance = distance;

  vpopts = vptree_default_options;
  vpopts.user_data = &ctx;
  vpopts.distance = vpdistance;

  vp = vptree_create(sizeof(vpopts), &vpopts);

  // Fill vp tree
  ints = (int *)malloc(sizeof(int) * V);
  ptrs = (const void **)malloc(sizeof(const void *) * V);
  for(i = 0; i < V; i++) {
    ints[i] = i;
    ptrs[i] = &(ints[i]);
  }

  if(verbose) {
    timer = time(NULL);
    vptree_add_many_progress(vp, V, ptrs, &timer, progress_callback);
  }
  else {
    vptree_add_many(vp, V, ptrs);
  }

  if(verbose) {
    fprintf(stderr, "\nQuerying VP tree...\n");
    timer = time(NULL);
  }

  // Perform nearest-neighbor queries
  // Build graph in process
  nn = (const void **)malloc(sizeof(const void *) * (k+1));
  edges = (edge_t *)malloc(sizeof(edge_t) * k * V);

  for(i = 0; i < V; i++) {
    if(max_nodes == -1) {
      vptree_nearest_neighbor(vp, &i, k+1, nn);
    }
    else {
      vptree_nearest_neighbor_approx(vp, &i, k+1, nn, max_nodes);
    }

    for(j = 0; j < k; j++) {
      // Add to graph
      e.u = i;
      e.v = *(const int *)(nn[j+1]);
      e.w = weight(knn_user_data, i, e.v);

      edges[k*i + j] = e;
    }

    if(verbose) {
      print_progress(stderr, timer, i, V, 1);
    }
  }


  // VP_tree cleanup
  vptree_destroy(vp);
  free(nn);
  free(ptrs);
  free(ints);

  // Build graph structure
  G = graph_adjacency_csr_create(V, k*V, false, true, edges);

  // Final cleanup
  free(edges);

  if(verbose) {
    fprintf(stderr, "\n");
  }

  return G;
}
