#include <stdio.h>
#include <pthread.h>

#include "knn.h"
#include "util/vptree_parallel.h"
#include "util/progress.h"
#include "adjacency_csr.h"

typedef struct {
  int V;

  void *knn_user_data;
  double (*distance)(void *knn_user_data, int i, int j);
} context;

static double vpdistance(void *user_data, const void *p1, const void *p2)
{
  context *ctx;
  int i, j;

  ctx = (context *)user_data;
  i = *(const int *)p1;
  j = *(const int *)p2;

  return ctx->distance(ctx->knn_user_data, i, j);
}

static double oneweight(void *knn_user_data, int i, int j)
{
  return 1;
}

static void progress_callback(void *user_data, int i, int n)
{
  time_t *timer;

  timer = (time_t *)user_data;

  //print_progress(stderr, *timer, i, n, 1);
  fprintf(stderr, "Built node %d of %d (%.2f%%)\n", i, n, (((float)i)*100)/n);
}

struct query_ctx
{
  pthread_t pid;

  vptree *vp;
  int k, max_nodes;
  int off, n, allV;
  edge_t *edges;
  int verbose;

  void *knn_user_data;
  double (*weight)(void *knn_user_data, int i, int j);

  time_t timer;
  int *progressi;
  pthread_mutex_t *progress_lock;
};

static void *query_routine(void *ptr)
{
  struct query_ctx *ctx;
  
  int i, j, k, ind;
  const void **nn;
  edge_t e;

  ctx = (struct query_ctx *)ptr;
  k = ctx->k;
  nn = (const void **)malloc(sizeof(const void *) * (k+1));

  for(i = 0; i < ctx->n; i++) {
    ind = ctx->off + i;
    if(ctx->max_nodes == -1) {
      vptree_nearest_neighbor(ctx->vp, &ind, k+1, nn);
    }
    else {
      vptree_nearest_neighbor_approx(ctx->vp, &ind, k+1, nn, ctx->max_nodes);
    }

    for(j = 0; j < k; j++) {
      // Add to graph
      e.u = ind;
      e.v = *(const int *)(nn[j+1]);
      e.w = ctx->weight(ctx->knn_user_data, ind, e.v);

      ctx->edges[k*i + j] = e;
    }

    if(ctx->verbose && (i + 1) % 8 == 0) {
      pthread_mutex_lock(ctx->progress_lock);
      *ctx->progressi += 8;
      print_progress(stderr, ctx->timer, *ctx->progressi, ctx->allV, 1);
      pthread_mutex_unlock(ctx->progress_lock);
    }
  }

  free(nn);

  return NULL;
}

graph *graph_build_knn_parallel(
  int V, int k, int max_nodes, int verbose, int nthreads,
  void *knn_user_data,
  double (*distance)(void *knn_user_data, int i, int j),
  double (*weight)(void *knn_user_data, int i, int j))
{
  int i, j;

  context ctx;
  vptree_options vpopts;
  vptree *vp;

  int *ints;
  const void **ptrs;

  edge_t *edges;
  struct query_ctx *queries;

  graph *G;

  time_t timer;
  int query_progress;
  pthread_mutex_t query_progress_lock;

  if(weight == NULL) {
    weight = oneweight;
  }

  if(verbose) {
    fprintf(stderr, "Building VP tree for graph...\n");
  }

  // Build VP-tree
  ctx.V = V;
  ctx.knn_user_data = knn_user_data;
  ctx.distance = distance;

  vpopts = vptree_default_options;
  vpopts.user_data = &ctx;
  vpopts.distance = vpdistance;

  vp = vptree_create(sizeof(vpopts), &vpopts);

  // Fill vp tree
  ints = (int *)malloc(sizeof(int) * V);
  ptrs = (const void **)malloc(sizeof(const void *) * V);
  for(i = 0; i < V; i++) {
    ints[i] = i;
    ptrs[i] = &(ints[i]);
  }

  timer = time(NULL);
  vptree_add_many_parallel(vp, V, ptrs, nthreads);

  if(verbose) {
    fprintf(stderr, "\nQuerying VP tree...\n");
  }
  timer = time(NULL);

  // Spawn threads to build graph
  edges = (edge_t *)malloc(sizeof(edge_t) * k * V);
  queries = (struct query_ctx *)malloc(sizeof(struct query_ctx) * nthreads);

  query_progress = 0;
  pthread_mutex_init(&query_progress_lock, NULL);

  for(i = 0; i < nthreads; i++) {
    queries[i].vp = vp;
    queries[i].k = k;
    queries[i].max_nodes = max_nodes;
    queries[i].knn_user_data = knn_user_data;
    queries[i].weight = weight;
    queries[i].progressi = &query_progress;
    queries[i].progress_lock = &query_progress_lock;
    queries[i].timer = timer;
    queries[i].verbose = verbose;
    queries[i].allV = V;

    queries[i].off = i*(V/nthreads);
    if(i == nthreads - 1) {
      queries[i].n = V - queries[i].off;
    }
    else {
      queries[i].n = V/nthreads;
    }
    queries[i].edges = edges + queries[i].off;

    if(i == nthreads - 1) {
      query_routine(&queries[i]);
    }
    else {
      pthread_create(&queries[i].pid, NULL, query_routine, &queries[i]);
    }
  }

  for(i = 0; i < nthreads - 1; i++) {
    pthread_join(queries[i].pid, NULL);
  }

  pthread_mutex_destroy(&query_progress_lock);
  free(queries);

  // VP_tree cleanup
  vptree_destroy(vp);
  free(ptrs);
  free(ints);

  // Build graph structure
  G = graph_adjacency_csr_create(V, k*V, false, true, edges);

  // Final cleanup
  free(edges);

  if(verbose) {
    fprintf(stderr, "\n");
  }

  return G;
}
