#include <assert.h>
#include <math.h>
#include <memory.h>

#include "laplacian.h"
#include "util/sort.h"

static void sort_rows(cs_di *A)
{
  int n, j, k, *perm, alloc, colnnz;
  double *xj;

  alloc = 0;
  perm = NULL;
  xj = NULL;

  for(j = 0; j < A->n; j++) {
    colnnz = A->p[j+1] - A->p[j];
    if(colnnz > alloc) {
      alloc = colnnz;
      perm = realloc(perm, (sizeof(int) + sizeof(double)) * alloc);
    }

    xj = (double *)(perm + colnnz);

    memcpy(xj, A->x + A->p[j], sizeof(double) * colnnz);
    sort_ints_permutation((size_t)colnnz, A->i + A->p[j], perm);

    for(k = 0; k < colnnz; k++) {
      A->x[A->p[j] + k] = xj[perm[k]];
    }
  }
}

/////////////////////////// Un-normalized //////////////////////

struct laplacian_context
{
  int i, E;
  cs_di *L;
};

static void Loffdiag(void *data, const edge_t *e)
{
  struct laplacian_context *ctx;
  int i, E;
  cs_di *L;

  int *Lp, *Li;
  double *Lx;

  int u, v;
  double w;

  ctx = (struct laplacian_context *)data;
  i = ctx->i++;
  E = ctx->E;
  L = ctx->L;
  Lp = L->p;
  Li = L->i;
  Lx = L->x;

  u = e->u;
  v = e->v;
  w = e->w;

  // Add off-diagonal entries
  Li[2*i] = u;
  Lp[2*i] = v;
  Lx[2*i] = -w;

  Li[2*i+1] = v;
  Lp[2*i+1] = u;
  Lx[2*i+1] = -w;

  // Add to diagonal degree entries
  Lx[2*E + e->u] += w;
  Lx[2*E + e->v] += w;
}

cs_di *graph_laplacian(graph *G)
{
  int i, V, E;
  cs_di *L, *Lcmp;
  struct laplacian_context ctx;

  assert(!graph_directed(G));

  V = graph_vertex_count(G);
  E = graph_edge_count(G);

  L = cs_di_spalloc(V, V, 2*E + V, 1, 1);
  if(L == NULL) {
    return NULL;
  }
  L->nz = 2*E + V;

  // Zero out diagonal entries
  for(i = 0; i < V; i++) {
    L->i[2*E + i] = i;
    L->p[2*E + i] = i;
    L->x[2*E + i] = 0;
  }

  // Construct matrix with an iterator over the edges
  ctx.i = 0;
  ctx.E = E;
  ctx.L = L;
  graph_edges_visit(G, Loffdiag, &ctx);

  // Reduce to CSC format
  Lcmp = cs_di_compress(L);
  cs_di_free(L);

  sort_rows(Lcmp);

  return Lcmp;
}

////////////////////////////// Normalized Laplacians /////////////////////

static void degree_closure_undirected(void *data, const edge_t *e)
{
  double *deg;

  deg = (double *)data;

  deg[e->u] += e->w;
  deg[e->v] += e->w;
}

static double *graph_degrees(graph *G)
{
  int i, V;
  double *deg;

  V = graph_vertex_count(G);
  deg = (double *)malloc(sizeof(double) * V);
  if(deg == NULL) {
    return NULL;
  }

  // Zero out degrees
  for(i = 0; i < V; i++) {
    deg[i] = 0;
  }

  // Iterate over edges
  assert(!graph_directed(G)); // TODO
  graph_edges_visit(G, degree_closure_undirected, deg);

  return deg;
}

cs_di *graph_laplacian_sym(graph *G)
{
  double *D;
  int i, j, k, V;
  cs_di *L;

  V = graph_vertex_count(G);
  D = graph_degrees(G);
  if(D == NULL) {
    return NULL;
  }
  
  // Take square root
  for(i = 0; i < V; i++) {
    D[i] = sqrt(D[i]);
  }

  // Get un-normalized laplacian
  L = graph_laplacian(G);
  if(L == NULL) {
    free(D);
    return NULL;
  }

  // Normalization
  assert(V == L->m && V == L->n);
  for(j = 0; j < V; j++) {
    for(k = L->p[j]; k < L->p[j+1]; k++) {
      i = L->i[k];

      if(D[i] == 0 || D[j] == 0) {
        assert(L->x[k] == 0);
      }
      else {
        L->x[k] /= D[i] * D[j];
      }
    }
  }

  return L;
}
