#include <assert.h>
#include <stdio.h>

#include "randcluster.h"

#include "util/debug.h"
#include "graph/clustering_graph.h"
#include "graph/laplacian.h"
#include "util/geom.h"

struct sample_data
{
  size_t ndims, n;
  double *p;
  double wsigma;
};

static double distance(void *user_data, int i, int j);
static double weight(void *user_data, int i, int j);

static void print_edge(void *data, const edge_t *e)
{
  printf("%d %d %lg\n", e->u, e->v, e->w);
}

static void nancheck(void *data, const edge_t *e)
{
  assert(!isnan(e->w));
}

graph *sample_MoG_graph(sample_t *samp,
                        size_t ndims, size_t n, double *p, int *label,
                        size_t k, const double *pi, const double *mu,
                        const double *sigma,
                        double grapheps, double weightsigma)
{
  struct sample_data data;
  graph *G;

  data.ndims = ndims;
  data.n = n;
  data.p = p;
  data.wsigma = weightsigma;

  // Sample points
  sample_MoG(samp, ndims, n, p, label, k, pi, mu, sigma);
  //sample_gaussian(samp, ndims, n, p, mu, sigma);

  assert(!anynan((int)(ndims*n), p));

  // Print points for debug
  //print_matrix(stderr, "%lg", "", ndims, n, p);

  // Build graph
  if(grapheps < 0) {
    fprintf(stderr, "eps = %lg\n", -grapheps);
    G = graph_distance_clustering(data.n, -grapheps, &data, distance, distance);
  }
  else {
    fprintf(stderr, "k = %d\n", (int)grapheps);
    G = graph_clustering(data.n, (int)grapheps, 1024, &data, distance, distance);
  }

  // Print graph
  //graph_edges_visit(G, print_edge, NULL);
  graph_edges_visit(G, nancheck, NULL);

  return G;
}

cs_di *sample_MoG_laplacian(
  sample_t *samp,
  size_t ndims, size_t n, double *p, int *label,
  size_t k, const double *pi, const double *mu,
  const double *sigma,
  double grapheps, double weightsigma)
{
  graph *G;
  cs_di *L;

  G = sample_MoG_graph(samp, ndims, n, p, label, k, pi, mu, sigma, grapheps, weightsigma);

  // Construct laplacian
  L = graph_laplacian_sym(G);

  graph_destroy(G);

  return L;
}

static double distance(void *user_data, int i, int j)
{
  struct sample_data *data;
  const double *p, *q;
  size_t ndims;

  data = (struct sample_data *)user_data;

  assert(i >= 0 && i < (int)data->n);
  assert(j >= 0 && j < (int)data->n);

  ndims = data->ndims;
  p = data->p + ndims * i;
  q = data->p + ndims * j;

  return geom_distance(ndims, p, q);
}

static double weight(void *user_data, int i, int j)
{
  struct sample_data *data;
  double dist, w, sigma;

  data = (struct sample_data *)user_data;
  dist = distance(user_data, i, j);
  sigma = data->wsigma;
  w = exp(-(dist*dist)/(2 * sigma * sigma));

  //printf("dist = %g, w = %g\n", dist, w);

  return w;
}

