#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#include <assert.h>

#include "util/geom.h"
#include "util/debug.h"

#include "block_stiefel.h"
#include "util/linalg.h"
#include "util/randsample.h"
#include "util/setdiff.h"
#include "sample_independent.h"

#ifndef INFINITY
#define INFINITY -(1.0/0.0)
#endif

static void select_rows(int n, int k, int p, double *Vk, const double *V, const int *K);
static void install_rows(int n, int k, int p, double *V, const double *Vk, const int *ind);

static void select_cols(int m, int n, int k, const int *cols, const double *A, double *sub);

static void select_Vkind(struct stiefel_ctx *ctx, int *out_rank, int *indcols, int *depcols, const double *Vk, double *Vkind, double *coeff);
static void compute_VK(struct stiefel_ctx *ctx, int rank, const int *indcols, const int *depcols, double *Vk, const double *Vkind, const double *coeff);

struct stiefel_ctx {
  // Problem parameters
  int n,p,k;

  double (*obj)(void *user_data, int k, int p, const int *K, const double *V, const double *Vk);
  void (*grad)(void *user_data, int k, int p, const int *K, const double *V, double *Gk);
  void *user_data;

  // Workspace and intermediate variables
  double *Ql, *Qr;
  double *Utau;

  double *P;
  double *Plambda;
  double *Psqrt;
  double *U0;
  double *G;
  double *A;

  int *indcols, *depcols;
  double *Vkind, *Vkdep, *coeff, *Gind, *Gdep;

  double *Vktauind, *Vktaudep;

  double *Vk;
  double *Gk;
  int *K;

  double dF;
  double tau;
  double objtau;

  optim_params params;
};

struct stiefel_ctx *block_stiefel_ctx_init(const struct stiefel_prob *prob)
{
  struct stiefel_ctx *ctx;
  int n, p, k;

  ctx = (struct stiefel_ctx *)malloc(sizeof(struct stiefel_ctx));

  // Problem data
  ctx->n = n = prob->n;
  ctx->p = p = prob->p;
  ctx->k = k = prob->k;

  ctx->obj = prob->obj;
  ctx->grad = prob->grad;
  ctx->user_data = prob->user_data;

  ctx->params = prob->params;

  // Workspace
  ctx->Ql = (double *)malloc(sizeof(double) * k * k * 2);
  ctx->Qr = ctx->Ql + k*k;

  ctx->Utau = (double *)malloc(sizeof(double) * k * p);

  ctx->P = (double *)malloc(sizeof(double) * p * p);
  ctx->Plambda = (double *)malloc(sizeof(double) * p);
  ctx->Psqrt = (double *)malloc(sizeof(double) * p * p);
  ctx->U0 = (double *)malloc(sizeof(double) * k * p);
  ctx->G = (double *)malloc(sizeof(double) * k * p);
  ctx->A = (double *)malloc(sizeof(double) * k * k);

  ctx->indcols = (int *)malloc(sizeof(int) * p);
  ctx->depcols = (int *)malloc(sizeof(int) * p);
  ctx->Vkind = (double *)malloc(sizeof(double) * k * p);
  ctx->Vkdep = (double *)malloc(sizeof(double) * k * p);
  ctx->coeff = (double *)malloc(sizeof(double) * k * p);
  ctx->Gind = (double *)malloc(sizeof(double) * k * p);
  ctx->Gdep = (double *)malloc(sizeof(double) * k * p);

  ctx->Vktauind = (double *)malloc(sizeof(double) * k * p);
  ctx->Vktaudep = (double *)malloc(sizeof(double) * k * p);

  ctx->Vk = (double *)malloc(sizeof(double) * k * p);
  ctx->Gk = (double *)malloc(sizeof(double) * k * p);
  ctx->K = (int *)malloc(sizeof(int) * k);

  return ctx;
}

void block_stiefel_ctx_deinit(struct stiefel_ctx *ctx)
{
  free(ctx->Ql);
  free(ctx->Utau);

  free(ctx->P);
  free(ctx->Plambda);
  free(ctx->Psqrt);
  free(ctx->U0);
  free(ctx->G);
  free(ctx->A);

  free(ctx->indcols);
  free(ctx->depcols);
  free(ctx->Vkind);
  free(ctx->Vkdep);
  free(ctx->coeff);
  free(ctx->Gind);
  free(ctx->Gdep);

  free(ctx->Vktauind);
  free(ctx->Vktaudep);

  free(ctx->Vk);
  free(ctx->Gk);
  free(ctx->K);
}

/**
 * Calculate points along a geodesic on the stiefel manifold
 *
 * A is k x k
 * X0 is k x p
 * X is k x p
 */
static void geodesic(struct stiefel_ctx *ctx, int rank, const double *A, const double *X0, double tau, double *X)
{
  double *Ql, *Qr;
  int i, j, p, k;

  p = ctx->p;
  k = ctx->k;

  // Allocate workspace
  Ql = ctx->Ql;
  Qr = ctx->Qr;

  // Calcualte necessary matrices
  for(i = 0; i < k*k; i++) {
    Qr[i] = -(Ql[i] = A[i] * (tau/2));
  }

  for(i = 0; i < k; i++) {
    Ql[k*i + i] += 1;
    Qr[k*i + i] += 1;
  }

  // Find geodesic point
  // X <- (I - B) * X0;
  dgemm_wrap('N', 'N', k, rank, k, 1, Qr, k, X0, k, 0, X, k);

  // X <- (I + B) \ X;
  lusolve(k, rank, Ql, X, NULL, NULL);
}

static double line_search_obj(struct stiefel_ctx *ctx, const double *U0, const double *A, const double *Psqrt, const double *coeff, double tau, const int *K, int rank, const int *indcols, const int *depcols, const double *V, double *Vktau)
{
  int i, j, p, k;
  double *Utau, *Vktauind, *Vktaudep, objtau;

  //fprintf(stderr, "tau = %lg\n", tau);

  p = ctx->p;
  k = ctx->k;

  // Allocate workspace
  Utau = ctx->Utau;
  Vktauind = ctx->Vktauind;
  Vktaudep = ctx->Vktaudep;
  
  // Construct trial point
  geodesic(ctx, rank, A, U0, tau, Utau);
  //printf("Utau = \n"); print_matrix(stderr, "%lg", "", k, rank, Utau);

  dgemm_wrap('N', 'N', k, rank, rank, 1, Utau, k, Psqrt, rank, 0, Vktauind, k);
  if(rank < p) {
    dgemm_wrap('N', 'N', k, p-rank, rank, 1, Vktauind, k, coeff, rank, 0, Vktaudep, k);
  }

  for(j = 0; j < rank; j++) {
    for(i = 0; i < k; i++) {
      Vktau[k*indcols[j] + i] = Vktauind[k*j + i];
    }
  }

  for(j = 0; j < p - rank; j++) {
    for(i = 0; i < k; i++) {
      Vktau[k*depcols[j] + i] = Vktaudep[k*j + i];
    }
  }

  //fprintf(stderr, "Vktau = \n"); print_matrix(stderr, "%lg", "   ", k, p, Vktau);

  // Calculate new objective
  objtau = ctx->obj(ctx->user_data, k, p, K, V, Vktau);

  return objtau;
}

static double reconstruction_error(int k, int p, int rank, const int *indcols, const int *depcols, const double *Vkind, const double *Vk, const double *coeff)
{
  double *Vkdep, err, diff;
  int i, j;

  err = 0;

  //printf("k = %d, p - rank = %d\n", k, p-rank);
  Vkdep = (double *)malloc(sizeof(double) * k * (p - rank));
  dgemm_wrap('N', 'N', k, p - rank, rank, 1, Vkind, k, coeff, rank, 0, Vkdep, k);

  for(j = 0; j < p - rank; j++) {
    for(i = 0; i < k; i++) {
      diff = fabs(Vkdep[j*k + i] - Vk[depcols[j]*k + i]);
      if(diff > err) {
        err = diff;
      }
    }
  }

  free(Vkdep);

  return err;
}

double step_block(struct stiefel_ctx *ctx, const double *V, const int *K, double *Vk, const double *Gk)
{
  double *P, Pij, *Plambda, *Psqrt, *U0, *G, *A;
  double dF, tau, objtau, obj0;
  int i, j, l, n, p, k;
  const optim_params *params;

  int rank, *indcols, *depcols;
  double *Vkind, *coeff, *Gind, *Gdep;

  n = ctx->n;
  p = ctx->p;
  k = ctx->k;
  params = &ctx->params;

  indcols = ctx->indcols;
  depcols = ctx->depcols;
  Vkind = ctx->Vkind;
  coeff = ctx->coeff;
  Gind = ctx->Gind;
  Gdep = ctx->Gdep;
  G = ctx->G;

  obj0 = ctx->obj(ctx->user_data, k, p, K, V, Vk);

  // Find linearily independent columns
  select_Vkind(ctx, &rank, indcols, depcols, Vk, Vkind, coeff);
  if(rank == 0) {
    fprintf(stderr, "Discarded rank-0 update\n");
    return 0;
  }

  // Calculate partial inner products
  P = ctx->P;
  dgemm_wrap('T', 'N', rank, rank, k, 1, Vkind, k, Vkind, k, 0, P, rank);

  // Get sqrtm of P
  Plambda = ctx->Plambda;
  dsyev_wrap('V', 'U', rank, P, rank, Plambda);

  for(i = 0; i < rank; i++) {
    Plambda[i] = sqrt(Plambda[i]);
  }

  Psqrt = ctx->Psqrt;
  sqrtm_calc(rank, P, Plambda, Psqrt);

  // Find initial rotation
  U0 = ctx->U0;
  invsqrtm_mult_right(k, rank, P, Plambda, Vkind, U0);
  
  // Restrict gradient to U space
  select_cols(k, p, rank, indcols, Gk, Gind);
  if(rank < p) {
    select_cols(k, p, p-rank, depcols, Gk, Gdep);
    dgemm_wrap('N', 'T', k, rank, p - rank, 1, Gdep, k, coeff, rank, 1, Gind, k);
  }
  dgemm_wrap('N', 'N', k, rank, rank, 1, Gind, k, Psqrt, rank, 0, G, k);

  // Calculate A = G*U' - U*G' 
  A = ctx->A;
  dgemm_wrap('N', 'T', k, k, rank, 1, G, k, U0, k, 0, A, k);
  dgemm_wrap('N', 'T', k, k, rank, -1, U0, k, G, k, 1, A, k);

  // Line search
  dF = 0;
  for(i = 0; i < k*rank; i++) {
    dF += A[i]*A[i];
  }
  dF *= -0.5;
  ctx->dF = dF;
  
  tau = 1;
  objtau = line_search_obj(ctx, U0, A, Psqrt, coeff, tau, K, rank, indcols, depcols, V, Vk);
  while(objtau > obj0 + params->rho1 * tau * dF) {
    tau = tau/2;
    objtau = line_search_obj(ctx, U0, A, Psqrt, coeff, tau, K, rank, indcols, depcols, V, Vk);
    
    if(tau < 1e-3) {
      break;
    }
  }

  ctx->tau = tau;
  ctx->objtau = objtau;

  return objtau - obj0;
}

static const char *status_header = "\n\
 iter           |dF|           dobj            tau\n\
";

static void print_status(struct stiefel_ctx *ctx, int iter)
{
  printf("%5d   %le   %le   %le\n", iter, -ctx->dF, -ctx->objtau, ctx->tau);
}

extern double trace_full_obj(cs_di *A, int p, const double *V);

void block_stiefel(const struct stiefel_prob *prob, double *V)
{
  struct stiefel_ctx *ctx;
  int i, j, iter, n, p, k;
  int *K, *indcols, *depcols, rank;
  double *Vk, *Vkind, *coeff, *Gk;
  const optim_params *params;

  // Unpack problems
  n = prob->n;
  p = prob->p;
  k = prob->k;
  params = &prob->params;

  // Allocate workspace
  ctx = block_stiefel_ctx_init(prob);

  K = ctx->K;
  Vk = ctx->Vk;
  Gk = ctx->Gk;

  if(params->print_interval > 0) {
    fputs(status_header, stdout);
  }

  for(iter = 0; iter < params->max_iter; iter++) {
    // Randomly pick indices
    randsample(K, n, k, 0);

    // Select out rows
    select_rows(n, k, p, Vk, V, K);
    /*
    fprintf(stderr, "ind =");
    for(i = 0; i < k; i++) {
      fprintf(stderr, " %d", Vind[i]);
    }
    fprintf(stderr, "\n");
    */

    // Get gradient
    prob->grad(prob->user_data, k, p, K, V, Gk);

    // Take a step
    step_block(ctx, V, K, Vk, Gk);

    //break;

    // Copy back into V
    install_rows(n, k, p, V, Vk, K);
    // TODO
    //for(j = 0; j < p; j++) {
    //  for(i = 0; i < k; i++) {
    //    V[n*j + Vind[i]] = Vk[k*j + i];
    //  }
    //}

    if(-ctx->dF < params->grad_epsilon) {
      //break;
    }

    if(params->print_interval > 0 && iter % params->print_interval == 0) {
      print_status(ctx, iter);

      double obj = prob->fullobj(ctx->user_data, ctx->p, V);
      printf("obj %d %lg\n", iter, obj);
    }

  }
  print_status(ctx, iter);

  // Cleanup
  block_stiefel_ctx_deinit(ctx);
}

///////////////////////////////// Parallel ///////////////////////////

/*
static void itoa_list(int buflen, char *buf, int n, const int *ls)
{
  int i, written;

  assert(buflen > 0);

  written = 0;
  for(i = 0; i < n && buflen > 0; i++) {
    written += snprintf(buf + written, buflen - written, "%d ", ls[i]);
  }

  if(written > 0) {
    buf[written-1] = '\0';
  }
}

static pthread_mutex_t debug_lock = PTHREAD_MUTEX_INITIALIZER;

static void print_indices(const char *msg, int k, const int *indices)
{
  char buf[256];
  
  itoa_list(256, buf, k, indices);

  pthread_mutex_lock(&debug_lock);
  printf("%s %s\n", msg, buf);
  pthread_mutex_unlock(&debug_lock);
}

struct thread_context
{
  struct stiefel_ctx ctx;

  int threadi;
  pthread_t id;

  pthread_mutex_t *sample_lock;
  independent_sampler *sample;

  void (*call_client)(const void *pdata, struct itermsg *msg);
  const void *pdata;

  double *V;
};

static void *thread_routine(void *ptr)
{
  struct thread_context *tctx;
  int iter, n, p, k, nnbr, *ind, *nbr;

  pthread_mutex_t *sample_lock;
  independent_sampler *sample;

  struct itermsg msg;
  int nsampled;

  // Setup
  tctx = (struct thread_context *)ptr;

  n = tctx->ctx->n;
  msg.p = p = tctx->ctx->p;
  k = tctx->ctx->k;

  sample_lock = tctx->sample_lock;
  sample = tctx->sample;

  msg.Vind = ind = (int *)malloc(sizeof(int) * k);
  msg.Vk = (double *)malloc(sizeof(double) * k * p);

  for(iter = 0; iter < 1; iter++) {
    // Sample independent
    do {
      pthread_mutex_lock(sample_lock);
      nsampled = independent_sampler_sample(sample, k, ind, &nnbr, &nbr);
      print_indices("sampled", nsampled, ind);
      print_indices("nbr", nnbr, nbr);
      pthread_mutex_unlock(sample_lock);
    } while(nsampled < 2);

    // Assemble iter message
    msg.k = nsampled;
    msg.nnbr = nnbr;
    msg.nbr = nbr;

    select_rows(n, k, p, msg.Vk, tctx->V, ind);
    msg.Vnbr = (double *)malloc(sizeof(double) * nnbr * p);
    select_rows(n, nnbr, p, msg.Vnbr, tctx->V, nbr);
    
    // Call client
    tctx->call_client(tctx->pdata, &msg);
    
    // Propagate changes
    install_rows(n, k, p, tctx->V, msg.Vk, ind);

    // Return sample
    pthread_mutex_lock(sample_lock);
    print_indices("return", nsampled, ind);
    print_indices("return", nnbr, nbr);

    independent_sampler_return(sample, nsampled, ind);
    independent_sampler_return(sample, nnbr, nbr);
    pthread_mutex_unlock(sample_lock);

    // Clean up neighbor set
    free(nbr);
    free(msg.Vnbr);
  }

  // Cleanup
  free(ind);

  return ptr;
}

void block_stiefel_parallel(
  const struct stiefel_prob *prob, double *V,
  int nthreads, struct cs_di_sparse *indep,
  void (*call_client)(const void *pdata, struct itermsg *msg),
  const void *pdata)
{
  int i;
  struct thread_context *tctx;

  pthread_mutex_t sample_lock;
  independent_sampler *sample;

  // Set up sampler
  pthread_mutex_init(&sample_lock, NULL);
  sample = independent_sampler_create(indep);

  // Create thread contexts
  tctx = (struct thread_context *)malloc(sizeof(struct thread_context) * nthreads);

  for(i = 0; i < nthreads; i++) {
    block_stiefel_ctx_init(&tctx[i].ctx, prob);
    tctx[i].threadi = i;
    tctx[i].sample_lock = &sample_lock;
    tctx[i].sample = sample;
    tctx[i].call_client = call_client;
    tctx[i].pdata = pdata;
    tctx[i].V = V;
  }

  // Start up thread workers
  for(i = 0; i < nthreads; i++) {
    pthread_create(&tctx[i].id, NULL, thread_routine, &tctx[i]);
  }

  // Join threads
  for(i = 0; i < nthreads; i++) {
    pthread_join(tctx[i].id, NULL);
  }

  // Cleanup
  for(i = 0; i < nthreads; i++) {
    block_stiefel_ctx_deinit(&tctx[i].ctx);
  }
  free(tctx);
  independent_sampler_destroy(sample);
}


static void thread_client(const void *pdata, struct itermsg *msg)
{
  print_indices("client call", msg->k, msg->Vind);
}

void block_stiefel_multithread(const struct stiefel_prob *prob, double *V, int nthreads, struct cs_di_sparse *indep)
{
  block_stiefel_parallel(prob, V, nthreads, indep, thread_client, NULL);
}
*/

///////////////////////////////// Utility ////////////////////////

/**
 * Select out certain rows of V, place them into Vk
 */
static void select_rows(int n, int k, int p, double *Vk, const double *V, const int *K)
{
  int i, j;

  for(j = 0; j < p; j++) {
    for(i = 0; i < k; i++) {
      Vk[k*j + i] = V[n*j + K[i]];
    }
  }
}

/**
 * Place certain rows from Vk into V
 */
static void install_rows(int n, int k, int p, double *V, const double *Vk, const int *ind)
{
  int i, j;
  
  for(j = 0; j < p; j++) {
    for(i = 0; i < k; i++) {
      V[n*j + ind[i]] = Vk[k*j + i]; 
    }
  }
}

static void select_cols(int m, int n, int k, const int *cols, const double *A, double *sub)
{
  int i, j;

  for(j = 0; j < k; j++) {
    assert(cols[j] >= 0 && cols[j] < n);
    for(i = 0; i < m; i++) {
      sub[m*j + i] = A[m*cols[j] + i];
    }
  }
}

/**
 * Select linearily independent columns from subset of rows of V
 */
static void select_Vkind(struct stiefel_ctx *ctx, int *out_rank, int *indcols, int *depcols, const double *Vk, double *Vkind, double *coeff)
{
  int i, j, n, p, k, rank;

  n = ctx->n;
  p = ctx->p;
  k = ctx->k;

  //printf("Vk = \n"); print_matrix(stderr, "%lg", "", k, p, Vk);

  // Find linearly independent columns
  memcpy(coeff, Vk, sizeof(double) * k * p);

  //printf("k = %d, p = %d\n", k, p);
  //print_matrix(stdout, "%lg", "  ", k, p, coeff);

  rref(k, p, 1e-6, coeff, indcols, &rank);
  indsetdiff(p, rank, indcols, depcols);
  for(j = 0; j < rank; j++) {
    for(i = 0; i < k; i++) {
      Vkind[k*j + i] = Vk[k*indcols[j] + i];
    }
  }
  *out_rank = rank;

  // Put coeff entries into a contiguous matrix
  // TODO: replace this with a solve, coeff must be rank x p - rank
  //lusolve(int n, int nrhs, const double *A, double *x, NULL, NULL);
  for(j = 0; j < p - rank; j++) {
    for(i = 0; i < rank; i++) {
      coeff[k*j + i] = coeff[rank*depcols[j] + indcols[i]];
    }
  }
}

/**
 * Compute Vk from the linearly independent columns
 */
static void compute_VK(struct stiefel_ctx *ctx, int rank, const int *indcols, const int *depcols, double *Vk, const double *Vkind, const double *coeff)
{
  int i, j, k, p;
  double *Vkdep;

  k = ctx->k;
  p = ctx->p;
  Vkdep = ctx->Vkdep;

  // Place independent into Vk
  for(j = 0; j < rank; j++) {
    for(i = 0; i < k; i++) {
      Vk[indcols[j] * k + i] = Vkind[j*k + i];
    }
  }

  // Place dependent columns into Vk
  dgemm_wrap('N', 'N', k, p - rank, rank, 1, Vkind, k, coeff, rank, 0, Vkdep, k);
  for(j = 0; j < p - rank; j++) {
    for(i = 0; i < k; i++) {
      Vk[depcols[j] * k + i] = Vkdep[j*k + i];
    }
  }
}
