#include <stdlib.h>
#include <assert.h>

#include "util/linalg.h"
#include "util/debug.h"
#include "util/geom.h"

#include "cayley.h"

/**
 * Add the identity matrix to n x n matrix @c 
 */
static void add_identity(int n, double *X);

void stiefel_cayley_general(int m, int n, double tau, const double *W, const cs_di *M, const double *X0, double *X)
{
  int i, j, k, l;
  double *WM, *Ql, *Qr;

  // Check sizes
  assert(M->m == m && M->n == m);

  // Allocate workspace
  WM = (double *)malloc(m * m * sizeof(double) * 3);
  Ql = WM + m*m;
  Qr = Ql + m*m;

  // Calculate product W * M
  for(j = 0; j < m; j++) {
    for(i = 0; i < m; i++) {
      WM[m*j + i] = 0;

      for(l = M->p[j]; l < M->p[j+1]; l++) {
        k = M->i[l];
        WM[m*j + i] += W[m*k + i] * M->x[l];
      }
    }
  }
  assert(!anynan(m*m, WM));

  // Calculate Q matrices
  for(i = 0; i < m*m; i++) {
    Ql[i] = (tau/2) * WM[i];
    Qr[i] = -Ql[i];
  }
  add_identity(m, Ql);
  add_identity(m, Qr);

  // Calculate X <- Ql \ (Qr * X0)
  dgemm_wrap('N', 'N', m, n, m, 1, Qr, m, X0, m, 0, X, m);
  lusolve(m, n, Ql, X, NULL, NULL);

  assert(!anynan(m*n, X));

  // Cleanup
  free(WM);
}

static void quadform(const cs_di *M, int n, const double *X, double *XTMX)
{
  int i, j, k, m, u, v;
  double Mij;

  assert(M->m == M->n);

  m = M->m;

  for(i = 0; i < n*n; i++) {
    XTMX[i] = 0;
  }

  for(j = 0; j < m; j++) {
    for(k = M->p[j]; k < M->p[j+1]; k++) {
      i = M->i[k];
      Mij = M->x[k];

      for(v = 0; v < n; v++) {
        for(u = 0; u < n; u++) {
          XTMX[u + n*v] += Mij * X[m*i + u] * X[m*j + v];
        }
      }
    }
  }
}

double cayley_check(const cs_di *M, int n, const double *X0, const double *X)
{
  int i, j;
  double *XTMX0, *XTMX, err;
  
  XTMX0 = (double *)malloc(sizeof(double) * 2 * n * n);
  XTMX = XTMX0 + n*n;

  quadform(M, n, X0, XTMX0);
  quadform(M, n, X, XTMX);

  err = geom_linftydistance((size_t)n*n, XTMX0, XTMX);

  free(XTMX0);

  return err;
}

static void add_identity(int n, double *X)
{
  int i;

  for(i = 0; i < n; i++) {
    X[n*i + i] += 1.0;
  }
}
