#include <assert.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <pthread.h>

#include "client.h"
#include "msg.h"
#include "colmsg.h"
#include "msgbuf.h"

#include "util/debug.h"
#include "util/lazymat.h"
#include "clientterms.h"

#include "genorth.h"
#include "trace.h"

struct client_state;

struct client_matrices
{
  int nterms;
  termdef **objdef;

  lazymat *B;

  int loaded;
  pthread_rwlock_t lock;
};

static void client_matrices_init(struct client_state *st, int nterms, int *objtypes, int sock);
static void client_matrices_prerecv(struct client_state *st, int sock);

static intarray *missing_columns(client_matrices *cm, int k, const int *K);

struct client_state
{
  int nterms, n;
  int k, p;

  config_params params;
  client_matrices *cm;

  rowset *Vk;
  rowset *VnbrA;
  rowset *VnbrB;

  genorth_ctx *ctx;
};


static double obj(void *user_data, const rowset *Vk);
static void grad(void *user_data, const rowset *Vk, double *Gk);
static double fullobj(void *user_data, int n, int p, const double *V);

double client_iter(struct client_state *st, struct itermsg *iter, double *Vout)
{
  int n, k, p;
  double *Gk, chg;

  // Unpack messages
  n = st->n;
  k = rowset_nrows(iter->Vk);
  p = rowset_ncols(iter->Vk);
  assert(p == st->p && k <= st->k);

  st->Vk = iter->Vk;
  st->VnbrA = iter->VnbrA;
  st->VnbrB = iter->VnbrB;

  pthread_rwlock_rdlock(&st->cm->lock);

  // Calculate gradient
  Gk = (double *)malloc(sizeof(double) * k * p);
  grad(st, iter->Vk, Gk);

  // Perform step
  chg = genorth_step(st->ctx, iter->Vk, iter->VnbrB, Gk, Vout);

  pthread_rwlock_unlock(&st->cm->lock);

  // Cleanup
  free(Gk);

  return chg;
}


int client_run_socket(int sock, client_matrices *cm)
{
  int i, j, k, s, p;

  ssize_t read;
  int32_t iterlen;
  msgbuf *buf;

  intarray *missing_cols;
  double chg, *Vout;

  struct setupmsg *setup;
  struct itermsg *iter;
  struct colmsg *cols;

  struct client_state st;
  genorth_prob prob;
  genorth_ctx *ctx;

  printf("Starting up client... (%d)\n", sock);

  buf = msgbuf_create();

  // Read setup message
  s = msgbuf_recv(buf, sock);
  if(s != 0) {
    return s;
  }
  setup = setupmsg_deserialize(msgbuf_getlen(buf), msgbuf_data(buf));
  assert(setup != NULL);

  // Initialize client state
  st.nterms = setup->nterms;
  st.n = setup->n;
  st.k = k = setup->k;
  st.p = p = setup->p;
  st.params = setup->params;
  st.Vk = NULL;
  st.VnbrA = NULL;
  st.VnbrB = NULL;

  st.cm = cm;
  client_matrices_init(&st, st.nterms, setup->objtypes, sock);

  // Set up Stiefel problem
  prob.n = setup->n;
  prob.k = setup->k;
  prob.p = setup->p;

  prob.obj = obj;
  prob.grad = grad;
  prob.fullobj = fullobj;
  prob.user_data = &st;

  prob.B = (sparsemat *)cm->B;

  prob.params = setup->params.optim;

  st.ctx = ctx = genorth_ctx_init(&prob);

  Vout = (double *)malloc(sizeof(double) * k * p);

  printf("Client setup complete (%d)\n", sock);

  // Client loop
  while(1) {
    // Read iteration message
    s = msgbuf_recv(buf, sock);
    if(s != 0) {
      break;
    }
    if(msgbuf_getlen(buf) == 0) {
      printf("Termination message recieved by client.\n");
      break;
    }
    iter = itermsg_deserialize(msgbuf_getlen(buf), msgbuf_data(buf));
    k = rowset_nrows(iter->Vk);
    //assert(k == rowset_nrows(iter->Vk));
 
    // Request missing columns
    // TODO: request all dependent columns?
    missing_cols = missing_columns(st.cm, k, rowset_inds(iter->Vk));
    //fprintf(stderr, "Requesting columns: "); print_ints(stderr, nmissing, missing_cols); fprintf(stderr, "\n");

    s = msgbuf_senddata(sock, sizeof(int) * intarray_length(missing_cols), intarray_contents_const(missing_cols));
    if(s != 0) {
      break;
    }

    // Recieve columns for objective
    s = msgbuf_recv(buf, sock);
    if(s != 0) {
      break;
    }

    s = sumterm_indupdate_deserialize(cm->nterms, cm->objdef, msgbuf_getlen(buf), msgbuf_data(buf), intarray_length(missing_cols), intarray_contents_const(missing_cols));
    assert(s == 0);

    // Receive columns for constraints
    s = msgbuf_recv(buf, sock);
    assert(s == 0);

    s = colmsg_deserialize(cm->B, intarray_length(missing_cols), intarray_contents_const(missing_cols), msgbuf_getlen(buf), msgbuf_data(buf));
    assert(s == 0);

    intarray_destroy(missing_cols);
    
    // Perform iteration
    //printf("Client iterating\n");
    chg = client_iter(&st, iter, Vout);
    
    // Send Vk back to server
    s = msgbuf_senddata(sock, sizeof(double) * (size_t)(k * p), Vout);
    if(s != 0) {
      break;
    }

    s = msgbuf_senddata(sock, sizeof(double), &chg);
    if(s != 0) {
      break;
    }

    itermsg_dealloc(iter);
  }

  // Cleanup
  free(Vout);
  msgbuf_destroy(buf);
  genorth_ctx_deinit(ctx);
  setupmsg_dealloc(setup);

  return s;
}

static double obj(void *user_data, const rowset *Vk)
{
  struct client_state *st;
  client_matrices *cm;
  double obj;

  st = (struct client_state *)user_data;
  cm = st->cm;

  //pthread_rwlock_rdlock(&cm->lock);
  obj = sumterm_obj(cm->nterms, (const termdef * const *)cm->objdef, Vk, st->VnbrA);
  //pthread_rwlock_unlock(&cm->lock);

  return obj;
}

static void grad(void *user_data, const rowset *Vk, double *Gk)
{
  struct client_state *st;
  client_matrices *cm;
  
  st = (struct client_state *)user_data;
  cm = st->cm;

  //pthread_rwlock_rdlock(&cm->lock);
  sumterm_grad(cm->nterms, (const termdef * const *)cm->objdef, Vk, st->VnbrA, Gk);
  //pthread_rwlock_unlock(&cm->lock);
}

static double fullobj(void *user_data, int n, int p, const double *V)
{
  assert(0);
}

///////////////////////// Shared Matrices //////////////////////

client_matrices *client_matrices_alloc()
{
  client_matrices *cm;
  pthread_rwlockattr_t lockattr;

  cm = (client_matrices *)malloc(sizeof(client_matrices));
  cm->nterms = 0;
  cm->objdef = NULL;
  cm->B = NULL;
  cm->loaded = 0;

  // Initialize lock
  pthread_rwlock_init(&cm->lock, NULL);


  return cm;
}

void client_matrices_destroy(client_matrices *cm)
{
  int i;

  if(cm == NULL) {
    return;
  }

  pthread_rwlock_destroy(&cm->lock);
  if(cm->objdef != NULL) {
    for(i = 0; i < cm->nterms; i++) {
      term_destroy(cm->objdef[i]);
    }
    free(cm->objdef);
  }
  sparsemat_destroy((sparsemat *)cm->B);
  free(cm);
}

static void client_matrices_init(struct client_state *st, int nterms, int *objtypes, int sock)
{
  int i, dims[2], n, z, s;
  client_matrices *cm;
  msgbuf *buf;

  cm = st->cm;
  n = st->n;
  assert(nterms == st->nterms);

  z = 0;

  while(pthread_rwlock_trywrlock(&cm->lock) != 0) {
    if(cm->loaded) {
      assert(cm->nterms == nterms);
      s = msgbuf_senddata(sock, sizeof(int), &z);
      assert(s == 0);

      client_matrices_prerecv(st, sock);
      return;
    }
  }
  //pthread_rwlock_wrlock(&cm->lock);

  if(cm->objdef == NULL) {
    // Not already initialized by another client

    dims[0] = dims[1] = n;

    cm->nterms = nterms;
    cm->objdef = (termdef **)malloc(sizeof(termdef *) * nterms);
    for(i = 0; i < nterms; i++) {
      cm->objdef[i] = client_term_create(objtypes[i], n);
    }
    
    cm->B = lazymat_create(dims);
    assert(cm->B != NULL);

    // Get setup messages
    //fprintf(stderr, "Sending # requested terms\n");
    s = msgbuf_senddata(sock, sizeof(int), &cm->nterms);
    assert(s == 0);

    buf = msgbuf_create();

    for(i = 0; i < cm->nterms; i++) {
      s = msgbuf_recv(buf, sock);
      s = s || term_setup_deserialize(cm->objdef[i], msgbuf_getlen(buf), msgbuf_data(buf));
      assert(s == 0);
    }

    msgbuf_destroy(buf);
  }
  else {
    assert(cm->nterms == nterms);

    fprintf(stderr, "Sending no requested terms\n");
    s = msgbuf_senddata(sock, sizeof(int), &z);
    assert(s == 0);
  }

  // Prerecieve matrices
  client_matrices_prerecv(st, sock);

  cm->loaded = 1;
  pthread_rwlock_unlock(&cm->lock);
}

static void client_matrices_prerecv(struct client_state *st, int sock)
{
  msgbuf *buf;
  int i, ncols, *cols, s;
  struct colmsg *msg;
  client_matrices *cm;

  cm = st->cm;

  // Request all or no columns, depending on status
  if(st->params.presend_matrices && !cm->loaded) {
    ncols = st->n;
    cols = (int *)malloc(sizeof(int) * ncols);
    for(i = 0; i < ncols; i++) {
      cols[i] = i;
    }
  }
  else {
    ncols = 0;
    cols = NULL;
  }

  // Send list of requested columns
  s = msgbuf_senddata(sock, sizeof(int) * (size_t)ncols, cols);
  assert(s == 0);
  
  // Recieve columns for objective
  buf = msgbuf_create();
  s = msgbuf_recv(buf, sock);
  assert(s == 0);

  s = sumterm_indupdate_deserialize(cm->nterms, cm->objdef, msgbuf_getlen(buf), msgbuf_data(buf), ncols, cols);
  assert(s == 0);

  // Receive columns for constraints
  s = msgbuf_recv(buf, sock);
  assert(s == 0);

  s = colmsg_deserialize(cm->B, ncols, cols, msgbuf_getlen(buf), msgbuf_data(buf));
  assert(s == 0);

  // Cleanup
  free(cols);
  msgbuf_destroy(buf);
}

static intarray *missing_columns(client_matrices *cm, int k, const int *K)
{
  int i, n;
  const lazymat *B;
  int *Bmissing;
  intarray *missing, *unique;

  assert(cm->nterms > 0);

  pthread_rwlock_rdlock(&cm->lock);

  missing = sumterm_indupdate_request(cm->nterms, (const termdef * const *)cm->objdef, k, K);
  assert(missing != NULL);

  // Missing contraint columns
  B = cm->B;
  Bmissing = (int *)malloc(sizeof(int) * k);
  n = 0;
  for(i = 0; i < k; i++) {
    if(sparsemat_col((sparsemat *)B, K[i]).nnz == -1) {
      Bmissing[n++] = K[i];
    }
  }
  intarray_append(missing, n, Bmissing);

  pthread_rwlock_unlock(&cm->lock);

  unique = intarray_unique(missing);
  intarray_destroy(missing);

  return unique;
}
