#include <assert.h>
#include <stdint.h>

#include "clientterms.h"
#include "trace.h"
#include "sqrtterm.h"
#include "colmsg.h"

#include "util/lazymat.h"

/////////////////////////////// Trace terms ////////////////////////

static double client_traceterm_obj(void *user_data, const rowset *V, const rowset *Vnbr)
{
  return trace_partial_obj((sparsemat *)user_data, V, Vnbr);
}

static void client_traceterm_grad(void *user_data, const rowset *V, const rowset *Vnbr, double *Gk)
{
  return trace_partial_grad((sparsemat *)user_data, V, Vnbr, Gk);
}

static intarray *client_traceterm_indupdate_request(void *user_data, int n, const int *inds)
{
  sparsemat *A;
  int i, m, *missing;

  A = (sparsemat *)user_data;

  missing = (int *)malloc(sizeof(int) * n);
  m = 0;
  for(i = 0; i < n; i++) {
    if(sparsemat_col(A, inds[i]).nnz == -1) {
      missing[m++] = inds[i];
    }
  }
  missing = (int *)realloc(missing, sizeof(int) * m);

  return intarray_create_from(m, missing);
}

static int client_traceterm_indupdate_deserialize(void *user_data, size_t buflen, const void *buf, int n, const int *inds)
{
  return colmsg_deserialize((lazymat *)user_data, n, inds, buflen, buf);
}

static void client_traceterm_destroy(void *user_data)
{
  sparsemat_destroy((sparsemat *)user_data);
}

void client_traceterm_create(termdef *def, int n)
{
  int dims[2];

  // Set up methods
  def->obj = client_traceterm_obj;
  def->grad = client_traceterm_grad;

  def->dependent = NULL;
  def->indupdate_request = client_traceterm_indupdate_request;

  def->setup_buflen = NULL;
  def->setup_serialize = NULL;
  def->setup_deserialize = NULL;

  def->indupdate_buflen = NULL;
  def->indupdate_serialize = NULL;
  def->indupdate_deserialize = client_traceterm_indupdate_deserialize;

  def->destroy = client_traceterm_destroy;

  // Allocate matrix
  dims[0] = dims[1] = n;
  def->user_data = lazymat_create(dims);
  def->mult = 1;
  def->type = OBJTERM_TYPE_TRACE;

}

/////////////////////////////// sqrt terms ////////////////////////////

static double client_sqrtterm_obj(void *user_data, const rowset *V, const rowset *Vnbr)
{
  const mustlink *ml;

  ml = *(const mustlink **)user_data;
  assert(ml != NULL);

  return sqrt_obj(ml, V, Vnbr);
}

static void client_sqrtterm_grad(void *user_data, const rowset *V, const rowset *Vnbr, double *Gk)
{
  const mustlink *ml;

  ml = *(const mustlink **)user_data;
  assert(ml != NULL);

  sqrt_grad(ml, V, Vnbr, Gk);
}

static int client_sqrtterm_deserialize(void *user_data, size_t buflen, const void *buf)
{
  mustlink *ml;

  ml = mustlink_deserialize(buflen, buf);
  *(mustlink **)user_data = ml;

  if(ml == NULL) {
    return -1;
  }
  else {
    return 0;
  }
}

static intarray *client_sqrtterm_indupdate_request(void *user_data, int n, const int *inds)
{
  intarray *req;

  req = intarray_create(0);
  // TODO: can retrieve request from dependent columns given in iteration message?

  return req;
}

static int client_sqrtterm_indupdate_deserialize(void *user_data, size_t buflen, const void *buf, int n, const int *inds)
{
  assert(buflen == 0);
  // TODO?
  return 0;
}

static void client_sqrtterm_destroy(void *user_data)
{
  mustlink_destroy(*(mustlink **)user_data);
  free(user_data);
}

void client_sqrtterm_create(termdef *def, int n)
{
  def->obj = client_sqrtterm_obj;
  def->grad = client_sqrtterm_grad;

  def->dependent = NULL;

  def->setup_buflen = NULL;
  def->setup_serialize = NULL;
  def->setup_deserialize = client_sqrtterm_deserialize;

  def->indupdate_request = client_sqrtterm_indupdate_request;
  def->indupdate_buflen = NULL;
  def->indupdate_serialize = NULL;
  def->indupdate_deserialize = client_sqrtterm_indupdate_deserialize;

  def->destroy = client_sqrtterm_destroy;

  def->type = OBJTERM_TYPE_SQRT;
  def->user_data = malloc(sizeof(mustlink *));
  *(mustlink **)def->user_data = NULL;
  def->mult = 1;
}

/////////////////////////// Top-level constructor //////////////////////


termdef *client_term_create(int type, int n)
{
  termdef *def;

  def = (termdef *)malloc(sizeof(termdef));
  if(def == NULL) {
    return NULL;
  }

  switch(type) {
  case OBJTERM_TYPE_TRACE:
    client_traceterm_create(def, n);
    break;

  case OBJTERM_TYPE_SQRT:
    client_sqrtterm_create(def, n);
    break;

  default:
    assert(0);
    break;
  }

  return def;
}
