#include <assert.h>
#include <memory.h>
#include <stdint.h>

#include "colmsg.h"
#include "serialize.h"

#include "util/debug.h"

size_t colmsg_buflen(sparsemat *A, int n, const int *cols)
{
  int i, totalnnz;
  sparsecol col;

  totalnnz = 0;
  for(i = 0; i < n; i++) {
    //fprintf(stderr, "cols[%d] = %d\n", i, cols[i]); 
    col = sparsemat_col(A, cols[i]);
    //fprintf(stderr, "col.nnz = %d\n", col.nnz);
    assert(col.nnz >= 0);
    totalnnz += col.nnz;
  }

  return serial_length(SERIALIZE_INT,
                       SERIALIZE_INT_VECTOR, n,
                       SERIALIZE_INT_VECTOR, totalnnz,
                       SERIALIZE_DOUBLE_VECTOR, totalnnz,
                       SERIALIZE_END);
  
}

int colmsg_serialize(sparsemat *A, int n, const int *cols, size_t buflen, void *buf)
{
  int j, k, totalnnz, s;
  int *colnnz, *i;
  double *v;
  sparsecol col;

  //fprintf(stderr, "Serializing colmsg for:");  print_ints(stderr, n, cols); fprintf(stderr, "\n");

  colnnz = (int *)malloc(sizeof(int) * n);
  if(colnnz == NULL) {
    return -1;
  }

  // Get nonzero count of each column
  totalnnz = 0;
  for(j = 0; j < n; j++) {
    col = sparsemat_col(A, cols[j]);
    colnnz[j] = col.nnz;
    totalnnz += col.nnz;
  }

  // Assemble column data
  i = (int *)malloc(sizeof(int) * totalnnz);
  v = (double *)malloc(sizeof(double) * totalnnz);
  if(i == NULL || v == NULL) {
    free(colnnz);
    free(i);
    free(v);
    return -1;
  }

  k = 0;
  for(j = 0; j < n; j++) {
    col = sparsemat_col(A, cols[j]);
    if(col.nnz == -1) {
      free(colnnz);
      free(i);
      free(v);
      return -1;
    }

    assert(k + col.nnz <= totalnnz);

    memcpy(i + k, col.i, sizeof(int) * col.nnz);
    memcpy(v + k, col.v, sizeof(double) * col.nnz);

    k += col.nnz;
  }
  assert(k == totalnnz);

  // Serialize
  s = serialize(buflen, buf,
                SERIALIZE_INT, n,
                SERIALIZE_INT_VECTOR, n, colnnz,
                SERIALIZE_INT_VECTOR, totalnnz, i,
                SERIALIZE_DOUBLE_VECTOR, totalnnz, v,
                SERIALIZE_END);

  // Cleanup
  free(colnnz);
  free(i);
  free(v);

  return s;
}

int colmsg_deserialize(lazymat *A, int n, const int *cols, size_t buflen, const void *buf)
{
  int j, k;
  int32_t *colnnz, totalnnz, *i;
  double *v;
  uintptr_t p, pend;
  
  p = (uintptr_t)buf;
  pend = p + buflen;

  // Check column count
  if(buflen < 4 || ((int32_t *)p)[0] != (int32_t)n) {
    return -1;
  }
  p += 4;

  // Get nonzero counts
  if(p + 4 * n > pend) {
    return -1;
  }
  colnnz = (int32_t *)p;
  p += 4 * n;

  totalnnz = 0;
  for(j = 0; j < n; j++) {
    totalnnz += colnnz[j];
  }

  // Get points to column data
  i = (int32_t *)p;
  if(p + 4 * totalnnz > pend) {
    return -1;
  }
  p += 4 * totalnnz;

  v = (double *)p;
  if(p + 8 * totalnnz > pend) {
    return -1;
  }
  p += 8 * totalnnz;

  if(p != pend) {
    return -1;
  }

  // Add columns
  k = 0;
  for(j = 0; j < n; j++) {
    lazymat_col_add(A, cols[j], colnnz[j], i + k, v + k);
    k += colnnz[j];
  }
  assert(k == totalnnz);

  return 0;
}
