#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "config.h"

static void strip(char *out, const char *in);

/**
 * Get the length of a var array
 */
static int nvars(const config_var *vars)
{
  int n;

  for(n = 0; vars[n].name != NULL; n++);

  return n;
}

static void readvar(const char *name, const char *value, const config_var *vars, void **outptrs)
{
  int i, n, ind, vlen, read;
  config_type type;
  char *outstr;
  void *out;

  // Find variable index
  ind = -1;
  n = nvars(vars);
  for(i = 0; i < n; i++) {
    if(strcmp(vars[i].name, name) == 0) {
      ind = i;
      break;
    }
  }
  if(ind == -1) {
    fprintf(stderr, "Unrecognized config variable \"%s\"\n", name);
  }

  type = vars[ind].type;
  out = outptrs[ind];

  // Interpret variable
  switch(type) {
  case CONFIG_INT:
    read = sscanf(value, "%d", (int *)out);
    if(read != 1) {
      fprintf(stderr, "Malformed int argument %s\n", value);
    }
    break;
  case CONFIG_STRING:
    vlen = strlen(value);
    if(value[0] != '"' || value[vlen-1] != '"') {
      fprintf(stderr, "Malformed string argument %s\n", value);
      return;
    }

    outstr = (char *)malloc(vlen - 1);
    strncpy(outstr, value + 1, vlen - 2);
    outstr[vlen-2] = '\0';

    *(char **)out = outstr;
    break;
  case CONFIG_FLOAT:
    read = sscanf(value, "%lg", (double *)out);
    if(read != 1) {
      fprintf(stderr, "Malformed float argument %s\n", value);
    }
    break;
  case CONFIG_BOOL:
    if(strcmp(value, "true") == 0) {
      *(bool *)out = true;
    }
    else if(strcmp(value, "false") == 0) {
      *(bool *)out = false;
    }
    else {
      fprintf(stderr, "Invalid boolean argument %s\n", value);
    }
  }
}

static void config_readline(char *line, const config_var *vars, void **out)
{
  char *name, *value, *eq;
  int line_len;

  // Find =
  for(eq = line; *eq != '=' && *eq != '\0'; eq++);
  if(eq == '\0') {
    fprintf(stderr, "Malformed line \"%s\"\n", line);
    return;
  }

  *eq = '\0';

  line_len = strlen(line);
  name = (char *)malloc(line_len);
  value = (char *)malloc(line_len);

  strip(name, line);
  strip(value, eq+1);
  
  readvar(name, value, vars, out);

  // Cleanup
  free(name);
  free(value);
}

void config_read_file(FILE *fp, const config_var *vars, void **out)
{
  char *line, *stripline, *c;
  size_t line_len, alloc;

  line_len = 0;
  line = NULL;
  alloc = 0;
  stripline = NULL;

  do {
    // Read in a line
    line_len = getline(&line, &line_len, fp);
    if((int)line_len == -1) {
      break;
    }

    // End line at a comment
    for(c = line; *c != '#' && *c != '\0'; c++);
    *c = '\0';
    line_len = c - line;

    // Strip whitespace
    if(line_len + 1 > alloc) {
      alloc = line_len + 1;
    }
    stripline = realloc(stripline, alloc);

    strip(stripline, line);

    // Empty line
    if(stripline[0] == '\0') {
      continue;
    }

    config_readline(stripline, vars, out);

  } while(!feof(fp));

  // Cleanup
  free(stripline);
  free(line);
}

void config_read_path(const char *path, const config_var *vars, void **out)
{
  FILE *fp;

  fp = fopen(path, "r");
  config_read_file(fp, vars, out);
  fclose(fp);
}

static void strip(char *out, const char *in)
{
  const char *start, *end;
  
  end = in + strlen(in);
  for(start = in; start != end && isspace(*start); start++);
  for(end--; end > start && isspace(*end); end--);
  end++;

  strncpy(out, start, end - start);
  out[end-start] = '\0';
}
