#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <stdint.h>

#include "util/geom.h"
#include "util/debug.h"

#include "util/linalg.h"
#include "util/randsample.h"
#include "util/setdiff.h"
#include "util/sparse_submatrix.h"
#include "util/indcols.h"
#include "util/neighbors.h"
#include "util/sort.h"

#include "cayley.h"
#include "infeas.h"

#include "genorth.h"

/**
 * prod <- prod + A*B
 */
static void axpy_spmat_rowset(char transA, cs_di *A, const rowset *B, double *prod);
static double line_search_obj(genorth_ctx *ctx, double tau);

struct genorth_ctx {
  // Problem parameters
  int n, p, k, maxk;

  double (*obj)(void *user_data, const rowset *Vk);
  void (*grad)(void *user_data, const rowset *Vk, double *Gk);
  void *user_data;

  sparsemat *B;

  optim_params params;

  // Workspace and intermediate variables
  int iter;

  const int *K;
  double *VK, *nbr_term;

  int rank, *indcols, *depcols;
  double *R;

  double *U0;
  double *Utau;
  double *Vtauind, *Vtaudep, *Vtau;
  double *Vout;

  double *G, *Gind, *Gdep;
  double *A;

  cs_di *BKK;

  double dF;
  double tau;
  double objtau;

};

genorth_ctx *genorth_ctx_init(const genorth_prob *prob)
{
  genorth_ctx *ctx;
  int k, p;

  ctx = (genorth_ctx *)malloc(sizeof(genorth_ctx));
  if(ctx == NULL) {
    return NULL;
  }

  // Copy problem data
  ctx->n = prob->n;
  ctx->p = p = prob->p;
  ctx->maxk = ctx->k = k = prob->k;

  ctx->obj = prob->obj;
  ctx->grad = prob->grad;
  ctx->user_data = prob->user_data;

  ctx->B = prob->B;

  ctx->params = prob->params;

  // Allocate workspace
  ctx->VK = (double *)malloc(sizeof(double) * k * p);
  ctx->nbr_term = (double *)malloc(sizeof(double) * k * p);

  ctx->indcols = (int *)malloc(sizeof(int) * p);
  ctx->depcols = (int *)malloc(sizeof(int) * p);
  ctx->R = (double *)malloc(sizeof(double) * p * p);

  ctx->U0 = (double *)malloc(sizeof(double) * k * p);
  ctx->Utau = (double *)malloc(sizeof(double) * k * p);
  ctx->Vtauind = (double *)malloc(sizeof(double) * k * p);
  ctx->Vtaudep = (double *)malloc(sizeof(double) * k * p);
  ctx->Vtau = (double *)malloc(sizeof(double) * k * p);

  ctx->Gind = (double *)malloc(sizeof(double) * k * p);
  ctx->Gdep = (double *)malloc(sizeof(double) * k * p);

  ctx->A = (double *)malloc(sizeof(double) * k * k);

  return ctx;
}

void genorth_ctx_deinit(genorth_ctx *ctx)
{
  if(ctx == NULL) {
    return;
  }

  free(ctx->VK);
  free(ctx->nbr_term);
  free(ctx->indcols);
  free(ctx->depcols);
  free(ctx->R);
  free(ctx->U0);
  free(ctx->Utau);
  free(ctx->Vtauind);
  free(ctx->Vtaudep);
  free(ctx->Vtau);
  free(ctx->Gind);
  free(ctx->Gdep);
  free(ctx->A);

  free(ctx);
}

double genorth_step(genorth_ctx *ctx, const rowset *V, const rowset *Vnbr, const double *Gk, double *Vout)
{
  cs_di *BKK, *BCK;
  int i, j, p, k, nnbr, rank;
  double *U0, *VK, *A;
  double dF, tau, objtau, obj0;


  p = ctx->p;
  k = rowset_nrows(V);
  nnbr = rowset_nrows(Vnbr);

  //fprintf(stderr, "norm(G) = %lg\n", geom_l2norm(k * p, Gk));

  ctx->k = k;
  assert(k <= ctx->maxk);

  ctx->BKK = BKK = sparse_submatrix(ctx->B, k, rowset_inds(V), k, rowset_inds(V));
  BCK = sparse_submatrix(ctx->B, nnbr, rowset_inds(Vnbr), k, rowset_inds(V));

  VK = ctx->VK;
  rowset_getall_colmajor(V, VK);
  ctx->K = rowset_inds(V);

  ctx->Vout = Vout;

  // Calculate V_K + B_KK \ (B_CK' * V_C)
  axpy_spmat_rowset('T', BCK, Vnbr, ctx->nbr_term);
  for(j = 0; j < p; j++) {
    cs_di_lusol(1, BKK, ctx->nbr_term + k*j, 1e-8); // TODO: multiple columns
  }

  U0 = ctx->U0;
  for(i = 0; i < k * p; i++) {
    U0[i] = VK[i] + ctx->nbr_term[i];
  }
  assert(!anynan(k*p, U0));

  // Separate out into independent and dependent columns
  indcols(k, p, U0, &rank, ctx->indcols, ctx->depcols, ctx->R);
  select_cols(k, p, rank, ctx->indcols, U0, U0);
  ctx->rank = rank;
  assert(!anynan(k*rank, U0));

  // Restrict gradient to U space
  select_cols(k, p, rank, ctx->indcols, Gk, ctx->Gind);
  if(rank < p) {
    select_cols(k, p, p-rank, ctx->depcols, Gk, ctx->Gdep);
    dgemm_wrap('N', 'T', k, rank, p - rank, 1, ctx->Gdep, k, ctx->R, rank, 1, ctx->Gind, k);
  }

  // Calculate A = G*U' - U*G' 
  A = ctx->A;
  dgemm_wrap('N', 'T', k, k, rank, 1, ctx->Gind, k, U0, k, 0, A, k);
  dgemm_wrap('N', 'T', k, k, rank, -1, U0, k, ctx->Gind, k, 1, A, k);
  assert(!anynan(k*rank, A));

  // Line search
  obj0 = ctx->obj(ctx->user_data, V);

  dF = 0;
  for(i = 0; i < k*rank; i++) {
    dF += A[i]*A[i];
  }
  dF *= -0.5;
  ctx->dF = dF;

  tau = 1e1;
  objtau = line_search_obj(ctx, tau);
  //printf("objdiff = %lg\n", obj0 - objtau);
  //exit(0);
  while(objtau > obj0 + ctx->params.rho1 * tau * dF) {
    tau = tau/2;
    objtau = line_search_obj(ctx, tau);
    
    if(tau < 1e-3) {
      break;
    }
  }

  ctx->tau = tau;
  ctx->objtau = objtau;

  assert(!anynan(k*p, Vout));

  return objtau - obj0;
}

static double line_search_obj(genorth_ctx *ctx, double tau)
{
  int i, j, k, p, rank;
  double obj;

  rowset *rwst;

  k = ctx->k;
  p = ctx->p;
  rank = ctx->rank;

  //printf("tau = %lg\n", tau);

  // Generate rotation
  stiefel_cayley_general(k, rank, tau, ctx->A, ctx->BKK, ctx->U0, ctx->Vtauind);
  assert(!anynan(k*rank, ctx->Vtauind));

  // Create overall rows
  dgemm_wrap('N', 'N', k, p - rank, rank, 1, ctx->Vtauind, k, ctx->R, rank, 0, ctx->Vtaudep, k);

  install_cols(k, p, rank, ctx->indcols, ctx->Vtau, ctx->Vtauind);
  install_cols(k, p, p - rank, ctx->depcols, ctx->Vtau, ctx->Vtaudep);

  for(i = 0; i < k*p; i++) {
    ctx->Vtau[i] -= ctx->nbr_term[i];
  }

  // Construct row-major rowset
  for(i = 0; i < k; i++) {
    for(j = 0; j < p; j++) {
      ctx->Vout[p*i + j] = ctx->Vtau[i + k*j];
    }
  }

  // Calculate objective
  rwst = rowset_create(ctx->n, ctx->p, ctx->k, ctx->K, ctx->Vout);
  obj = ctx->obj(ctx->user_data, rwst);
  rowset_destroy(rwst);

  return obj;
}

static void axpy_spmat_rowset(char transA, cs_di *A, const rowset *Br, double *prod)
{
  int i, j, k, ind, m, n, p;
  double *B, v;

  assert(transA == 'N' || transA == 'T');

  // Get sizes
  m = transA == 'N' ? A->m : A->n;
  p = transA == 'N' ? A->n : A->m;
  n = rowset_ncols(Br);
  assert(p == rowset_nrows(Br));

  for(i = 0; i < m*n; i++) {
    prod[i] = 0;
  }

  // Extract contents of B
  B = (double *)malloc(sizeof(double) * p * n);
  assert(B != NULL);
  rowset_getall_colmajor(Br, B);


  // Perform multiplication
  if(transA == 'N') {
    for(j = 0; j < n; j++) {
      cs_di_gaxpy(A, B + p*j, prod + m*j);
    }
  }
  else {
    assert(transA == 'T');

    for(i = 0; i < m; i++) {
      for(ind = A->p[i]; ind < A->p[i+1]; ind++) {
        k = A->i[ind];
        v = A->x[ind];

        for(j = 0; j < n; j++) {
          prod[i + m*j] += v * B[k + p*j];
        }
      }
    }
  }

  // Cleanup
  free(B);
}


//////////////////////// Basic single-threaded solver ////////////////////

static const char *status_header = "\n\
 iter           |dF|           dobj            tau\n\
";

static void print_status(struct genorth_ctx *ctx, int iter, double objchg)
{
  printf("%5d   %le   %le   %le\n", iter, -ctx->dF, objchg, ctx->tau);
}

void genorth_solve(const genorth_prob *prob, double *V)
{
  struct genorth_ctx *ctx;
  int i, j, iter, n, p, k, *K, nnbr, *nbr;
  const optim_params *params;
  double *Gk, *Vout, objchg;
  rowset *Vk, *Vnbr;

  // Unpack problems
  n = prob->n;
  p = prob->p;
  k = prob->k;
  params = &prob->params;

  // Allocate workspace
  ctx = genorth_ctx_init(prob);

  K = (int *)malloc(sizeof(int) * k);
  Gk = (double *)malloc(sizeof(double) * k * p);
  Vout = (double *)malloc(sizeof(double) * k * p);

  if(params->print_interval > 0) {
    fputs(status_header, stdout);
  }

  objchg = 0;
  for(iter = 0; iter < params->max_iter; iter++) {
    ctx->iter = iter;

    // Randomly pick indices
    randsample(K, n, k, 0);
    sort_ints((size_t)k, K);
    Vk = rowset_extract(n, p, k, K, V);

    nbr = sparsemat_neighbors(1, &prob->B, k, K, &nnbr);
    Vnbr = rowset_extract(n, p, nnbr, nbr, V);
    free(nbr);

    // Get gradient
    prob->grad(prob->user_data, Vk, Gk);

    // Take a step
    objchg += genorth_step(ctx, Vk, Vnbr, Gk, Vout);

    rowset_destroy(Vk);
    rowset_destroy(Vnbr);

    // Copy back into V
    Vk = rowset_create(n, p, k, K, Vout);
    rowset_install(V, Vk);

    if(-ctx->dF < params->grad_epsilon) {
      //break;
    }

    // Print status
    if(params->print_interval > 0 && iter % params->print_interval == 0) {
      print_status(ctx, iter, objchg);
     
      /*
      printf("obj %d %lg %lg\n", iter,
             prob->fullobj(prob->user_data, n, p, V),
             genorth_infeas(prob->B, p, V));
      */
    }

    rowset_destroy(Vk);

    //break;
  }
  print_status(ctx, iter, objchg);
  printf("obj %d %lg %lg\n", iter,
         prob->fullobj(prob->user_data, n, p, V),
         genorth_infeas(prob->B, p, V));

  // Cleanup
  genorth_ctx_deinit(ctx);
  free(K);
  free(Gk);
  free(Vout);
}
