#include <assert.h>
#include <stdint.h>
#include <memory.h>

#include "serialize.h"
#include "msg.h"
#include "util/int.h"

/////////////////////////// Basic Serialization /////////////

static void serialize_doubles(intptr_t p, int n, const double *x)
{
  memcpy((void *)p, x, sizeof(double) * n);
}

static void deserialize_doubles(intptr_t p, int n, double *x)
{
  memcpy(x, (const void *)p, sizeof(double) * n);
}

static void serialize_ints(intptr_t p, int n, const int *src)
{
  int32_t *dst;
  int i;

  dst = (int32_t *)p;
  for(i = 0; i < n; i++) {
    dst[i] = (int32_t)src[i];
  }
}

static void deserialize_ints(intptr_t p, int n, int *dst)
{
  const int32_t *src;
  int i;

  src = (const int32_t *)p;
  for(i = 0; i < n; i++) {
    dst[i] = (int)src[i];
  }
}

/////////////////////////////// Setupmsg ///////////////////////////

size_t setupmsg_buflen(const struct setupmsg *msg)
{
  return serial_length(SERIALIZE_INT,
                       SERIALIZE_INT,
                       SERIALIZE_INT,
                       SERIALIZE_INT,
                       SERIALIZE_INT_VECTOR, msg->nterms,
                       SERIALIZE_INT,
                       SERIALIZE_OTHER, sizeof(config_params),
                       SERIALIZE_END);
}

int setupmsg_serialize(const struct setupmsg *msg, size_t buflen, void *buf)
{
  return serialize(buflen, buf,
                   SERIALIZE_INT, msg->n,
                   SERIALIZE_INT, msg->k,
                   SERIALIZE_INT, msg->p,
                   SERIALIZE_INT, msg->nterms,
                   SERIALIZE_INT_VECTOR, msg->nterms, msg->objtypes,
                   SERIALIZE_INT, (int)sizeof(config_params),
                   SERIALIZE_OTHER, sizeof(config_params), &msg->params,
                   SERIALIZE_END);
}

struct setupmsg *setupmsg_deserialize(size_t buflen, const void *buf)
{
  uintptr_t p, pend;
  struct setupmsg *msg;
  int32_t param_size;

  msg = (struct setupmsg *)malloc(sizeof(struct setupmsg));
  if(msg == NULL) {
    return NULL;
  }
  msg->objtypes = NULL;

  p = (intptr_t)buf;
  pend = p + buflen;

  // Get sizes
  if(buflen < 4*4) {
    free(msg);
    return NULL;
  }
  msg->n = ((int32_t *)p)[0];
  msg->k = ((int32_t *)p)[1];
  msg->p = ((int32_t *)p)[2];
  msg->nterms = ((int32_t *)p)[3];
  p += sizeof(int32_t) * 4;

  // Read obj types
  if(p + 4 * msg->nterms > pend) {
    free(msg);
    return NULL;
  }
  msg->objtypes = (int *)malloc(sizeof(int) * msg->nterms);
  if(msg->objtypes == NULL) {
    free(msg);
    return NULL;
  }
  deserialize_ints(p, msg->nterms, msg->objtypes);
  p += 4 * msg->nterms;

  // Read params
  if(p + 4 > pend) {
    setupmsg_dealloc(msg);
    return NULL;
  }
  param_size = *(int32_t *)p;
  p += sizeof(int32_t);

  if(p + param_size > pend) {
    setupmsg_dealloc(msg);
    return NULL;
  }
  msg->params = config_default_params;
  memcpy(&msg->params, (const void *)p, param_size);

  return msg;
}

void setupmsg_dealloc(struct setupmsg *msg)
{
  int i;

  if(msg == NULL) {
    return;
  }

  free(msg->objtypes);
  free(msg);
}

////////////////////////////// Rowsets //////////////////////////////

static size_t rowset_buflen(const rowset *r)
{
  return serial_length(SERIALIZE_INT,
                       SERIALIZE_INT,
                       SERIALIZE_INT,
                       SERIALIZE_INT_VECTOR, rowset_nrows(r),
                       SERIALIZE_DOUBLE_VECTOR, rowset_nelts(r),
                       SERIALIZE_END);
}

int rowset_serialize(const rowset *r, size_t buflen, void *buf)
{
  int m, n, k;

  m = rowset_nrows_all(r);
  n = rowset_ncols(r);
  k = rowset_nrows(r);

  return serialize(buflen, buf,
                   SERIALIZE_INT, m,
                   SERIALIZE_INT, n,
                   SERIALIZE_INT, k,
                   SERIALIZE_INT_VECTOR, k, rowset_inds(r),
                   SERIALIZE_DOUBLE_VECTOR, k*n, rowset_data(r),
                   SERIALIZE_END);
}

rowset *rowset_deserialize(size_t buflen, const void *buf)
{
  int m, n, k, dims[3], *inds;
  intptr_t ptr, pend;
  double *X;
  rowset *r;
  
  ptr = (intptr_t)buf;
  pend = ptr + buflen;

  // Deserialize dimensions
  if(ptr + 4*3 > pend) {
    return NULL;
  }
  deserialize_ints(ptr, 3, dims);
  ptr += 4*3;

  m = dims[0];
  n = dims[1];
  k = dims[2];

  // Allocate indices and elements
  if(ptr + 4*k + 8*k*n > pend) {
    return NULL;
  }

  inds = (int *)malloc(sizeof(int) * k);
  X = (double *)malloc(sizeof(double) * k * n);

  if(inds == NULL || X == NULL) {
    free(inds);
    free(X);
    return NULL;
  }

  // Deserialize indices
  deserialize_ints(ptr, k, inds);
  ptr += 4*k;

  // Deserialize elements
  deserialize_doubles(ptr, k*n, X);
  
  // Allocate actual rowset
  r = rowset_create(m, n, k, inds, X);

  // Cleanup
  free(inds);
  free(X);

  return r;
}

//////////////////////////////// itermsg ///////////////////////////


size_t itermsg_buflen(const struct itermsg *msg)
{
  return rowset_buflen(msg->Vk) + rowset_buflen(msg->VnbrA) + rowset_buflen(msg->VnbrB);
}

int itermsg_serialize(const struct itermsg *msg, size_t buflen, void *buf)
{
  rowset *rowset_vec[3];
  int i, r;
  uintptr_t ptr, pend;
  size_t sz;

  ptr = (uintptr_t)buf;
  pend = ptr + buflen;

  rowset_vec[0] = msg->Vk;
  rowset_vec[1] = msg->VnbrA;
  rowset_vec[2] = msg->VnbrB;

  for(i = 0; i < 3; i++) {
    sz = rowset_buflen(rowset_vec[i]);
    if(ptr + sz > pend) {
      return -1;
    }
    r = rowset_serialize(rowset_vec[i], sz, (void *)ptr);
    if(r < 0) {
      return r;
    }
    ptr += sz;
  }

  return 0;
}

struct itermsg *itermsg_deserialize(size_t buflen, const void *buf)
{
  struct itermsg *msg;
  uintptr_t ptr, pend;
  size_t sz;
  int i, j;
  rowset *rowset_vec[3];

  ptr = (uintptr_t)buf;
  pend = ptr + buflen;

  // Create structure
  msg = (struct itermsg *)malloc(sizeof(struct itermsg));
  if(msg == NULL) {
    return NULL;
  }

  // Read in rowsets
  for(i = 0; i < 3; i++) {
    rowset_vec[i] = rowset_deserialize(pend - ptr, (const void *)ptr);
    if(rowset_vec[i] == NULL) {
      for(j = 0; j < i; j++) {
        rowset_destroy(rowset_vec[j]);
      }
      free(msg);
      return NULL;
    }

    sz = rowset_buflen(rowset_vec[i]);
    ptr += sz;
  }

  msg->Vk = rowset_vec[0];
  msg->VnbrA = rowset_vec[1];
  msg->VnbrB = rowset_vec[2];

  return msg;
}

void itermsg_dealloc(struct itermsg *msg)
{
  if(msg == NULL) {
    return;
  }

  rowset_destroy(msg->Vk);
  rowset_destroy(msg->VnbrA);
  rowset_destroy(msg->VnbrB);

  free(msg);
}
