#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/socket.h>
#include <arpa/inet.h>

#include "msgbuf.h"

#ifndef MIN
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#endif

#define MAXLEN (MIN(UINT32_MAX, SIZE_MAX/2))
#define MAX_RETRIES (128)

struct msgbuf
{
  size_t len, alloc;
  void *data;
};

msgbuf *msgbuf_create()
{
  msgbuf *buf;

  buf = (msgbuf *)malloc(sizeof(msgbuf));
  if(buf == NULL) {
    return NULL;
  }

  buf->len = buf->alloc = 0;
  buf->data = NULL;

  return buf;
}


void msgbuf_destroy(msgbuf *buf)
{
  if(buf == NULL) {
    return;
  }

  free(buf->data);
  free(buf);
}


int msgbuf_setlen(msgbuf *buf, size_t len)
{
  if(len > MAXLEN) {
    return -1;
  }

  buf->len = len;
  if(len > buf->alloc) {
    buf->alloc = len;
    buf->data = realloc(buf->data, len);
  }

  if(buf->data == NULL && len != 0) {
    return -1;
  }

  return 0;
}

size_t msgbuf_getlen(msgbuf *buf)
{
  return buf->len;
}

size_t msgbuf_getalloc(msgbuf *buf)
{
  return buf->alloc;
}

void *msgbuf_data(msgbuf *buf)
{
  return buf->data;
}

int msgbuf_senddata(int sock, size_t len, const void *data)
{
  uint32_t sendlen;
  ssize_t sent, more_sent;
  char errbuf[256];
  int retries;

  if(len > MAXLEN) {
    return -1;
  }

  //printf("Sending length %lu\n", len);

  // Send length
  sendlen = htonl((uint32_t)len);
  sent = send(sock, &sendlen, 4, 0);
  if(sent != 4) {
    strerror_r(errno, errbuf, 256);
    errbuf[255] = '\0';
    fprintf(stderr, "Could not send size: %s\n", errbuf);
    return -1;
  }

  // Send the message buffer
  retries = 0;
  sent = 0;
  while(sent < (ssize_t)len && retries < MAX_RETRIES) {
    more_sent = send(sock, data + sent, len - sent, 0);
    if(more_sent < 0) {
      strerror_r(errno, errbuf, 256);
      errbuf[255] = '\0';
      fprintf(stderr, "send error: %s\n", errbuf);
      return -1;
    }

    sent += more_sent;
    retries++;
  }

  if(sent != (ssize_t)len) {
    strerror_r(errno, errbuf, 256);
    errbuf[255] = '\0';
    fprintf(stderr, "Could not send buffer: %s\n", errbuf);
    return -1;
  }

  //printf("Sent %lu: 0x%lx\n", len, *(unsigned long *)data);

  return 0;
}

int msgbuf_send(msgbuf *buf, int sock)
{
  return msgbuf_senddata(sock, buf->len, buf->data);
}

int msgbuf_recv(msgbuf *buf, int sock)
{
  uint32_t recvlen;
  ssize_t read, more_read;
  int status, retries;

  char errbuf[256];

  // Get length
  read = recv(sock, &recvlen, 4, MSG_WAITALL);
  if(read != 4) {
    strerror_r(errno, errbuf, 256);
    errbuf[255] = '\0';
    fprintf(stderr, "Failed to read length: %s\n", errbuf);
    return -1;
  }
  recvlen = ntohl(recvlen);

  //printf("Receiving length %d\n", (int)recvlen);

  // Make room for message
  status = msgbuf_setlen(buf, (size_t)recvlen);
  if(status != 0) {
    strerror_r(errno, errbuf, 256);
    errbuf[255] = '\0';
    fprintf(stderr, "Failed to resize buffer: %s\n", errbuf);
    return -1;
  }

  // Read message into buffer
  if(recvlen == 0) {
    return 0;
  }

  // Retry as necessary until buffer is filled
  retries = 0;
  read = 0;
  while(read < recvlen && retries < MAX_RETRIES) {
    more_read = recv(sock, buf->data + read, recvlen - read, MSG_WAITALL);
    if(more_read < 0) {
      strerror_r(errno, errbuf, 256);
      errbuf[255] = '\0';
      fprintf(stderr, "Socket read failed: %s\n", errbuf);
      return -1;
    }
    else if(more_read == 0) {
      retries++;
      usleep(100000);

      if(retries % 64 == 0) {
        fprintf(stderr, "%d recv retried %d times\n", sock, retries);
      }
      continue;
    }

    retries = 0;
    read += more_read;
  }

  if(read != recvlen) {
    strerror_r(errno, errbuf, 256);
    errbuf[255] = '\0';
    fprintf(stderr, "Did not read full message (%ld of %u): %s (0x%lx)\n", read, recvlen, errbuf, *(unsigned long *)buf->data);
    return -1;
  }

  return 0;
}
