#include <assert.h>
#include <memory.h>
#include <stdint.h>

#include "mustlink.h"

#include "util/sort.h"
#include "serialize.h"

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

struct mustlink
{
  int ngroups, nmembers;

  int **groups;

  int *ningroups;
  int **membership;
};

mustlink *mustlink_create(int nmembers)
{
  mustlink *ml;
  int i;

  ml = (mustlink *)malloc(sizeof(mustlink));
  if(ml == NULL) {
    return NULL;
  }

  ml->ngroups = 0;
  ml->nmembers = nmembers;

  ml->groups = NULL;

  ml->ningroups = (int *)malloc(sizeof(int) * nmembers);
  if(ml->ningroups == NULL) {
    free(ml);
    return NULL;
  }
  ml->membership = (int **)malloc(sizeof(int *) * nmembers);
  if(ml->membership == NULL) {
    free(ml->ningroups);
    free(ml);
    return NULL;
  }

  for(i = 0; i < nmembers; i++) {
    ml->ningroups[i] = 0;
    ml->membership[i] = NULL;
  }

  return ml;
}

void mustlink_destroy(mustlink *ml)
{
  if(ml == NULL) {
    return;
  }
}

int mustlink_nmembers(const mustlink *ml)
{
  return ml->nmembers;
}

int mustlink_ngroups(const mustlink *ml)
{
  return ml->ngroups;
}

int mustlink_ningroups(const mustlink *ml, int i)
{
  return ml->ningroups[i];
}

const int *mustlink_membership(const mustlink *ml, int i)
{
  return ml->membership[i];
}

int mustlink_group_nmembers(const mustlink *ml, int g)
{
  assert(g >= 0 && g < ml->ngroups);
  return ml->groups[g][0];
}

const int *mustlink_group_members(const mustlink *ml, int g)
{
  assert(g >= 0 && g < ml->ngroups);
  return ml->groups[g] + 1;
}

int mustlink_in_group(const mustlink *ml, int i, int g)
{
  int n, *q;
  const int *members;

  n = mustlink_group_nmembers(ml, g);
  members = mustlink_group_members(ml, g);

  q = bsearch_ints((size_t)n, i, (int *)members);

  return q != NULL;
}

int mustlink_add_group(mustlink *ml, int nmembers, const int *members)
{
  int i, j, n;

  n = ml->ngroups;

  ml->groups = (int **)realloc(ml->groups, sizeof(int *) * (n + 1));
  if(ml->groups == NULL) {
    return -1;
  }

  ml->groups[n] = (int *)malloc(sizeof(int) * (nmembers + 1));
  if(ml->groups[n] == NULL) {
    return -1;
  }

  ml->ngroups++;
  ml->groups[n][0] = nmembers;
  memcpy(ml->groups[n] + 1, members, sizeof(int) * nmembers);

  // Add backreference in membership
  for(i = 0; i < nmembers; i++) {
    j = members[i];
    assert(j >= 0 && j < ml->nmembers);

    ml->ningroups[j]++;
    ml->membership[j] = (int *)realloc(ml->membership[j], sizeof(int) * ml->ningroups[j]);
    ml->membership[j][ml->ningroups[j] - 1] = n;
  }

  return n;
}

///////////////////////////////// FILE I/O ////////////////////////////////

mustlink *mustlink_read_from_file(FILE *fp)
{
  mustlink *ml;
  ssize_t r;
  size_t nline;
  char *lineptr, *lasts, *tok;
  int nmembers, *group, i, grouplen, groupalloc, g;

  lineptr = NULL;
  nline = 0;

  // Read member count
  r = getdelim(&lineptr, &nline, '\n', fp);
  if(r < 0) {
    return NULL;
  }
  nmembers = atoi(lineptr);
  //fprintf(stderr, "%s:%d \"%s\" %d\n", __FILE__, __LINE__, lineptr, nmembers);

  // Create structure
  ml = mustlink_create(nmembers);
  if(ml == NULL) {
    return NULL;
  }

  // Parse groups
  groupalloc = 0;
  group = NULL;

  r = getline(&lineptr, &nline, fp);
  while(r >= 0) {
    //fprintf(stderr, "%s:%d \"%s\"\n", __FILE__, __LINE__, lineptr);
    grouplen = 0;

    // Tokenize string by whitespace
    tok = strtok_r(lineptr, " \t", &lasts);
    while(tok != NULL) {
      i = atoi(tok);

      // Add to group
      if(grouplen >= groupalloc) {
        groupalloc = MAX(grouplen + 1, groupalloc * 2);
        group = realloc(group, sizeof(int) * groupalloc);
      }
      group[grouplen++] = i;
      //fprintf(stderr, "%d\n", i);

      tok = strtok_r(NULL, " \t", &lasts);
    }

    // Add group to contraints
    //fprintf(stderr, "adding %d %d %d\n", group[0], group[1], group[2]);
    g = mustlink_add_group(ml, grouplen, group);
    if(g < 0) {
      free(lineptr);
      mustlink_destroy(ml);
      return NULL;
    }

    r = getline(&lineptr, &nline, fp);
  }

  free(lineptr);

  return ml;
}

mustlink *mustlink_read_from_path(const char *path)
{
  FILE *fp;
  mustlink *ml;

  fp = fopen(path, "r");
  if(fp == NULL) {
    return NULL;
  }

  ml = mustlink_read_from_file(fp);

  fclose(fp);

  return ml;
}

//////////////////////////////// Serialization ////////////////////////

static int mustlink_total_entries(const mustlink *ml)
{
  int i, m, n;

  m = mustlink_ngroups(ml);
  n = 0;
  for(i = 0; i < m; i++) {
    n += mustlink_group_nmembers(ml, i);
  }

  return n;
}

size_t mustlink_serial_length(const mustlink *ml)
{
  return serial_length(SERIALIZE_INT, SERIALIZE_INT,
                       SERIALIZE_INT_VECTOR, mustlink_ngroups(ml) + mustlink_total_entries(ml),
                       SERIALIZE_END);
}

int mustlink_serialize(const mustlink *ml, size_t buflen, void *buf)
{
  int i, ngroups, s;
  size_t sz;
  uintptr_t p;

  ngroups = mustlink_ngroups(ml);

  p = (uintptr_t)buf;

  // Serialize sizes
  sz = serial_length(SERIALIZE_INT, SERIALIZE_INT, SERIALIZE_END);
  s = serialize(buflen, (void *)p,
                SERIALIZE_INT, mustlink_nmembers(ml),
                SERIALIZE_INT, ngroups,
                SERIALIZE_END);
  if(s != 0) {
    return s;
  }

  buflen -= sz;
  p += sz;

  // Serialize groups
  for(i = 0; i < ngroups; i++) {
    sz = serial_length(SERIALIZE_INT_VECTOR, ml->groups[i][0] + 1, SERIALIZE_END);
    s = serialize(buflen, (void *)p, SERIALIZE_INT_VECTOR, ml->groups[i][0] + 1, ml->groups[i], SERIALIZE_END);
    if(s != 0) {
      return s;
    }
    
    
    //fprintf(stderr, "ml->groups[%d] = %d %d %d %d\n", i, ml->groups[i][0], ml->groups[i][1], ml->groups[i][2], ml->groups[i][3]);

    buflen -= sz;
    p += sz;
  }

  return 0;
}

mustlink *mustlink_deserialize(size_t buflen, const void *buf)
{
  int i, ngroups, nmembers, n, *g, galloc, s;
  size_t sz;
  uintptr_t p;
  mustlink *ml;
  
  p = (uintptr_t)buf;
  galloc = 0;
  g = NULL;

  // Deserialize sizes
  sz = serial_length(SERIALIZE_INT, SERIALIZE_INT, SERIALIZE_END);
  s = deserialize(buflen, (const void *)p,
                  SERIALIZE_INT, &nmembers,
                  SERIALIZE_INT, &ngroups,
                  SERIALIZE_END);
  if(s != 0) {
    return NULL;
  }
  buflen -= sz;
  p += sz;

  //fprintf(stderr, "sizes = %d %d\n", nmembers, ngroups);

  // Allocate structure
  ml = mustlink_create(nmembers);
  if(ml == NULL) {
    return NULL;
  }

  // Deserialize groups
  for(i = 0; i < ngroups; i++) {
    // Get group size
    sz = serial_length(SERIALIZE_INT, SERIALIZE_END);
    s = deserialize(buflen, (const void *)p, SERIALIZE_INT, &n, SERIALIZE_END);
    if(s != 0) {
      mustlink_destroy(ml);
      return NULL;
    }
    buflen -= sz;
    p += sz;

    //fprintf(stderr, "Read group of size %d, p = %lu, len = %lu\n", n, p - (uintptr_t)buf, buflen);
    
    // Make space in group buffer
    if(n > galloc) {
      galloc = MAX(n, galloc * 2);
      g = realloc(g, sizeof(int) * galloc);
      if(g == NULL) {
        mustlink_destroy(ml);
        return NULL;
      }
    }

    // Read in group
    sz = serial_length(SERIALIZE_INT_VECTOR, n, SERIALIZE_END);
    s = deserialize(buflen, (const void *)p, SERIALIZE_INT_VECTOR, n, g, SERIALIZE_END);
    if(s != 0) {
      mustlink_destroy(ml);
      return NULL;
    }
    buflen -= sz;
    p += sz;

    //fprintf(stderr, "g = %d %d %d\n", g[0], g[1], g[2]);

    // Add the ml
    s = mustlink_add_group(ml, n, g);
    if(s < 0) {
      mustlink_destroy(ml);
      return NULL;
    }
  }

  return ml;
}
