#include <stdlib.h>
#include <assert.h>
#include <stdint.h>

#include "objterm.h"
#include "util/sort.h"
#include "util/setdiff.h"
#include "util/debug.h"
#include "serialize.h"


double term_obj(const termdef *def, const rowset *V, const rowset *Vnbr)
{
  assert(def->obj != NULL);
  return def->mult * def->obj(def->user_data, V, Vnbr);
}

void term_grad(const termdef *def, const rowset *V, const rowset *Vnbr, double *Gk)
{
  int i, nelts;

  assert(def->grad != NULL);
  def->grad(def->user_data, V, Vnbr, Gk);

  nelts = rowset_nelts(V);
  //fprintf(stderr, "def->mult = %g, nelts = %d\n", def->mult, nelts);
  for(i = 0; i < nelts; i++) {
    Gk[i] *= def->mult;
  }
}

intarray *term_dependent(const termdef *def, int n, const int *inds)
{
  if(def->dependent == NULL) {
    return NULL;
  }
  return def->dependent(def->user_data, n, inds);
}

size_t term_setup_buflen(const termdef *def)
{
  if(def->setup_buflen == NULL) {
    return 0;
  }
  return def->setup_buflen(def->user_data) + sizeof(double);
}

int term_setup_serialize(const termdef *def, size_t buflen, void *buf)
{
  uintptr_t p;

  if(def->setup_serialize == NULL) {
    if(buflen == 0) {
      return 0;
    }
    else {
      return -1;
    }
  }
 
  p = (uintptr_t)buf;

  // Serialize multiplier
  if(buflen < sizeof(double)) {
    return -1;
  }
  *(double *)p = def->mult;
  p += sizeof(double);
  buflen -= sizeof(double);

  // Serialize child data
  return def->setup_serialize(def->user_data, buflen, (void *)p);
}

int term_setup_deserialize(termdef *def, size_t buflen, const void *buf)
{
  uintptr_t p;

  if(def->setup_deserialize == NULL) {
    if(buflen == 0) {
      return 0;
    }
    else {
      return -1;
    }
  }

  p = (uintptr_t)buf;
  
  // Deserialize multiplier
  if(buflen < sizeof(double)) {
    return -1;
  }
  def->mult = *(const double *)p;
  p += sizeof(double);
  buflen -= sizeof(double);

  return def->setup_deserialize(def->user_data, buflen, (const void *)p);
}

intarray *term_indupdate_request(const termdef *def, int n, const int *inds)
{
  if(def->indupdate_request == NULL) {
    return NULL;
  }
  return def->indupdate_request(def->user_data, n, inds);
}

size_t term_indupdate_buflen(const termdef *def, int n, const int *inds)
{
  assert(def->indupdate_buflen != NULL);
  return def->indupdate_buflen(def->user_data, n, inds);
}

int term_indupdate_serialize(const termdef *def, size_t buflen, void *buf, int n, const int *inds)
{
  if(def->indupdate_serialize == NULL) {
    return -1;
  }
  return def->indupdate_serialize(def->user_data, buflen, buf, n, inds);
}

int term_indupdate_deserialize(termdef *def, size_t buflen, const void *buf, int n, const int *inds)
{
  if(def->indupdate_deserialize == NULL) {
    return -1;
  }
  return def->indupdate_deserialize(def->user_data, buflen, buf, n, inds);
}

void term_destroy(termdef *def)
{
  if(def->destroy == NULL) {
    return;
  }
  return def->destroy(def->user_data);
}


double sumterm_obj(int nterms, const termdef * const *defs, const rowset *V, const rowset *Vnbr)
{
  double obj;
  int i;

  obj = 0;
  for(i = 0; i < nterms; i++) {
    obj += term_obj(defs[i], V, Vnbr);
  }

  return obj;
}

void sumterm_grad(int nterms, const termdef * const *defs, const rowset *V, const rowset *Vnbr, double *Gk)
{
  double *Gki;
  int i, j, k, p;
  
  k = rowset_nrows(V);
  p = rowset_ncols(V);
  assert(rowset_ncols(Vnbr) == p);

  Gki = (double *)malloc(sizeof(double) * k * p);

  for(j = 0; j < k*p; j++) {
    Gk[j] = 0;
  }

  for(i = 0; i < nterms; i++) {
    term_grad(defs[i], V, Vnbr, Gki);

    for(j = 0; j < k*p; j++) {
      Gk[j] += Gki[j];
    }
  }

  free(Gki);
}

intarray *sumterm_dependent(int nterms, const termdef * const *defs, int n, const int *inds)
{
  intarray *accum, *termdeps, *unique;
  int t, outndeps, *onlydeps;

  // Get all dependent indices
  accum = intarray_create(0);
  for(t = 0; t < nterms; t++) {
    termdeps = term_dependent(defs[t], n, inds);
    intarray_append(accum, intarray_length(termdeps), intarray_contents_const(termdeps));
    intarray_destroy(termdeps);
  }
  
  // Reduce to unique entries
  unique = intarray_unique(accum);

  intarray_destroy(accum);

  // Filter out original indices
  onlydeps = (int *)malloc(sizeof(int) * intarray_length(unique));
  intsetdiff(intarray_length(unique), intarray_contents_const(unique), n, inds, &outndeps, onlydeps);

  intarray_destroy(unique);

  return intarray_create_from(outndeps, onlydeps);
}

intarray *sumterm_indupdate_request(int nterms, const termdef * const *defs, int n, const int *inds)
{
  intarray *accum, *termreq, *unique;
  int t;

  // Get all dependent indices
  accum = intarray_create(0);
  for(t = 0; t < nterms; t++) {
    termreq = term_indupdate_request(defs[t], n, inds);
    intarray_append(accum, intarray_length(termreq), intarray_contents_const(termreq));
    intarray_destroy(termreq);
  }

  // Reduce to unique entries
  unique = intarray_unique(accum);
  intarray_destroy(accum);

  return unique;
}

size_t sumterm_indupdate_buflen(int nterms, const termdef * const *def, int n, const int *inds)
{
  size_t sz;
  int i;

  sz = serial_length(SERIALIZE_INT_VECTOR, nterms, SERIALIZE_END);
  for(i = 0; i < nterms; i++) {
    //fprintf(stderr, "i = %d\n", i);
    sz += term_indupdate_buflen(def[i], n, inds);
  }

  return sz;
}

int sumterm_indupdate_serialize(int nterms, const termdef * const *def, size_t buflen, void *buf, int n, const int *inds)
{
  intptr_t p, end;
  size_t sz, szints, sztotal;
  int i, *sizes, s;

  p = (intptr_t)buf;
  end = p + buflen;

  // Get sizes for individual terms
  sizes = (int *)malloc(sizeof(int) * nterms);
  if(sizes == NULL) {
    return -1;
  }

  sztotal = szints = serial_length(SERIALIZE_INT_VECTOR, nterms, SERIALIZE_END);
  for(i = 0; i < nterms; i++) {
    sz = term_indupdate_buflen(def[i], n, inds);
    sizes[i] = (int)sz;
    sztotal += sz;
  }

  if(buflen < sztotal) {
    free(sizes);
    return -1;
  }
  
  // Serialize sizes
  s = serialize(end - p, (void *)p, SERIALIZE_INT_VECTOR, nterms, sizes, SERIALIZE_END);
  p += szints;

  if(s != 0) {
    free(sizes);
    return s;
  }

  // Serialize each term in turn
  for(i = 0; i < nterms; i++) {
    assert(p + (intptr_t)sizes[i] <= end);
    s = term_indupdate_serialize(def[i], (size_t)sizes[i], (void *)p, n, inds);
    if(s != 0) {
      free(sizes);
      return s;
    }

    p += (intptr_t)sizes[i];
  }
  

  // Cleanup
  free(sizes);

  return 0;
}

int sumterm_indupdate_deserialize(int nterms, termdef * const *def, size_t buflen, const void *buf, int n, const int *inds)
{
  intptr_t p, end;
  size_t szints, sztotal;
  int i, *sizes, s;

  p = (intptr_t)buf;
  end = p + buflen;

  // Deserialize sizes for individual terms
  sizes = (int *)malloc(sizeof(int) * nterms);
  if(sizes == NULL) {
    return -1;
  }

  s = deserialize(buflen, buf, SERIALIZE_INT_VECTOR, nterms, sizes, SERIALIZE_END);
  if(s != 0) {
    free(sizes);
    return 0;
  }

  // Get total size
  sztotal = szints = serial_length(SERIALIZE_INT_VECTOR, nterms, SERIALIZE_END);
  for(i = 0; i < nterms; i++) {
    sztotal += (size_t)sizes[i];
  }

  if(buflen < sztotal) {
    free(sizes);
    return -1;
  }

  // Deserialize each term in turn
  p += szints;

  for(i = 0; i < nterms; i++) {
    assert(p + (intptr_t)sizes[i] <= end);
    s = term_indupdate_deserialize(def[i], (size_t)sizes[i], (const void *)p, n, inds);
    if(s != 0) {
      free(sizes);
      return s;
    }

    p += (intptr_t)sizes[i];
  }

  // Cleanup
  free(sizes);

  return 0;
}
