#include <assert.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <signal.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "msg.h"
#include "stiefelprob.h"

#include "server.h"
#include "util/randorth.h"
#include "util/debug.h"
#include "util/matio.h"
#include "util/denseio.h"
#include "util/cachemat.h"
#include "optim_config.h"

#include "util/timing.h"

#define PORTNUM (5428)

const char * const USAGE = "parallel-multiview-server <config> <problem> <Vout>";

static const char *signal_name(int signum);
static void signal_handler(int signum);

int main(int argc, char **argv)
{
  const char *config_path, *problem_path, *outpath;
  
  int i, j, n, p, dims[2];
  struct sockaddr_in serv;
  int sock, consock;

  struct sigaction new_action;

  config_params config;
  stiefel_prob *prob;

  srand(0);

  // We have a better way to handle SIGPIPE than to simply kill the whole deal
  sigemptyset(&new_action.sa_mask);
  new_action.sa_handler = signal_handler;
  new_action.sa_flags = 0;
  sigaction(SIGPIPE, &new_action, NULL);

  // Interpret arguments
  if(argc < 4) {
    fprintf(stderr, "Usage: %s\n", USAGE);
    return EX_USAGE;
  }

  config_path = argv[1];
  problem_path = argv[2];
  outpath = argv[3];

  // Read configuration
  optim_config_read_path(&config, config_path);
  
  // Read in the problem
  fprintf(stderr, "Reading problem data...\n"); fflush(stderr);
  prob = stiefel_prob_from_path(problem_path);
  if(prob == NULL) {
    fprintf(stderr, "Error reading problem data from \"%s\"\n", problem_path);
    return 1;
  }

  // Start up server
  sock = socket(AF_INET, SOCK_STREAM, 0);
  assert(sock != -1);

  serv.sin_family = AF_INET;
  serv.sin_addr.s_addr = htonl(INADDR_ANY);
  serv.sin_port = htons(PORTNUM);

  server_run_sockets(&config, prob, sock, (struct sockaddr *)&serv);

  // Output V
  n = stiefel_prob_n(prob);
  p = stiefel_prob_p(prob);
  dense_file_write_rowmajor(outpath, 1, n, p, prob->V0);

  // Cleanup
  close(sock);
  stiefel_prob_destroy(prob);

  return 0;
}

static const char *signal_name(int signum)
{
  switch(signum) {
  case SIGPIPE:
    return "SIGPIPE";
  default:
    return "UNKNOWN";
  }
}

static void signal_handler(int signum)
{
  fprintf(stderr, "recieved signal %s (%d)\n", signal_name(signum), signum);
}
