#include <assert.h>

#include "sample_independent.h"

static void print_ints(int k, const int *p);

struct independent_sampler
{
  sparsemat *adj;

  int nfree;

  /**
   * A flat list of indices which may be sampled.
   */
  int *freelist;

  /**
   * Reverse mapping of @c freelist.
   *
   * freerev[i] is the location of i in the freelist.
   */
  int *freerev;

  /**
   * The number of neighbors of each node  which are currently "sampled"
   */
  int *sampled_neighbors;

  /**
   * Workspace for computing total number of neighbors of a vertex set.
   */
  int *nbrcnt;
};

independent_sampler *independent_sampler_create(sparsemat *adj)
{
  int i, n;
  independent_sampler *is;

  n = sparsemat_nrows(adj);
  if(n != sparsemat_ncols(adj)) { // matrix must be square
    return NULL;
  }

  // Allocate
  is = (independent_sampler *)malloc(sizeof(independent_sampler));
  if(is == NULL) {
    return NULL;
  }
  is->adj = adj;

  is->nfree = n;
  is->freelist = (int *)malloc(sizeof(int) * 4 * n);
  if(is->freelist == NULL) {
    free(is);
    return NULL;
  }
  is->freerev = is->freelist + n;
  is->sampled_neighbors = is->freerev + n;
  is->nbrcnt = is->sampled_neighbors + n;

  // Initialize
  for(i = 0; i < n; i++) {
    is->freelist[i] = i;
    is->freerev[i] = i;
    is->sampled_neighbors[i] = 0;
    is->nbrcnt[i] = 0;
  }

  return is;
}

void independent_sampler_destroy(independent_sampler *is)
{
  if(is == NULL) {
    return;
  }

  free(is->freelist);
  free(is);
}

/**
 * Remove a given entry from the free list.
 *
 * @arg @c i Index into the free list.
 * @returns Index of the invalidated vertex
 */
static int invalidate(independent_sampler *is, int i)
{
  int mov, num;

  assert(i >= 0 && i < is->nfree);

  num = is->freelist[i];

  // Move the last free entry over the sampled point
  mov = is->freelist[is->nfree-1];
  is->freelist[i] = mov;
  is->nfree--;

  // Update reverse mapping
  is->freerev[mov] = i;
  is->freerev[num] = -1;

  return num;
}

int independent_sampler_sample(independent_sampler *is, int k, int *ind, int *out_nnbr, int **out_nbr)
{
  int i, j, samp, mov;
  int nnbr, nnbr2, nbrj, *nbr;
  sparsecol col;

  if(is->nfree < k) {
    k = is->nfree;
  }

  // Sample
  for(i = 0; i < k; i++) {
    samp = rand() % is->nfree;
    ind[i] = invalidate(is, samp);
  }


  // Count neighbors
  for(i = 0; i < k; i++) {
    is->nbrcnt[ind[i]] = 1;
  }

  nnbr = 0;
  for(i = 0; i < k; i++) {
    col = sparsemat_col(is->adj, ind[i]);

    for(j = 0; j < col.nnz; j++) {
      nbrj = col.i[j];

      if(!is->nbrcnt[nbrj]) {
        nnbr++;
        is->nbrcnt[nbrj] = 1;
      }

      is->sampled_neighbors[nbrj]++;
    }
  }

  for(i = 0; i < k; i++) {
    is->nbrcnt[ind[i]] = 0;
  }

  // Get list of neighbors
  nbr = (int *)malloc(sizeof(int) * nnbr);
  nnbr2 = 0;
  for(i = 0; i < k; i++) {
    col = sparsemat_col(is->adj, ind[i]);

    for(j = 0; j < col.nnz; j++) {
      nbrj = col.i[j];
      
      if(is->nbrcnt[nbrj]) {
        nbr[nnbr2++] = nbrj;
        is->nbrcnt[nbrj] = 0;
      }
    }
  }
  assert(nnbr == nnbr2);

  // Invalidate neighbors
  for(i = 0; i < nnbr; i++) {
    if(is->freerev[nbr[i]] != -1) {
      invalidate(is, is->freerev[nbr[i]]);
    }
  }

  if(out_nnbr != NULL && out_nbr != NULL) {
    *out_nnbr = nnbr;
    *out_nbr = nbr;
  }
  else {
    free(nbr);
  }
  return k;
}

static void reallow(independent_sampler *is, int i)
{
  int mov, num;

  //printf("freelist = "); print_ints(is->nfree, is->freelist);
  //printf("freerev = "); print_ints(is->adj->n, is->freerev);

  assert(is->freerev[i] == -1);
  assert(is->nfree < sparsemat_nrows(is->adj));

  // Append to the end
  is->freelist[is->nfree] = i;
  is->freerev[i] = is->nfree;
  is->nfree++;
}

void independent_sampler_return(independent_sampler *is, int k, const int *ind)
{
  int i, j, nbrj;
  sparsecol col;

  // Reallow all entries of k
  for(i = 0; i < k; i++) {
    if(is->sampled_neighbors[ind[i]] == 0) {
      // Note that the condition will only be false if there are edges
      // inside ind, so those will be reallowed in the subsequenct paragraph
      reallow(is, ind[i]);
    }
  }

  // Reallow neighbors which do not have
  for(i = 0; i < k; i++) {
    col = sparsemat_col(is->adj, ind[i]);

    for(j = 0; j < col.nnz; j++) {
      nbrj = col.i[j];
      is->sampled_neighbors[nbrj]--;

      if(is->sampled_neighbors[nbrj] == 0) {
        reallow(is, nbrj);
      }
    }
  }
}


static void print_ints(int k, const int *p)
{
  int i;

  for(i = 0; i < k; i++) {
    printf("%d ", p[i]);
  }
  printf("\n");
}
