#include <stdarg.h>
#include <stdint.h>
#include <memory.h>
#include <assert.h>
#include <stdio.h>

#include "serialize.h"

/////////////////////////// serialize ///////////////////////////

#define SERIALIZE_SCALAR_CASE(typename, type, casttype)  \
  case typename: {                                       \
    type arg;                                            \
                                                         \
    if(p + sizeof(casttype) > pend) {                    \
      return -1;                                         \
    }                                                    \
                                                         \
    arg = va_arg(ap, type);                              \
    *(casttype *)p = (casttype)arg;                      \
    p += sizeof(casttype);                               \
  } break

#define SERIALIZE_VECTOR_CASE(typename, type)            \
  case typename: {                                       \
    int len;                                             \
    const type *arg;                                     \
                                                         \
    len = va_arg(ap, int);                               \
                                                         \
    if(p + len * sizeof(type) > pend) {                  \
      return -1;                                         \
    }                                                    \
                                                         \
    arg = va_arg(ap, const type *);                      \
    memcpy((void *)p, arg, sizeof(type) * len);          \
    p += len * sizeof(type);                             \
  } break

int serialize(size_t buflen, void *buf, ...)
{
  va_list ap;
  uintptr_t p, pend;
  serial_type type;
  
  p = (uintptr_t)buf;
  pend = p + buflen;

  va_start(ap, buf);

  type = va_arg(ap, serial_type);
  while(type != SERIALIZE_END) {
    switch(type) {
      SERIALIZE_SCALAR_CASE(SERIALIZE_INT, int, int32_t);
      SERIALIZE_SCALAR_CASE(SERIALIZE_DOUBLE, double, double);

      SERIALIZE_VECTOR_CASE(SERIALIZE_INT_VECTOR, int);
      SERIALIZE_VECTOR_CASE(SERIALIZE_DOUBLE_VECTOR, double);

    case SERIALIZE_OTHER: {
      int sz;
      const void *ptr;

      sz = va_arg(ap, int);
      ptr = va_arg(ap, const void *);

      if(p + sz > pend) {
        return -1;
      }

      memcpy((void *)p, ptr, (size_t)sz);
      p += sz;
    } break;

    default:
      return -1;
    }

    type = va_arg(ap, serial_type);
  }

  va_end(ap);

  return 0;
}

///////////////////////////// deserialize //////////////////////

#define DESERIALIZE_SCALAR_CASE(typename, type, casttype)  \
  case typename: {                                       \
    type *arg;                                           \
                                                         \
    if(p + sizeof(casttype) > pend) {                    \
      return -1;                                         \
    }                                                    \
                                                         \
    arg = va_arg(ap, type *);                            \
    *arg = (type)(*(casttype *)p);                       \
    p += sizeof(casttype);                               \
  } break

#define DESERIALIZE_VECTOR_CASE(typename, type)          \
  case typename: {                                       \
    int len;                                             \
    type *arg;                                           \
                                                         \
    len = va_arg(ap, int);                               \
                                                         \
    if(p + len * sizeof(type) > pend) {                  \
      return -1;                                         \
    }                                                    \
                                                         \
    arg = va_arg(ap, type *);                            \
    memcpy(arg, (const void *)p, sizeof(type) * len);    \
    p += len * sizeof(type);                             \
  } break

int deserialize(size_t buflen, const void *buf, ...)
{
  va_list ap;
  uintptr_t p, pend;
  serial_type type;
  
  p = (uintptr_t)buf;
  pend = p + buflen;

  va_start(ap, buf);

  type = va_arg(ap, serial_type);
  while(type != SERIALIZE_END) {
    switch(type) {
      DESERIALIZE_SCALAR_CASE(SERIALIZE_INT, int, int32_t);
      DESERIALIZE_SCALAR_CASE(SERIALIZE_DOUBLE, double, double);

      DESERIALIZE_VECTOR_CASE(SERIALIZE_INT_VECTOR, int);
      DESERIALIZE_VECTOR_CASE(SERIALIZE_DOUBLE_VECTOR, double);

    case SERIALIZE_OTHER: {
      int sz;
      void *ptr;

      sz = va_arg(ap, int);
      ptr = va_arg(ap, void *);

      if(p + sz > pend) {
        return -1;
      }

      memcpy(ptr, (const void *)p, (size_t)sz);
      p += sz;
    } break;

    default:
      return -1;
    }

    type = va_arg(ap, serial_type);
  }

  va_end(ap);

  return 0;
}

///////////////////////////// serialize_length /////////////////////

#define SERIALLEN_SCALAR_CASE(typename, casttype) \
  case typename: {                                \
    sz += sizeof(casttype);                       \
  } break

#define SERIALLEN_VECTOR_CASE(typename, type)  \
  case typename: {                             \
    int len;                                   \
                                               \
    len = va_arg(ap, int);                     \
    sz += (size_t)len * sizeof(type);          \
  } break

size_t serial_length(serial_type type, ...)
{
  va_list ap;
  size_t sz;

  sz = 0;
  va_start(ap, type);

  while(type != SERIALIZE_END) {
    switch(type) {
      SERIALLEN_SCALAR_CASE(SERIALIZE_INT, int32_t);
      SERIALLEN_SCALAR_CASE(SERIALIZE_DOUBLE, double);

      SERIALLEN_VECTOR_CASE(SERIALIZE_INT_VECTOR, int);
      SERIALLEN_VECTOR_CASE(SERIALIZE_DOUBLE_VECTOR, double);

    case SERIALIZE_OTHER: 
      sz += va_arg(ap, int);
      break;

    default:
      return 0;
    }

    type = va_arg(ap, serial_type);
  }

  va_end(ap);

  return sz;
}
