#include <stdint.h>
#include <assert.h>
#include <memory.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "server.h"

#include "sample_independent.h"
#include "msg.h"
#include "colmsg.h"
#include "msgbuf.h"
#include "util/randsample.h"
#include "util/sort.h"
#include "util/debug.h"


///////////// Server state /////////////
struct client_handler;

struct server_state
{
  int n, p, k;
  int sock;

  int done;

  // Problem data, packed in setup message sent to start clients
  struct setupmsg setup;
  size_t setuplen;
  void *setupbuf;

  int nterms;
  const termdef **objdef;
  sparsemat *B;
  double *V;
  double obj;

  // Structure for sampling rows to give to clients
  independent_sampler *samp;

  // Linked list of idle clients
  int nidle;
  struct client_handler *idle;
  struct client_handler **idle_tail;

  // Condition variable for existence of idle clients
  pthread_cond_t idlecond;

  // All clients
  int nsocks;

  // Head of a doubly-linked list of clients
  struct client_handler *clients;

  // Performance information
  struct timeval optim_start;

  // Lock for all common data accessed by server and client handler threads
  pthread_mutex_t lock;
};

static void server_state_init(struct server_state *st, int sock, const config_params *params, const stiefel_prob *prob);
static void server_state_cleanup(struct server_state *st);

static void *server_listen(void *arg);
static int server_count_idle(struct server_state *st);

/////////////////////// Client handlers /////////////////////

struct client_handler
{
  struct server_state *st;

  pthread_t thread;
  int sock;

  // Rows currently out to this client
  struct itermsg iter;

  // For linked list of idle clients
  struct client_handler *nextidle;

  // Doubly-linked list of all clients
  struct client_handler *prev, *next;

  // Signaling from main thread
  int started, zombie;
  pthread_cond_t cond;
  pthread_mutex_t condlock;
};

static struct client_handler *client_handler_start(struct server_state *st, int sock);
static void client_handler_shutdown(struct client_handler *hnd);
static void client_handler_cleanup(struct client_handler *hnd);

static void *client_routine(void *arg);

static int handle_colrequest(struct server_state *st, int sock);

///////////// Utility Functions /////////////
static rowset *getneighbors(int nmat, sparsemat **A, int n, int p, int k, const int *K, const double *V);
static void builditer(struct server_state *st, struct itermsg *iter, const double *V);
static struct colmsg *buildcolmsg(struct server_state *st, int ncols, const int *cols);

void server_run_sockets(const config_params *params, const stiefel_prob *prob, int sock, struct sockaddr *serv)
{
  struct server_state st;
  struct client_handler *hnd;
  pthread_t listen_thread;

  int i, iteri, n, k, dims[2];

  double timediff;
  double *V;

  sparsemat_dims(prob->B, dims);
  n = dims[0];
  k = params->optim.rowsel;
  V = prob->V0;

  // Initialize server
  server_state_init(&st, sock, params, prob);
  //fprintf(stderr, "obj start = %lg\n", multitrace_full_obj(nmat, A, p, V));
  fprintf(stderr, "Initialized\n");

  // Start listening to socket
  int reuseaddr = 1;
  setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(int));
  assert(0 == bind(sock, serv, sizeof(struct sockaddr)));
  assert(0 == pthread_create(&listen_thread, NULL, server_listen, &st));

  // Main loop
  for(iteri = 0; iteri < params->optim.max_iter; iteri++) {
    // Get first idle thread
    pthread_mutex_lock(&st.lock);
    if(st.idle == NULL) {
      pthread_cond_wait(&st.idlecond, &st.lock);
    }
    assert(st.idle != NULL);
    hnd = st.idle;

    // Construct iteration
    builditer(&st, &hnd->iter, V);

    // Pop from idle queue
    st.idle = hnd->nextidle;
    if(st.idle == NULL) {
      st.idle_tail = &st.idle;
    }
    st.nidle--;

    if(iteri % params->optim.print_interval == 0) {
      //fprintf(stderr, "obj %d = %lg, %d\n", iteri, multitrace_full_obj(st.setup.nmat, st.A, st.setup.p, st.setup.V0), st.nsocks);
      fprintf(stderr, "%5d %le %2d\n", iteri, st.obj, st.nsocks);
    }

    pthread_mutex_unlock(&st.lock);

    // Signal to send to client
    pthread_mutex_lock(&hnd->condlock);
    pthread_cond_signal(&hnd->cond);
    pthread_mutex_unlock(&hnd->condlock);
  }

  fprintf(stderr, "server loop complete\n");
  // Print performace information
  fprintf(stderr, "Optimization time: %lgs\n", timeelapsed(&st.optim_start));

  // Shut down and clean up clients
  i = 0;
  while(st.nidle < st.nsocks && i < 10) {
    fprintf(stderr, "st.nidle = %d, st.nsocks = %d\n", st.nidle, st.nsocks);
    fprintf(stderr, "manual nidle = %d\n", server_count_idle(&st));
    sleep(1); // TODO: replace with waiting on idle signal
    i++;
  }
  for(hnd = st.clients; hnd != NULL; hnd = hnd->next) {
    client_handler_shutdown(hnd);
  }
  while(st.clients != NULL) {
    sleep(1);
  }


  server_state_cleanup(&st);
}


//////////////////////////////// Utility Functions ////////////////////////////

static rowset *getneighbors(int nmat, sparsemat **A, int n, int p, int k, const int *K, const double *V)
{
  int i, j, l, ind, mati, nnbr, alloc, *nbr;
  sparsecol col;
  size_t nunique;
  rowset *r;

  nnbr = 0;
  alloc = 4 * nmat * k;
  nbr = (int *)malloc(sizeof(int) * alloc);

  for(i = 0; i < k; i++) {
    ind = K[i];

    for(mati = 0; mati < nmat; mati++) {
      col = sparsemat_col(A[mati], ind);

      // Allocate
      if(nnbr + col.nnz >= alloc) {
        alloc += col.nnz;
        nbr = (int *)realloc(nbr, sizeof(int) * alloc);
      }

      // Copy
      for(j = 0; j < col.nnz; j++) {
        nbr[nnbr++] = col.i[j];
      }
    }
  }

  // Retrieve unique neighbors
  nunique = (size_t)nnbr;
  unique_ints(&nunique, nbr);
  nnbr = (int)nunique;

  // Filter out actual selected nodes
  assert(issorted(k, K));
  assert(issorted(nnbr, nbr));

  j = 0, l = 0;
  for(i = 0; i < nnbr; ) {
    if(j == k || nbr[i] < K[j]) {
      nbr[l] = nbr[i];
      i++; l++;
    }
    else if(nbr[i] == K[j]) {
      i++; j++;
    }
    else {
      j++;
    }
  }
  nnbr = l;

  r = rowset_extract(n, p, nnbr, nbr, V);

  // Cleanup
  free(nbr);

  return r;
}

static void builditer(struct server_state *st, struct itermsg *iter, const double *V)
{
  intarray *objdeps;
  sparsemat *B;
  int n, k, p, *ind;

  B = st->B;

  n = st->n;
  p = st->p;

  // Sample indices
  ind = (int *)malloc(sizeof(int) * st->k);
  assert(ind != NULL);
  k = independent_sampler_sample(st->samp, st->k, ind, NULL, NULL);
  sort_ints((size_t)k, ind);
  assert(k <= st->k);

  // Select rows
  iter->Vk = rowset_extract(n, p, k, ind, st->V);
  assert(iter->Vk != NULL);

  // Select neighbor rows
  objdeps = sumterm_dependent(st->nterms, st->objdef, k, ind);
  iter->VnbrA = rowset_extract(n, p, intarray_length(objdeps), intarray_contents_const(objdeps), st->V);

  iter->VnbrB = getneighbors(1, &B, n, p, k, ind, st->V); // TODO: get neighbors from independent sampler?

  // Cleanup
  free(ind);
}


/////////////////////////////// Server state /////////////////////////////////

static void server_state_init(struct server_state *st, int sock, const config_params *params, const stiefel_prob *prob)
{
  int i, j;

  // Server state setup
  st->n = stiefel_prob_n(prob);
  st->p = stiefel_prob_p(prob);
  st->k = params->optim.rowsel;
  st->B = prob->B;
  st->nterms = stiefel_prob_nterms(prob);
  st->objdef = (const termdef **)malloc(sizeof(const termdef *) * st->nterms);
  for(i = 0; i < st->nterms; i++) {
    st->objdef[i] = stiefel_prob_get_term_const(prob, i);
  }

  st->V = prob->V0;
  st->obj = 0;
  st->sock = sock;

  st->samp = independent_sampler_create(prob->B);

  // Problem setup
  st->setup.nterms = stiefel_prob_nterms(prob);
  st->setup.n = stiefel_prob_n(prob);
  st->setup.k = params->optim.rowsel;
  st->setup.p = stiefel_prob_p(prob);
  st->setup.params = *params;

  st->setup.objtypes = (int *)malloc(sizeof(int) * st->nterms);
  for(i = 0; i < st->nterms; i++) {
    st->setup.objtypes[i] = st->objdef[i]->type;
  }

  st->setuplen = setupmsg_buflen(&st->setup);
  assert(st->setuplen > 0);
  st->setupbuf = malloc(st->setuplen);
  assert(st->setupbuf != NULL);
  setupmsg_serialize(&st->setup, st->setuplen, st->setupbuf);

  fprintf(stderr, "Built setup message of length %lu (0x%lx)\n", st->setuplen, (uintptr_t)st->setupbuf);

  // Client handler setup
  st->nidle = 0;
  st->idle = NULL;
  st->idle_tail = &st->idle;
  
  st->nsocks = 0;
  st->clients = NULL;

  pthread_mutex_init(&st->lock, NULL);
  pthread_cond_init(&st->idlecond, NULL);
  //pthread_mutex_init(&st->idlecondlock, NULL);
}

static void server_state_cleanup(struct server_state *st)
{
  int i;

  free(st->objdef);
  free(st->setup.objtypes);
  free(st->setupbuf);
  independent_sampler_destroy(st->samp);
  free(st->clients);
  pthread_mutex_destroy(&st->lock);
  pthread_cond_destroy(&st->idlecond);
  //pthread_mutex_destroy(&st->idlecondlock);
}

static int server_count_idle(struct server_state *st)
{
  struct client_handler *h;
  int nidle;

  pthread_mutex_lock(&st->lock);

  nidle = 0;
  for(h = st->idle; h != NULL; h = h->nextidle) {
    printf("%d ", h->sock);
    nidle++;
  }
  printf("\n");

  pthread_mutex_unlock(&st->lock);

  return nidle;
}

static void *server_listen(void *arg)
{
  struct server_state *st;
  struct sockaddr_in client;
  socklen_t socksize;
  int sock, consock;

  st = (struct server_state *)arg;
  sock = st->sock;

  assert(0 == listen(sock, 4));

  socksize = sizeof(struct sockaddr_in);
  consock = accept(sock, (struct sockaddr *)&client, &socksize);
  assert(consock != -1);

  while(1) {
    if(consock != -1) {
      fprintf(stderr, "Incoming connection from %s\n", inet_ntoa(client.sin_addr));
      client_handler_start(st, consock);
    }

    consock = accept(sock, (struct sockaddr *)&client, &socksize);
  }

  fprintf(stderr, "listen loop halted\n");

  while(1);

  return NULL;
}

//////////////////////////////// Client handlers ////////////////////////////

static struct client_handler *client_handler_start(struct server_state *st, int sock)
{
  struct client_handler *h;
  int k, p, n;

  // Build handler structure
  h = (struct client_handler *)malloc(sizeof(struct client_handler));
  h->st = st;
  h->sock = sock;

  h->next = h->prev = h->nextidle = NULL;

  h->started = 0;
  h->zombie = 0;
  pthread_cond_init(&h->cond, NULL);
  pthread_mutex_init(&h->condlock, NULL);

  // Iteration structure
  n = st->n;
  h->iter.Vk = NULL;
  h->iter.VnbrA = NULL;
  h->iter.VnbrB = NULL;

  // Spawn handler thread
  pthread_create(&h->thread, NULL, client_routine, h);

  return h;
}

static void *client_routine(void *arg)
{
  struct client_handler *h;
  struct server_state *st;
  int s, sock;
  ssize_t ret;
  int32_t termmsg;
  msgbuf *buf;
  int good;

  int iterk;
  const int *iterinds;
  double chg;

  rowset *Vout;

  pthread_detach(pthread_self());

  h = (struct client_handler *)arg;
  st = h->st;
  sock = h->sock;

  buf = msgbuf_create();

  good = 1;

  // Send client setup message
  s = msgbuf_senddata(sock, st->setuplen, st->setupbuf);
  if(s != 0) {
    fprintf(stderr, "Client failed to initialize (%d)\n", sock);
    msgbuf_destroy(buf);
    // TODO: other cleanup?
    return NULL;
  }

  {
    int i, n;

    // Get setup requests
    s = msgbuf_recv(buf, sock);
    if(s != 0 || msgbuf_getlen(buf) < 4) {
      fprintf(stderr, "Client failed to initialize (%d)\n", sock);
      msgbuf_destroy(buf);
      // TODO: other cleanup?
      return NULL;
    }

    n = *(int *)msgbuf_data(buf);
    if(n == st->nterms) {
      for(i = 0; i < n; i++) {
        s = msgbuf_setlen(buf, term_setup_buflen(st->objdef[i]));
        s = s || term_setup_serialize(st->objdef[i], msgbuf_getlen(buf), msgbuf_data(buf));
        s = s || msgbuf_send(buf, sock);
        if(s != 0) {
          fprintf(stderr, "Client failed to initialize (%d)\n", sock);
          msgbuf_destroy(buf);
          // TODO: other cleanup?
          return NULL;
        }
      }
    }
    else if(n != 0) {
      fprintf(stderr, "Invalid obj setup request (%d)\n", sock);
      msgbuf_destroy(buf);
      return NULL;
    }
  } 

  fprintf(stderr, "Client initialized (%d)\n", sock);

  // Initial column request
  s = handle_colrequest(st, sock);
  if(s != 0) {
    // TODO: other cleanup?
    msgbuf_destroy(buf);
    return NULL;
  }

  pthread_mutex_lock(&st->lock);
  
  // Add self to list of clients
  st->nsocks++;
  assert(h->prev == NULL);
  h->next = st->clients;
  if(st->clients != NULL) {
    assert(st->clients->prev == NULL);
    st->clients->prev = h;
  }
  st->clients = h;

  /*
  {
    int found = 0;
    struct client_handler *i;
    for(i = st->clients; i != NULL; i = i->next) {
      assert(i->prev != NULL || st->clients == i);
      if(i == h) {
        found = 1;
      }
      fprintf(stderr, "%d ", i->sock);
    }
    fprintf(stderr, "\n");
    assert(found);
  }
  */

  // Add self to idle queue
  assert(h->nextidle == NULL);
  *(st->idle_tail) = h;
  st->idle_tail = &(h->nextidle);
  st->nidle++;

  // Start optimization timer if I'm the first client
  if(st->nsocks == 1) {
    gettimeofday(&st->optim_start, NULL);
    fprintf(stderr, "First client started at "); printcurrenttime(stderr); fprintf(stderr, "\n");
  }

  pthread_mutex_unlock(&st->lock);

  // Wake up server if it's waiting for an idle thread
  pthread_cond_signal(&st->idlecond);

  //printf("Starting client handler loop\n");

  pthread_mutex_lock(&h->condlock);
  h->started = 1;
  while(1) {
    // Wait for job to be available
    s = pthread_cond_wait(&h->cond, &h->condlock);
    assert(s == 0);

    //printf("Client handler woken\n");

    if(h->zombie) { // woken for shutdown
      break;
    }

    iterk = rowset_nrows(h->iter.Vk);
    iterinds = rowset_inds(h->iter.Vk);
    assert(iterk <= st->k); // TODO: allow this to be less

    // Build iteration message
    s = msgbuf_setlen(buf, itermsg_buflen(&h->iter));
    if(s == -1) {
      fprintf(stderr, "Failed to allocate iteration message\n");
      good = 0;
      break;
    }
    itermsg_serialize(&h->iter, msgbuf_getlen(buf), msgbuf_data(buf));

    // Send to client
    s = msgbuf_send(buf, sock);
    if(s != 0) {
      fprintf(stderr, "Client send failed (%d)\n", sock);
      good = 0;
      break;
    }

    s = handle_colrequest(st, sock);
    if(s != 0) {
      good = 0;
      break;
    }

    // Retrieve updates
    s = msgbuf_recv(buf, sock);
    if(s == -1) {
      fprintf(stderr, "Client recv failed (%d)\n", sock);
      good = 0;
      break;
    }

    if(msgbuf_getlen(buf) != sizeof(double) * (size_t)(iterk * st->p)) {
      fprintf(stderr, "Invalid update recieved (%d)\n", sock);
      good = 0;
      break;
    }

    Vout = rowset_create(st->n, st->p, iterk, iterinds, msgbuf_data(buf));

    s = msgbuf_recv(buf, sock);
    if(s == -1 || msgbuf_getlen(buf) != sizeof(double)) {
      fprintf(stderr, "Client recv failed (%d)\n", sock);
      good = 0;
      break;
    }
    chg = *(double *)msgbuf_data(buf);

    // Done with this step, commit it to server
    pthread_mutex_lock(&st->lock);

    assert(h->prev != NULL || st->clients == h);

    // Put result back in
    independent_sampler_return(st->samp, iterk, iterinds);
    rowset_install(st->V, Vout);
    st->obj += chg;

    // Cleanup
    rowset_destroy(Vout);
    rowset_destroy(h->iter.Vk); h->iter.Vk = NULL;
    rowset_destroy(h->iter.VnbrA); h->iter.VnbrA = NULL;
    rowset_destroy(h->iter.VnbrB); h->iter.VnbrB = NULL;

    // Add self back to idle queue
    *(st->idle_tail) = h;
    h->nextidle = NULL;
    st->idle_tail = &(h->nextidle);
    st->nidle++;

    // Wake up server if it's waiting for an idle thread
    pthread_cond_signal(&st->idlecond);

    {
      int found = 0;
      struct client_handler *i;
      for(i = st->clients; i != NULL; i = i->next) {
        assert(i->prev != NULL || st->clients == i);
        if(i == h) {
          found = 1;
        }
      }
      assert(found);
    }

    pthread_mutex_unlock(&st->lock);
  }
  pthread_mutex_unlock(&h->condlock);

  fprintf(stderr, "Client handler stopped\n");

  // Send termination message
  if(good) {
    fprintf(stderr, "Sending termination message\n");
    termmsg = 0;
    send(sock, &termmsg, 4, 0);
  }

  msgbuf_destroy(buf);

  // Remove handler structure
  client_handler_cleanup(h);

  return NULL;
}

static int handle_colrequest(struct server_state *st, int sock)
{
  msgbuf *buf;
  int s, ncols;
  int *cols;

  buf = msgbuf_create();

  // Recieve requested columns
  s = msgbuf_recv(buf, sock);
  if(s != 0) {
    msgbuf_destroy(buf);
    fprintf(stderr, "Client recv interrupted (%d)\n", sock);
    return s;
  }

  if(msgbuf_getlen(buf) % sizeof(int) != 0) {
    fprintf(stderr, "Malformed column request (%d)\n", sock);
    msgbuf_destroy(buf);
    return s;
  }
  ncols = msgbuf_getlen(buf) / sizeof(int);
  cols = (int *)malloc(sizeof(int) * ncols);
  memcpy(cols, msgbuf_data(buf), msgbuf_getlen(buf));

  // Send requested columns for objective
  s = msgbuf_setlen(buf, sumterm_indupdate_buflen(st->nterms, st->objdef, ncols, cols));
  if(s != 0) {
    fprintf(stderr, "Failed to accumulate requested columns (%d)\n", sock);
    free(cols);
    msgbuf_destroy(buf);
    return s;
  }

  s = sumterm_indupdate_serialize(st->nterms, st->objdef, msgbuf_getlen(buf), msgbuf_data(buf), ncols, cols);
  if(s != 0) {
    fprintf(stderr, "Failed to accumulate requested columns (%d)\n", sock);
    free(cols);
    msgbuf_destroy(buf);
    return s;
  }

  s = msgbuf_send(buf, sock);
  if(s != 0) {
    fprintf(stderr, "Client send failed (%d)\n", sock);
    free(cols);
    msgbuf_destroy(buf);
    return s;
  }

  // Send requested columns for constraints
  s = msgbuf_setlen(buf, colmsg_buflen(st->B, ncols, cols));
  if(s != 0) {
    fprintf(stderr, "Failed to accumulate requested columns (%d)\n", sock);
    free(cols);
    msgbuf_destroy(buf);
    return s;
  }

  colmsg_serialize(st->B, ncols, cols, msgbuf_getlen(buf), msgbuf_data(buf));

  s = msgbuf_send(buf, sock);
  if(s != 0) {
    fprintf(stderr, "Client send failed (%d)\n", sock);
    free(cols);
    msgbuf_destroy(buf);
    return s;
  }


  // Cleanup
  free(cols);
  msgbuf_destroy(buf);

  return 0;
}

static void client_handler_shutdown(struct client_handler *h)
{
  fprintf(stderr, "Sending shutdown signal\n");

  pthread_mutex_lock(&h->condlock);
  h->zombie = 1;
  pthread_cond_signal(&h->cond);
  pthread_mutex_unlock(&h->condlock);
}

static void client_handler_cleanup(struct client_handler *h)
{
  struct server_state *st;

  st = h->st;

  // Remove self from list of clients
  pthread_mutex_lock(&st->lock);

  /*
  {
    int found = 0;
    struct client_handler *i;
    for(i = st->clients; i != NULL; i = i->next) {
      assert(i->prev != NULL || st->clients == i);
      if(i == h) {
        found = 1;
      }
      fprintf(stderr, "%d ", i->sock);
    }
    fprintf(stderr, "\n");
    assert(found);
  }
  */

  assert(h->prev != NULL || st->clients == h);

  if(h->prev == NULL) {
    assert(st->clients == h);
    st->clients = h->next;
  }
  else {
    assert(h->prev->next == h);
    h->prev->next = h->next;
  }
  if(h->next != NULL) {
    assert(h->next->prev == h);
    h->next->prev = h->prev;
  }
  st->nsocks--;


  {
    struct client_handler *i;
    for(i = st->clients; i != NULL; i = i->next) {
      assert(i->prev != NULL || st->clients == i);
    }
  }
  pthread_mutex_unlock(&st->lock);

  // Close connection
  close(h->sock);

  // Destroy structure
  rowset_destroy(h->iter.Vk);
  rowset_destroy(h->iter.VnbrA);
  rowset_destroy(h->iter.VnbrB);

  pthread_cond_destroy(&h->cond);
  pthread_mutex_destroy(&h->condlock);
  free(h);
}
