#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "sqrtterm.h"
#include "util/intarray.h"
#include "util/setdiff.h"
#include "util/geom.h"
#include "util/sort.h"

#include "util/debug.h"

///////////////////////////////////// MATH!!!!! ////////////////////////

static int group_extract_rowset(int n, const int *members, const rowset *V, double *GV)
{
  int i, j, nrows, p, nextracted;
  const int *rowinds;
  const double *Vdata;
 
  // Extract indices
  nrows = rowset_nrows(V);
  rowinds = rowset_inds(V);

  assert(issorted(n, members));
  assert(issorted(nrows, rowinds));

  p = rowset_ncols(V);
  Vdata = rowset_data(V);

  // Put rows into GV
  nextracted = i = j = 0;
  while(i < n && j < nrows) {
    if(members[i] == rowinds[j]) {
      memcpy(GV + p*i, Vdata + p*j, sizeof(double) * p);
      i++; j++; nextracted++;
    }
    else if(members[i] < rowinds[j]) {
      i++;
    }
    else {
      j++;
    }
  }

  return nextracted;
}

static void group_extract(int n, const int *members, const rowset *V, const rowset *Vnbr, double *GV)
{
  int nV, nnbr;

  nV = group_extract_rowset(n, members, V, GV);
  nnbr = group_extract_rowset(n, members, Vnbr, GV);

  assert(nV + nnbr == n);
}

/**
 * Calculate the mean of the rows in a must-link group
 */
static void rowmean(size_t n, size_t p, const double *GV, double *mu)
{
  size_t i;
  
  geom_vector_zero(p, mu);
  for(i = 0; i < n; i++) {
    geom_vector_add(p, mu, GV + p*i);
  }
  geom_vector_scale(p, mu, 1.0/((double)n));
}

static double squaredist(int n, const double *p, const double *q)
{
  int i;
  double diff, dist;

  dist = 0;
  for(i = 0; i < n; i++) {
    diff = p[i] - q[i];
    dist += diff*diff;
  }

  return dist;
}

static double group_obj(const mustlink *ml, int g, const rowset *V, const rowset *Vnbr, double *mu)
{
  int i, j, n, p;
  const int *members;
  double *GV, dist;

  // Get sizes and indices
  n = mustlink_group_nmembers(ml, g);
  members = mustlink_group_members(ml, g);
  p = rowset_ncols(V);
  assert(p == rowset_ncols(Vnbr));

  // Extract group
  GV = (double *)malloc(sizeof(double) * (n + 1) * p);
  assert(GV != NULL);
  if(mu == NULL) {
    mu = GV + n*p;
  }

  //fprintf(stderr, "g = %d\n", g);
  //fprintf(stderr, "members = %d %d %d\n", members[0], members[1], members[2]);
  group_extract(n, members, V, Vnbr, GV);

  // Get mean
  rowmean((size_t)n, (size_t)p, GV, mu);
  assert(!anynan(p, mu));

  // Iterate over members
  dist = 0;
  for(i = 0; i < n; i++) {
    dist += squaredist(p, GV + i*p, mu);
  }

  // Cleanup
  free(GV);

  assert(!isnan(dist));

  return sqrt(dist / (double)n);
}

/**
 * Determine the set of groups which includes rows in @c V
 */
static intarray *rowgroups(const mustlink *ml, const rowset *V)
{
  int i, n;
  const int *rowinds;
  intarray *groups, *unique;

  n = rowset_nrows(V);
  rowinds = rowset_inds(V);

  groups = intarray_create(0);
  for(i = 0; i < n; i++) {
    intarray_append(groups, mustlink_ningroups(ml, rowinds[i]), mustlink_membership(ml, rowinds[i]));
  }

  unique = intarray_unique(groups);
  intarray_destroy(groups);

  return unique;
}

double sqrt_obj(const mustlink *ml, const rowset *V, const rowset *Vnbr)
{
  intarray *groups;
  const int *groupinds;
  int i, j, m;
  double obj;

  // Assemble list of groups
  groups = rowgroups(ml, V);

  // Add objective term for each group
  obj = 0;
  m = intarray_length(groups);
  groupinds = intarray_contents_const(groups);
  for(i = 0; i < m; i++) {
    obj += group_obj(ml, groupinds[i], V, Vnbr, NULL);
  }

  intarray_destroy(groups);

  assert(!isnan(obj));

  return obj;
}

void sqrt_grad(const mustlink *ml, const rowset *V, const rowset *Vnbr, double *Gk)
{
  int i, j, k, l, m, n, p, *b, mi;
  intarray *groups;
  const int *rowinds, *groupinds, *groupsi;
  double *mu, *obj, *C;
  const double *Vdata, *row;

  n = rowset_nrows(V);
  rowinds = rowset_inds(V);
  p = rowset_ncols(V);
  Vdata = rowset_data(V);
  assert(p == rowset_ncols(Vnbr));

  geom_vector_zero(n*p, Gk);

  // Assemble list of groups
  groups = rowgroups(ml, V);

  m = intarray_length(groups);
  groupinds = intarray_contents_const(groups);

  //fprintf(stderr, "m = %d\n", m);

  // Calculate mean and objective for each group
  mu = (double *)malloc(sizeof(double) * m * (p+2));
  obj = mu + m*p;
  C = obj + m;

  for(i = 0; i < m; i++) {
    obj[i] = group_obj(ml, groupinds[i], V, Vnbr, mu + p*i);
    l = mustlink_group_nmembers(ml, groupinds[i]);
    C[i] = (double)l;
  }
  //fprintf(stderr, "group_obj = %g %g\n", obj[0], obj[1]);

  for(i = 0; i < n; i++) {
    mi = mustlink_ningroups(ml, rowinds[i]);
    groupsi = mustlink_membership(ml, rowinds[i]);

    //fprintf(stderr, "membership %d = ", i); print_ints(stderr, mi, groupsi); fprintf(stderr, "\n");

    row = Vdata + p*i;

    for(j = 0; j < mi; j++) {
      b = bsearch_ints((size_t)m, groupsi[j], (int *)groupinds);
      assert(b != NULL);
      k = b - groupinds;

      if(obj[k] == 0) {
        continue;
      }

      // Add component based on distance from kth row by groupinds
      for(l = 0; l < p; l++) {
        Gk[i*p + l] += (row[l] - mu[k*p + l]) / (C[k] * obj[k]);
      }
    }
  }

  // Cleanup
  free(mu);
  intarray_destroy(groups);

  assert(!anynan(n*p, Gk));
}

///////////////////////////// Server-side term definition ///////////////////

static intarray *dependent(void *user_data, int n, const int *inds)
{
  int i, j, m, *onlydeps;
  const int *l;
  intarray *groups, *depinds;
  mustlink *ml;

  ml = (mustlink *)user_data;

  // Get groups
  groups = intarray_create(0);
  for(i = 0; i < n; i++) {
    intarray_append(groups, mustlink_ningroups(ml, inds[i]), mustlink_membership(ml, inds[i]));
  }
  intarray_unique_update(groups);

  // Get all indices in those groups
  m = intarray_length(groups);
  l = intarray_contents_const(groups);

  depinds = intarray_create(0);
  for(i = 0; i < m; i++) {
    intarray_append(depinds, mustlink_group_nmembers(ml, l[i]), mustlink_group_members(ml, l[i]));
  }
  intarray_unique_update(depinds);

  // Filter out actual selected nodes
  m = intarray_length(depinds);
  l = intarray_contents_const(depinds);
  onlydeps = (int *)malloc(sizeof(int) * m);
  intsetdiff(m, l, n, inds, &m, onlydeps);

  // Cleanup
  intarray_destroy(groups);
  intarray_destroy(depinds);

  return intarray_create_from(m, onlydeps);
}

static size_t setup_buflen(void *user_data)
{
  return mustlink_serial_length((const mustlink *)user_data);
}

static int setup_serialize(void *user_data, size_t buflen, void *buf)
{
  return mustlink_serialize((const mustlink *)user_data, buflen, buf);
}

static size_t indupdate_buflen(void *user_data, int n, const int *inds)
{
  return 0;
}

static int indupdate_serialize(void *user_data, size_t buflen, void *buf, int n, const int *inds)
{
  // TODO?
  return 0;
}

static void destroy(void *user_data)
{
  mustlink_destroy((mustlink *)user_data);
}

void sqrtterm_create(termdef *def, double mult, mustlink *ml)
{
  def->obj = NULL;
  def->grad = NULL;

  def->dependent = dependent;

  def->setup_buflen = setup_buflen;
  def->setup_serialize = setup_serialize;
  def->setup_deserialize = NULL;

  def->indupdate_request = NULL;
  def->indupdate_buflen = indupdate_buflen;
  def->indupdate_serialize = indupdate_serialize;
  def->indupdate_deserialize = NULL;
 
  def->destroy = destroy;

  def->type = OBJTERM_TYPE_SQRT;
  def->mult = mult;
  def->user_data = ml;
}
