#include <assert.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>

#include "stiefelprob.h"

#include "traceterm.h"
#include "sqrtterm.h"

#include "util/debug.h"
#include "util/denseio.h"
#include "util/cachemat.h"
#include "util/matio.h"
#include "util/ident.h"
#include "util/path.h"

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

stiefel_prob *stiefel_prob_create(int nterms, int n, int p)
{
  stiefel_prob *prob;

  prob = (stiefel_prob *)malloc(sizeof(stiefel_prob));
  if(prob == NULL) {
    return NULL;
  }

  prob->nterms = nterms;
  prob->objdef = (termdef *)malloc(sizeof(termdef) * nterms);
  if(prob->objdef == NULL) {
    free(prob);
    return NULL;
  }

  prob->B = NULL;

  prob->n = n;
  prob->p = p;
  prob->V0 = (double *)malloc(sizeof(double) * n * p);
  if(prob->V0 == NULL) {
    free(prob->objdef);
    free(prob);
    return NULL;
  }

  return prob;
}

void stiefel_prob_destroy(stiefel_prob *prob)
{
  int i;

  if(prob == NULL) {
    return;
  }

  for(i = 0; i < prob->nterms; i++) {
    term_destroy(&prob->objdef[i]);
  }
  free(prob->objdef);

  free(prob->V0);
  sparsemat_destroy(prob->B);

  free(prob);
}

////////////////////////////// File parsing ////////////////////////////

typedef struct {
  double mult;
  objterm_type type;
  char *path;
} parse_objterm;

typedef struct {
  int nobj;
  parse_objterm *obj;

  char *V0;
  char *B;

  char *basepath;
} parse_prob;

static void parse_prob_init(parse_prob *parse, const char *basepath)
{
  parse->nobj = 0;
  parse->obj = NULL;
  parse->V0 = NULL;
  parse->B = NULL;

  if(basepath == NULL) {
    parse->basepath = NULL;
  }
  else {
    parse->basepath = strdup(basepath);
  }
}

static void parse_prob_cleanup(parse_prob *parse)
{
  int i;

  free(parse->basepath);

  free(parse->V0);
  free(parse->B);

  for(i = 0; i < parse->nobj; i++) {
    free(parse->obj[i].path);
  }
  free(parse->obj);
}

stiefel_prob *build_from_parse(parse_prob *parse)
{
  stiefel_prob *prob;
  int i, Vdims[2];
  double *V;

  cs_di *Af;
  sparsemat *A;
  mustlink *ml;

  termdef *defi;

  // Read in initial matrix
  if(parse->V0 == NULL) {
    return NULL;
  }
  V = dense_file_readall_rowmajor_path(parse->V0, 1, Vdims);
  if(V == NULL) {
    fprintf(stderr, "Could not read initial V from \"%s\"\n", parse->V0);
    return NULL;
  }

  assert(!anynan(Vdims[0] * Vdims[1], V));

  // Initialize prob
  prob = stiefel_prob_create(parse->nobj, Vdims[0], Vdims[1]);

  // Copy over matrix
  memcpy(prob->V0, V, sizeof(double) * Vdims[0] * Vdims[1]);
  free(V);

  // Read in objective
  for(i = 0; i < parse->nobj; i++) {
    defi = stiefel_prob_get_term(prob, i);

    switch(parse->obj[i].type) {
    case OBJTERM_TYPE_TRACE:
      // TODO: re-enable lazily loading matrices
      Af = sparsemat_file_read_all_path(parse->obj[i].path);
      if(Af == NULL) {
        fprintf(stderr, "Failed to sparse matrix from \"%s\"\n", parse->obj[i].path);
        stiefel_prob_destroy(prob);
        return NULL;
      }
      A = sparsemat_basic(Af, 1);
      traceterm_create(defi, parse->obj[i].mult, A); // TODO: have traceterm take ownership
      break;
    case OBJTERM_TYPE_SQRT:
      ml = mustlink_read_from_path(parse->obj[i].path);
      if(ml == NULL) {
        fprintf(stderr, "Failed to load mustlink constraints from \"%s\"\n", parse->obj[i].path);
        stiefel_prob_destroy(prob);
        return NULL;
      }
      sqrtterm_create(defi, parse->obj[i].mult, ml); // TODO: have sqrtterm take ownership
      break;
    default:
      stiefel_prob_destroy(prob);
      return NULL;
    }
  }

  // Read in ortho matrix
  if(parse->B == NULL) {
    A = sparsemat_basic(create_ident_matrix(Vdims[0]), 1);
  }
  else {
    // TODO: re-enable lazily loading matrices
    A = sparsemat_basic(sparsemat_file_read_all_path(parse->B), 1);
  }
  prob->B = A;
  
  // TODO : check sizes throughout

  return prob;
}

static int parse_line(parse_prob *parse, int ntok, char **tok)
{
  if(strcmp(tok[0], "obj") == 0) {
    if(ntok < 4) {
      fprintf(stderr, "Too few tokens\n");
      return 1;
    }

    // Allocate space
    parse->obj = (parse_objterm *)realloc(parse->obj, (parse->nobj + 1) * sizeof(parse_objterm));
    parse->obj[parse->nobj].mult = atof(tok[1]);
    parse->obj[parse->nobj].path = path_join(parse->basepath, tok[3]);

    if(strcmp(tok[2], "trace") == 0) {
      parse->obj[parse->nobj].type = OBJTERM_TYPE_TRACE;
    }
    else if(strcmp(tok[2], "sqrt") == 0) {
      parse->obj[parse->nobj].type = OBJTERM_TYPE_SQRT;
    }
    else {
      fprintf(stderr, "%s:%d Unrecognized objective term \"%s\"\n", __FILE__, __LINE__, tok[2]);
      return 1;
    }

    parse->nobj++;

    return 0;
  }
  else if(strcmp(tok[0], "V0") == 0) {
    if(ntok < 2) {
      fprintf(stderr, "Too few tokens\n");
      return 1;
    }

    parse->V0 = path_join(parse->basepath, tok[1]);
    return 0;
  }
  else if(strcmp(tok[0], "orthmat") == 0) {
    if(ntok < 2) {
      fprintf(stderr, "Too few tokens\n");
      return 1;
    }

    parse->B = path_join(parse->basepath, tok[1]);
    return 0;
  }


  fprintf(stderr, "Unrecognized line prefix \"%s\"\n", tok[0]);
  return 1;
}

stiefel_prob *stiefel_prob_from_file(FILE *fp, const char *basepath)
{
  stiefel_prob *prob;
  ssize_t r;

  size_t nline;
  char *line, *comment;

  int i, s, ntok, alloctok;
  char *toki, **tok, *lasts;

  parse_prob parse;

  nline = 0;
  line = NULL;

  alloctok = 0;
  tok = NULL;

  parse_prob_init(&parse, basepath);
  s = 0;

  r = getline(&line, &nline, fp);
  while(r >= 0) {
    // Remove comments
    comment = strchr(line, '#');
    if(comment != NULL) {
      *comment = '\0';
    }

    // Tokenize string by whitespace
    ntok = 0;
    toki = strtok_r(line, " \t\n", &lasts);
    while(toki != NULL) {
      if(*toki != '\0') {
        if(ntok >= alloctok) {
          alloctok = MAX(alloctok * 2, ntok + 1);
          tok = (char **)realloc(tok, sizeof(char *) * alloctok);
        }

        tok[ntok++] = strdup(toki);
      }

      toki = strtok_r(NULL, " \t\n", &lasts);
    }
    

    // Interpret tokens
    if(ntok != 0) {
      s = parse_line(&parse, ntok, tok);
    }
    else {
      s = 0;
    }


    // Cleanup tokens
    for(i = 0; i < ntok; i++) {
      free(tok[i]);
      tok[i] = NULL;
    }

    if(s != 0) {
      break;
    }

    r = getline(&line, &nline, fp);
  }

  // Construct problem structure
  prob = NULL;
  if(s == 0) {
    prob = build_from_parse(&parse);
  }

  // Cleanup
  free(tok);
  free(line);
  parse_prob_cleanup(&parse);

  return prob;
}

stiefel_prob *stiefel_prob_from_path(const char *path)
{
  FILE *fp;
  char *dir;
  stiefel_prob *prob;

  dir = path_dirname(path);

  fp = fopen(path, "r");
  prob = stiefel_prob_from_file(fp, dir);
  fclose(fp);

  free(dir);

  return prob;
}


////////////////////////////////// Accessors ///////////////////////////

int stiefel_prob_nterms(const stiefel_prob *prob)
{
  return prob->nterms;
}

int stiefel_prob_n(const stiefel_prob *prob)
{
  return prob->n;
}

int stiefel_prob_p(const stiefel_prob *prob)
{
  return prob->p;
}

termdef *stiefel_prob_get_term(stiefel_prob *prob, int i)
{
  return &prob->objdef[i];
}

const termdef *stiefel_prob_get_term_const(const stiefel_prob *prob, int i)
{
  return &prob->objdef[i];
}

sparsemat *stiefel_prob_get_ortho(stiefel_prob *prob)
{
  return prob->B;
}

const sparsemat *stiefel_prob_get_ortho_const(const stiefel_prob *prob)
{
  return prob->B;
}
