#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "util/geom.h"

#include "trace.h"

static double fudge_factor = 1e4;

static double dotprod(int p, const double *u, const double *v)
{
  double dotprod;
  int i;

  if(p < 1) {
    return 0;
  }

  dotprod = fudge_factor * fudge_factor * u[0] * v[0];
  for(i = 1; i < p; i++) {
    dotprod += u[i] * v[i];
  }

  return dotprod;
}

double trace_partial_obj(sparsemat *A, const rowset *V, const rowset *Vnbr)
{
  int j, u, v, ind, k, p, dims[2];
  const int *K;
  const double *Vu, *Vv;

  double trace, Auv;

  sparsecol col;
  int colnnz;
  const int *coli;
  const double *colx;

  k = rowset_nrows(V);
  K = rowset_inds(V);
  p = rowset_ncols(V);

  sparsemat_dims(A, dims);
  assert(dims[0] == dims[1]);
  assert(p == rowset_ncols(Vnbr));

  trace = 0;
  for(j = 0; j < k; j++) {
    u = K[j];
    Vu = rowset_get(V, u);
    assert(Vu != NULL);

    col = sparsemat_col(A, u);

    for(ind = 0; ind < col.nnz; ind++) {
      v = col.i[ind];
      Auv = col.v[ind];

      Vv = rowset_get(V, v);
      if(Vv == NULL) {
        Vv = rowset_get(Vnbr, v);
      }
      assert(Vv != NULL);

      trace += Auv * geom_dot_product((size_t)p, Vu, Vv);
      //trace += Auv * dotprod(p, Vu, Vv);
    }
  }

  return trace;
}

void trace_partial_grad(sparsemat *A, const rowset *V, const rowset *Vnbr, double *Gk)
{
  int dims[2];
  int i, j, k, p, ind, l, u, v;
  const int *K;
  double Auv;
  const double *Vv;
  sparsecol col;

  k = rowset_nrows(V);
  K = rowset_inds(V);
  p = rowset_ncols(V);

  sparsemat_dims(A, dims);
  assert(dims[0] == dims[1]);
  assert(p == rowset_ncols(Vnbr));

  // Zero out the gradient
  for(i = 0; i < k*p; i++) {
    Gk[i] = 0;
  }

  // Perform the matrix mulitiplications
  for(j = 0; j < k; j++) {
    u = K[j];
    col = sparsemat_col(A, u);

    for(ind = 0; ind < col.nnz; ind++) {
      v = col.i[ind];
      Auv = col.v[ind];

      Vv = rowset_get(V, v);
      if(Vv == NULL) {
        Vv = rowset_get(Vnbr, v);
      }
      assert(Vv != NULL);

      for(l = 0; l < p; l++) {
        Gk[k*l + j] += Auv * Vv[l];
      }
    }
  }

  /*
  for(i = 0; i < k; i++) {
    Gk[i] *= fudge_factor;
  }
  */

  // Multiply in factor of 2
  for(i = 0; i < k*p; i++) {
    Gk[i] *= 2;
  }
}

double sparse_trace_full_obj(sparsemat *A, int p, const double *V)
{
  int dims[2], i, ind, n, u, v;
  double trace, Auv;
  sparsecol col;

  int colnnz;
  const int *coli;
  const double *colx;

  sparsemat_dims(A, dims);
  assert(dims[0] == dims[1]);
  n = dims[0];

  trace = 0;
  for(u = 0; u < n; u++) {
    col = sparsemat_col(A, u);

    for(ind = 0; ind < col.nnz; ind++) {
      v = col.i[ind];
      Auv = col.v[ind];
      
      trace += Auv * geom_dot_product((size_t)p, V + u*p, V + v*p);
      //trace += Auv * dotprod(p, V + u*p, V + v*p);
    }
  }

  return trace;
}

double multitrace_partial_obj(int nmat, sparsemat **A, const rowset *V, const rowset *Vnbr)
{
  double obj;
  int i;

  obj = 0;
  for(i = 0; i < nmat; i++) {
    obj += trace_partial_obj(A[i], V, Vnbr);
  }

  return obj;
}

void multitrace_partial_grad(int nmat, sparsemat **A, const rowset *V, const rowset *Vnbr, double *Gk)
{
  double *Gki;
  int i, j, k, p;
  
  k = rowset_nrows(V);
  p = rowset_ncols(V);
  assert(rowset_ncols(Vnbr) == p);

  Gki = (double *)malloc(sizeof(double) * k * p);

  for(j = 0; j < k*p; j++) {
    Gk[j] = 0;
  }

  for(i = 0; i < nmat; i++) {
    trace_partial_grad(A[i], V, Vnbr, Gki);
    for(j = 0; j < k*p; j++) {
      Gk[j] += Gki[j];
    }
  }

  free(Gki);
}

