#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <cs.h>
#include <math.h>
#include <sys/resource.h>
#include <ctype.h>

#include "debug.h"


static void print_matrix_impl(
	FILE *stream,
	const char *float_format, const char *indent,
	int rows, int cols,
	const double *A, int numbering, int rowmajor)
{
	int i, j, k;
	char buf[256], pad;
  double entry;

	if(rows <= 0 || cols <= 0) {
		return;
	}

  if(numbering) {
    numbering = 1;
  }
  assert(numbering == 0 || numbering == 1);

	int len[(rows + numbering)*(cols + numbering)], width[cols + numbering];
	char *s[(rows + numbering)*(cols + numbering)];

	// Format numbers to strings
  for(j = 0; j < cols + numbering; j++) {
    for(i = 0; i < rows + numbering; i++) {
      // Print matrix entry or row/column number
      if(numbering) {
        if(i == 0) {
          if(j == 0) {
            buf[0] = '\0';
          }
          else {
            snprintf(buf, 256, "%d", j-1);
          }
        }
        else if(j == 0) {
          snprintf(buf, 256, "%d", i-1);
        }
        else {
          if(rowmajor) {
            entry = A[cols*(i-1) + (j-1)];
          }
          else {
            entry = A[rows*(j-1) + (i-1)];
          }
          snprintf(buf, 256, float_format, entry);
        }
      }
      else {
        if(rowmajor) {
          entry = A[cols*i + j];
        }
        else {
          entry = A[rows*j + i];
        }
        snprintf(buf, 256, float_format, entry);
      }
      buf[255] = '\0';

      // Copy into string array
      k = (rows + numbering) * j + i;
      if(strcmp(buf, "-0") == 0) {
        s[k] = strdup("0");
      }
      else {
        s[k] = strdup(buf);
      }
      len[k] = strlen(s[k]);
    }
  }

	// Calculate column width
	for(j = 0; j < cols + numbering; j++) {
		width[j] = 0;
	}

	for(i = 0; i < rows + numbering; i++) {
		for(j = 0; j < cols + numbering; j++) {
			if(len[(rows + numbering)*j + i] > width[j]) {
				width[j] = len[(rows + numbering)*j + i];
			}
		}
	}

	// Print the matrix
	for(i = 0; i < rows + numbering; i++) {
		fputs(indent, stream);

		for(j = 0; j < cols + numbering; j++) {
      if(j == numbering) {
        if(numbering) {
          fputc(' ', stream);
        }
        fputc((i == numbering) ? '[' : ' ', stream);
      }

			pad = width[j] - len[(rows + numbering)*j + i];

			if(!(j == 0 || (numbering && (j == 1 || i == 0)))) {
				fputc(',', stream);
				pad++;
			}

			for(k = 0; k < pad; k++) {
				fputc(' ', stream);
			}

			fputs(s[(rows + numbering)*j + i], stream);
		}

    if(!(numbering && i == 0)) {
      fputc((i == rows+numbering-1) ? ']' : ';', stream);
    }
		fputc('\n', stream);
	}

	// Cleanup
	for(i = 0; i < (rows + numbering)*(cols + numbering); i++) {
		free(s[i]);
	}
}

void print_matrix(FILE *stream,	const char *float_format, const char *indent, 	int rows, int cols, const double *A)
{
  print_matrix_impl(stream, float_format, indent, rows, cols, A, 0, 0);
}

void print_matrix_rowmajor(FILE *stream, const char *float_format, const char *indent, int rows, int cols, const double *A)
{
  print_matrix_impl(stream, float_format, indent, rows, cols, A, 0, 1);
}

void print_matrix_numbered(FILE *stream, const char *float_format, const char *indent, int rows, int cols, const double *A)
{
  print_matrix_impl(stream, float_format, indent, rows, cols, A, 1, 0);
}

void print_sparse(FILE *stream, const char *indent, cs_di *A)
{
  int i, j;

  if(A->nz == -1) {
    for(j = 0; j < A->n; j++) {
      for(i = A->p[j]; i < A->p[j+1]; i++) {
        fprintf(stream, "%s%d %d %g\n", indent, A->i[i], j, A->x[i]);
      }
    }
  }
  else {
    for(i = 0; i < A->nz; i++) {
      fprintf(stream, "%s%d %d %g\n", indent, A->i[i], A->p[i], A->x[i]);
    }
  }
}

void print_ints(FILE *stream, int n, const int *l)
{
  int i;

  if(n == 0) {
    return;
  }

  fprintf(stream, "%d", l[0]);
  for(i = 1; i < n; i++) {
    fprintf(stream, " %d", l[i]);
  }
}

int anynan(int n, const double *ds)
{
  int i;

  for(i = 0; i < n; i++) {
    if(isnan(ds[i])) {
      return 1;
    }
  }

  return 0;
}

int issorted(int n, const int *l)
{
  int i;

  for(i = 0; i < n - 1; i++) {
    if(l[i+1] < l[i]) {
      return 0;
    }
  }

  return 1;
}

#define KB_SIZE (1024)
#define MB_SIZE (KB_SIZE * 1024)
#define GB_SIZE (MB_SIZE * 1024)
#define TB_SIZE (GB_SIZE * 1024)

void strmemsize(size_t sz, size_t buflen, char *buf)
{
  double szflt;
  const char *suffix;

  if(sz > 1280 * (size_t)GB_SIZE) {
    szflt = (double)sz / 1099511627776.0;
    suffix = "TB";
  }
  else if(sz > 1280 * MB_SIZE) {
    szflt = (double)sz / GB_SIZE;
    suffix = "GB";
  }
  else if(sz > 1280 * KB_SIZE) {
    szflt = (double)sz / MB_SIZE;
    suffix = "MB";
  }
  else if(sz > 1280) {
    szflt = (double)sz / KB_SIZE;
    suffix = "KB";
  }
  else {
    snprintf(buf, buflen, "%luB", sz);
    buf[buflen-1] = '\0';
    return;
  }

  snprintf(buf, buflen, "%.2lf%s", szflt, suffix);
  buf[buflen-1] = '\0';
}

void printmemsize(size_t sz, FILE *stream)
{
  char buf[1024];
  
  strmemsize(sz, 1024, buf);
  fputs(buf, stream);
}

size_t getmemusage()
{
  FILE *statfp;
  ssize_t r;
  char *lineptr, *lineend;
  size_t n;
  int sz;

  char *name, *value, *valueend;

  lineptr = NULL;
  n = 0;
  r = 0;

  sz = 0;
  
  statfp = fopen("/proc/self/status", "r");
  while(!feof(statfp) && r >= 0) {
    r = getline(&lineptr, &n, statfp);
    lineend = lineptr + strlen(lineptr);

    for(name = lineptr; name < lineend && isspace(*name); name++);
    if(name >= lineend) {
      continue;
    }

    for(value = name; value < lineend && (*value) != ':'; value++);
    if(value + 1 >= lineend) {
      continue;
    }

    value[0] = '\0';
    value++;

    if(strcmp(name, "VmSize") == 0) {
      sscanf(value, "%d kB", &sz);
    }
  }
  fclose(statfp);

  free(lineptr);

  return (size_t)sz;
}



void strtime(double t, size_t buflen, char *buf)
{
	if(t < 2 * 60) {
		snprintf(buf, buflen, "%.2lfs", t);
	}
	else if(t < 5 * 60) {
		snprintf(buf, buflen, "%dm%.2lfs", (int)(t/60), fmod(t, 60));
	}
	else if(t < 2 * 60 * 60) {
		snprintf(buf, buflen, "%.2lfm", t/60);
	}
	else if(t < 24 * 60 * 60) {
		snprintf(buf, buflen, "%dh%.2lfm", (int)(t/(60*60)), fmod(t/60, 60));
	}
	else {
		snprintf(buf, buflen, "%dd%.2lfh", (int)(t/(24*60*60)), fmod(t/(60*60), 24));
	}

  buf[buflen-1] = '\0';
}

void printtime(double sec, FILE *stream)
{
  char buf[256];

  strtime(sec, 256, buf);
  fputs(buf, stream);
}

void printcurrenttime(FILE *stream)
{
  time_t now;
  struct tm localnow;
  char buf[32];

  now = time(NULL);
  localtime_r(&now, &localnow);
  asctime_r(&localnow, buf);
  buf[31] = '\0';

  fputs(buf, stream);
}

double timeelapsed(struct timeval *start)
{
  struct timeval now;
  double s, us;

  gettimeofday(&now, NULL);

  s = (double)now.tv_sec - (double)start->tv_sec;
  us = (double)now.tv_usec - (double)start->tv_usec;

  return s + 1e-6 * us;
}
