#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <ctype.h>
#include <memory.h>
#include <assert.h>

#include "denseio.h"

struct dense_file
{
  FILE *fp;
  int bin;
  int dims[2];
  int pos[2];

  double *A;
};

//////////////////////////////// Reading ////////////////////////////

static int txtdims(FILE *fp, int dims[2]);

dense_file *dense_file_open(const char *path, int bin, int preload)
{
  dense_file *df;
  uint64_t fdims[2];
  size_t nread;
  int i, r, d, nel;

  // Check options
  df = (dense_file *)malloc(sizeof(dense_file));
  if(df == NULL) {
    return NULL;
  }

  df->fp = fopen(path, bin ? "rb" : "r");
  if(df->fp == NULL) {
    free(df);
    return NULL;
  }
  df->bin = bin;

  if(bin) {
    nread = fread(fdims, sizeof(uint64_t), 2, df->fp);
    if(nread != 2) {
      dense_file_close(df);
      return NULL;
    }

    for(d = 0; d < 2; d++) {
      df->dims[d] = (int)fdims[d];
    }
  }
  else {
    r = txtdims(df->fp, df->dims);
    if(r != 0) {
      dense_file_close(df);
      return NULL;
    }
  }

  df->A = NULL;
  if(preload) {
    nel = df->dims[0] * df->dims[1];
    df->A = (double *)malloc(sizeof(double) * nel);
    if(df->A == NULL) {
      dense_file_close(df);
      return NULL;
    }

    if(bin) {
      nread = fread(df->A, sizeof(double), nel, df->fp);
      if(nread != (size_t)nel) {
        dense_file_close(df);
        return NULL;
      }
    }
    else {
      for(i = 0; i < nel; i++) {
        r = fscanf(df->fp, "%lg", &df->A[i]);
        if(r != 1) {
          dense_file_close(df);
          return NULL;
        }
      }
    }
  }

  df->pos[0] = df->pos[1] = 0;
     

  return df;
}


void dense_file_close(dense_file *df)
{
  if(df == NULL) {
    return;
  }

  free(df->A);
  fclose(df->fp);
  free(df);
}

static int txtdims(FILE *fp, int dims[2])
{
  off_t prevoff;
  int c, inspace, m, n, j;

  prevoff = ftello(fp);
  fseeko(fp, 0, SEEK_SET);

  m = n = j = 0;
  inspace = 1;
  while((c = fgetc(fp)) != EOF) {
    if(inspace) {
      if(!isspace(c)) {
        j++;
        inspace = 0;
      }
    }
    else if(isspace(c)) {
      inspace = 1;
    }

    if(c == '\n') {
      if(m == 0) {
        n = j;
      }
      else if(n != j) {
        return -1;
      }
      m++;
      j = 0;
    }
  }

  fseeko(fp, prevoff, SEEK_SET);

  dims[0] = m;
  dims[1] = n;

  return 0;
}

void dense_file_dims(dense_file *df, int dims[2])
{
  int d;

  for(d = 0; d < 2; d++) {
    dims[d] = df->dims[d];
  }
}

int dense_file_eof(dense_file *df)
{
  return df->pos[0] == df->dims[0];
}

int dense_file_tell(dense_file *df, int pos[2])
{
  int d;

  for(d = 0; d < 2; d++) {
    pos[d] = df->pos[d];
  }

  return 0;
}


int dense_file_seek(dense_file *df, int pos[2])
{
  off_t off;
  int i, j, n, d, r;

  if(!df->bin) {
    return -1;
  }

  i = pos[0];
  j = pos[1];
  n = df->dims[1];

  off = sizeof(uint64_t) * 2 + sizeof(double) * (off_t)(n * i + j);
  r = fseeko(df->fp, off, SEEK_SET);
  if(r != 0) {
    return r;
  }

  for(d = 0; d < 2; d++) {
    df->pos[d] = pos[d];
  }

  return 0;
}

double dense_file_next(dense_file *df)
{
  double x;
  size_t r;
  int s;

  if(df->A != NULL) {
    x = df->A[df->pos[0] * df->dims[1] + df->pos[1]];
  }
  else if(df->bin) {
    r = fread(&x, sizeof(double), 1, df->fp);
    if(r != 1) {
      // TODO: error condition?
    }
  }
  else {
    s = fscanf(df->fp, "%lg", &x);
    if(s != 1) {
      // TODO: error condition?
    }
  }

  // Update position
  df->pos[1]++;
  if(df->pos[1] == df->dims[1]) {
    df->pos[0]++;
    df->pos[1] = 0;
  }

  return x;
}

int dense_file_row(dense_file *df, double *row)
{
  int i, n, s, read;
  size_t r;

  if(df->pos[1] != 0) {
    return -1;
  }

  n = df->dims[1];
  if(df->A != NULL) {
    memcpy(row, df->A + n * df->pos[0], sizeof(double) * n);
  }
  else if(df->bin) {
    r = fread(row, sizeof(double), (size_t)n, df->fp);
    if(r != (size_t)n) {
      return -1;
    }
  }
  else {
    for(i = 0; i < n; i++) {
      s = fscanf(df->fp, "%lg", &row[i]);
      if(s != 1) {
        return -1;
      }
    }
  }

  // Update position
  df->pos[0]++;

  return n;
}

int dense_file_readall(dense_file *df, double *A)
{
  double *row;
  int i, j, m, n, r;

  m = df->dims[0];
  n = df->dims[1];
  
  row = (double *)malloc(sizeof(double) * n);
  if(row == NULL) {
    return -1;
  }
  
  for(i = 0; i < m; i++) {
    r = dense_file_row(df, row);
    if(r != n) {
      return -1;
    }

    for(j = 0; j < n; j++) {
      A[i + m*j] = row[j];
    }
  }

  free(row);

  return m * n;
}

int dense_file_readall_rowmajor(dense_file *df, double *A)
{
  double *row;
  int i, j, m, n, r;

  m = df->dims[0];
  n = df->dims[1];

  row = (double *)malloc(sizeof(double) * n);
  if(row == NULL) {
    return -1;
  }

  for(i = 0; i < m; i++) {
    r = dense_file_row(df, row);
    if(r != n) {
      return -1;
    }

    memcpy(A + i*n, row, sizeof(double) * n);
  }

  free(row);

  return m * n;
}

double *dense_file_readall_path(const char *path, int bin, int dims[2])
{
  dense_file *df;
  int m, n, r;
  double *A;

  // Open file
  df = dense_file_open(path, bin, 0);
  if(df == NULL) {
    return NULL;
  }

  // Allocate output matrix
  dims[0] = m = df->dims[0];
  dims[1] = n = df->dims[1];
  A = (double *)malloc(sizeof(double) * m * n);
  if(A == NULL) {
    dense_file_close(df);
    return NULL;
  }

  // Read matrix
  r = dense_file_readall(df, A);
  if(r != m*n) {
    free(A);
    dense_file_close(df);
    return NULL;
  }

  dense_file_close(df);

  return A;
}

double *dense_file_readall_rowmajor_path(const char *path, int bin, int dims[2])
{
  dense_file *df;
  int m, n, r;
  double *A;

  // Open file
  df = dense_file_open(path, bin, 0);
  if(df == NULL) {
    return NULL;
  }

  // Allocate output matrix
  dims[0] = m = df->dims[0];
  dims[1] = n = df->dims[1];
  A = (double *)malloc(sizeof(double) * m * n);
  if(A == NULL) {
    dense_file_close(df);
    return NULL;
  }

  // Read matrix
  r = dense_file_readall_rowmajor(df, A);
  if(r != m*n) {
    free(A);
    dense_file_close(df);
    return NULL;
  }
  
  dense_file_close(df);

  return A;
}

//////////////////////////////// Writing ///////////////////////////

int dense_file_write(const char *path, int bin, int m, int n, const double *A)
{
  int i, j, r;
  FILE *fp;
  uint64_t fdims[2];
  size_t written;

  // Output
  fp = fopen(path, bin ? "wb+" : "w+");
  if(fp == NULL) {
    return -1;
  }

  if(bin) {
    // Write dimensions
    fdims[0] = (uint64_t)m;
    fdims[1] = (uint64_t)n;
    written = fwrite(fdims, sizeof(uint64_t), 2, fp);
    if(written != 2) {
      return -1;
    }

    // Write matrix doubles
    for(i = 0; i < m; i++) {
      for(j = 0; j < n; j++) {
        written = fwrite(&A[m*j + i], sizeof(double), 1, fp);
        if(written != 1) {
          return -1;
        }
      }
    }
  }
  else {
    for(i = 0; i < m; i++) {
      for(j = 0; j < n; j++) {
        r = fprintf(fp, "% .16e%c", A[m*j + i], j == n-1 ? '\n' : ' ');
        if(r < 0) {
          return -1;
        }
      }
    }
  }  

  r = fclose(fp);
  if(r != 0) {
    return -1;
  }

  return 0;
}

int dense_file_write_rowmajor(const char *path, int bin, int m, int n, const double *A)
{
  int i, j, r;
  FILE *fp;
  uint64_t fdims[2];
  size_t written;

  // Output
  fp = fopen(path, bin ? "wb+" : "w+");
  if(fp == NULL) {
    return -1;
  }

  if(bin) {
    // Write dimensions
    fdims[0] = (uint64_t)m;
    fdims[1] = (uint64_t)n;
    written = fwrite(fdims, sizeof(uint64_t), 2, fp);
    if(written != 2) {
      return -1;
    }

    // Write matrix doubles
    written = fwrite(A, sizeof(double), (size_t)m * (size_t)n, fp);
    if(written != (size_t)m * (size_t)n) {
      return -1;
    }
  }
  else {
    for(i = 0; i < m; i++) {
      for(j = 0; j < n; j++) {
        r = fprintf(fp, "% .16e%c", A[n*i + j], j == n-1 ? '\n' : ' ');
        if(r < 0) {
          return -1;
        }
      }
    }
  }  

  r = fclose(fp);
  if(r != 0) {
    return -1;
  }

  return 0;
}
