#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "randsample.h"

#include "file_list.h"

struct file_list
{
  size_t n;
  char **s;
};

file_list *file_list_create(int n)
{
  file_list *fl;
  int i;

  // Allocate
  fl = (file_list *)malloc(sizeof(file_list));
  if(fl == NULL) {
    return NULL;
  }

  fl->n = n;
  fl->s = (char **)calloc(n, sizeof(char *));
  if(fl->s == NULL) {
    free(fl);
    return NULL;
  }

  return fl;
}

void file_list_destroy(file_list *fl)
{
  int i, n;

  if(fl == NULL) {
    return;
  }

  n = fl->n;
  for(i = 0; i < n; i++) {
    free(fl->s[i]);
  }
  free(fl->s);
  free(fl);
}

static char *strip(const char *in)
{
  const char *start, *end;
  
  end = in + strlen(in);
  for(start = in; isspace(*start) && start != end; start++);
  for(end--; isspace(*end) && end > start; end--);
  end++;

  return strndup(start, end - start);
}

file_list *file_list_load_fp(FILE *fp)
{
  size_t alloc, linen;
  char *lineptr, *path, *c;

  file_list *fl;

  alloc = 32;
  linen = 0;
  lineptr = NULL;

  fl = (file_list *)malloc(sizeof(file_list));
  fl->n = 0;
  fl->s = (char **)malloc(sizeof(char *) * alloc);


  while(!feof(fp)) {
    if(getline(&lineptr, &linen, fp) == -1) {
      break;
    }

    for(c = lineptr; *c != '\0'; c++) { // remove comments
      if(*c == '#' && (c == lineptr || *(c - 1) != '\\')) {
        *c = '\0';
        break;
      }
    }

    path = strip(lineptr);

    // Skip blank lines and comments
    if(strlen(path) == 0) {
      free(path);
      continue;
    }

    // Ensure space
    if(fl->n >= alloc) {
      alloc *= 2;
      fl->s = (char **)realloc(fl->s, sizeof(char *) * alloc);
    }

    // Append to path list
    fl->s[fl->n++] = path;
  }

  // Cleanup
  free(lineptr);

  return fl;
}

file_list *file_list_load(const char *path)
{
  FILE *fp;
  file_list *fl;

  fp = fopen(path, "r");
  if(fp == NULL) {
    return NULL;
  }

  fl = file_list_load_fp(fp);
  
  fclose(fp);
  return fl;
}

void file_list_write_fp(FILE *fp, const file_list *fl)
{
  int i, n;

  n = file_list_length(fl);
  for(i = 0; i < n; i++) {
    fprintf(fp, "%s\n", file_list_get(fl, i));
  }
}


int file_list_length(const file_list *fl)
{
  return (int)fl->n;
}

const char *file_list_get(const file_list *fl, int i)
{
  assert(i >= 0 && (size_t)i < fl->n);
  return fl->s[i];
}

void file_list_set(file_list *fl, int i, const char *s)
{
  assert(i >= 0 && (size_t)i < fl->n);
  fl->s[i] = strdup(s);
}

file_list *file_list_merge(const file_list *a, const file_list *b)
{
  int i, na, nb, nm;
  file_list *m;

  na = file_list_length(a);
  nb = file_list_length(b);
  nm = na + nb;

  m = file_list_create(nm);

  for(i = 0; i < na; i++) {
    file_list_set(m, i, file_list_get(a, i));
  }
  for(i = 0; i < nb; i++) {
    file_list_set(m, na + i, file_list_get(b, i));
  }

  return m;
}

file_list *file_list_sample(const file_list *fl, int n, const int *ind)
{
  file_list *sub;
  int i;

  sub = file_list_create(n);
  for(i = 0; i < n; i++) {
    file_list_set(sub, i, file_list_get(fl, ind[i]));
  }

  return sub;
}

file_list *file_list_randsample(const file_list *fl, int n, int replacement)
{
  int m, *ind;
  file_list *sub;

  m = file_list_length(fl);
  ind = (int *)malloc(sizeof(int) * n);
  randsample(ind, m, n, replacement);

  sub = file_list_sample(fl, n, ind);

  free(ind);

  return sub;
}
