#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <assert.h>

#include "geom.h"
#include "gonzalez85.h"
#include "kmeans_util.h"

#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif // #ifndef INFINITY

size_t gonzalez85_closure(
  size_t n, size_t k, double mindist, int *part, size_t *head,
  double (*distance)(void *user_data, size_t i, size_t j), void *user_data)
{
  size_t i, j, l, vi;
  size_t *freehead;
  double h, *dist, d;
  int *freepart;

  freehead = NULL;
  if(head == NULL) {
    freehead = head = (size_t *)malloc(sizeof(size_t) * k);
  }
  freepart = NULL;
  if(part == NULL) {
    freepart = part = (int *)malloc(sizeof(int) * n);
  }

  dist = (double *)malloc(sizeof(double) * n);

  // Start everything in first cluster with first point at head
  assert(k >= 1);
  head[0] = 0;
  for(i = 0; i < n; i++) {
    part[i] = 0;
    dist[i] = distance(user_data, 0, i);
  }

  for(l = 1; l < k; l++) {
    // Find most distant point from its center
    h = -INFINITY;
    for(i = 1; i < n; i++) {
      if(dist[i] > h) {
        h = dist[i];
        head[l] = i;
      }
    }

    if(h < mindist) {
      k = l;
      break;
    }

    // Move points into new cluster
    for(i = 1; i < n; i++) {
      d = distance(user_data, head[l], i);

      if(d < dist[i]) {
        dist[i] = d;
        part[i] = l;
      }
    }
  }


  // Cleanup
  free(dist);
  free(freehead);
  free(freepart);

  return k;
}

static size_t gonzalez85_real(
  size_t n, size_t ndims, const double *p,
  size_t k, double mindist, double *C, int *part)
{
  size_t i, j, l, vi;
  const double **head;
  double h, *dist, d;
  double *ones;

  head = (const double **)malloc(sizeof(double *) * k);
  ones = (double *)malloc(sizeof(double) * n * 2);
  dist = ones + n;

  // Start everything in first cluster with first point at head
  assert(k >= 1);
  head[0] = p;
  for(i = 0; i < n; i++) {
    part[i] = 0;
    ones[i] = 1.0;
    dist[i] = geom_distance(ndims, p, p + ndims*i);
  }

  for(l = 1; l < k; l++) {
    // Find most distant point from its center
    h = -INFINITY;
    for(i = 1; i < n; i++) {
      if(dist[i] > h) {
        h = dist[i];
        head[l] = p + ndims*i;
      }
    }

    if(h < mindist) {
      k = l;
      break;
    }

    // Move points into new cluster
    for(i = 1; i < n; i++) {
      d = geom_distance(ndims, head[l], p + ndims * i);

      if(d < dist[i]) {
        dist[i] = d;
        part[i] = l;
      }
    }
  }

  // Calculate centers
  cluster_centers(n, ndims, p, ones, k, part, C);

  // Cleanup
  free(ones);
  free(head);

  return k;
}

size_t gonzalez85(
  size_t n, size_t ndims, const double *p,
  size_t k, double mindist,
  double *C, int *part)
{
  size_t ret;

  // Ensure C is non-null
  if(C == NULL) {
    C = (double *)malloc(n * ndims * sizeof(double));
    ret = gonzalez85(n, ndims, p, k, mindist, C, part);
    free(C);
    return ret;
  }

  // Ensure part is non-null
  if(part == NULL) {
    part = (int *)malloc(n * sizeof(int));
    ret = gonzalez85(n, ndims, p, k, mindist, C, part);
    free(part);
    return ret;
  }

  return gonzalez85_real(n, ndims, p, k, mindist, C, part);
}
