#include <stdlib.h>
#include <memory.h>
#include <assert.h>

#include "lazymat.h"
#include "sparsemat_internal.h"

static const sparsemat_vtable vtable;

struct lazymat
{
  sparsemat base;

  int *nnz;
  int **i;
  double **v;
};

static inline lazymat *cast(sparsemat *mat)
{
  assert(mat->v == &vtable);
  return (lazymat *)mat;
}

static inline const lazymat *cast_const(const sparsemat *mat)
{
  assert(mat->v == &vtable);
  return (const lazymat *)mat;
}

lazymat *lazymat_create(int dims[2])
{
  lazymat *lz;
  int i, j, n;

  lz = (lazymat *)malloc(sizeof(lazymat));
  if(lz == NULL) {
    return NULL;
  }

  lz->base.v = &vtable;
  for(i = 0; i < 2; i++) {
    lz->base.dims[i] = dims[i];
  }
  n = dims[1];

  lz->nnz = NULL;
  lz->i = NULL;
  lz->v = NULL;

  lz->nnz = (int *)malloc(sizeof(int) * n);
  if(lz->nnz == NULL) {
    free(lz);
    return NULL;
  }

  lz->i = (int **)malloc(sizeof(int *) * n);
  if(lz->i == NULL) {
    free(lz->nnz); free(lz);
    return NULL;
  }

  lz->v = (double **)malloc(sizeof(double *) * n);
  if(lz->v == NULL) {
    free(lz->nnz); free(lz->i); free(lz);
    return NULL;
  }

  for(j = 0; j < n; j++) {
    lz->nnz[j] = -1;
    lz->i[j] = NULL;
    lz->v[j] = NULL;
  }

  return lz;
}

static void destroy(sparsemat *base)
{
  lazymat *lz;
  int j, n;

  lz = cast(base);

  n = lz->base.dims[1];
  for(j = 0; j < n; j++) {
    free(lz->i[j]);
    free(lz->v[j]);
  }

  free(lz->nnz);
  free(lz->i);
  free(lz->v);

  free(lz);
}

static sparsecol getcol(sparsemat *base, int j)
{
  sparsecol col;
  lazymat *lz;

  lz = cast(base);

  col.nnz = lz->nnz[j];
  col.i = lz->i[j];
  col.v = lz->v[j];

  return col;
}

void lazymat_col_add(lazymat *lz, int j, int nnz, const int *i, const double *v)
{
  int *lzi;
  double *lzv;

  if(lz->nnz[j] != -1) { // Ignore repeated additions
    return;
  }

  assert(nnz >= 0);
  if(nnz == 0) {
    lz->nnz[j] = 0;
    return;
  }

  lzi = (int *)malloc(sizeof(int) * nnz);
  lzv = (double *)malloc(sizeof(double) * nnz);
  if(lzi == NULL || lzv == NULL) {
    return;
  }

  memcpy(lzi, i, sizeof(int) * nnz);
  memcpy(lzv, v, sizeof(double) * nnz);

  lz->nnz[j] = nnz;
  lz->i[j] = lzi;
  lz->v[j] = lzv;
}

static const sparsemat_vtable vtable = {
  .destroy = destroy,
  .getcol = getcol
};
