#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <memory.h>
#include <stdio.h>
#include <ctype.h>

#include "linalg.h"
#include "debug.h"

#include <f2c.h>

#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

extern int dsyev_(const char *jobz, const char *uplo, const integer *n,
                  double *A, const integer *lda, double *w,
                  double *work, integer *lwork,
                  integer *info);


int dsyev_wrap(char jobz, char uplo, int n, double *A, int lda, double *w)
{
  integer ln, llda, lwork, info;
  double *work, worksize;

  ln = (integer)n;
  llda = (integer)lda;

  // Get workspace size
  lwork = -1;
  dsyev_(&jobz, &uplo, &ln, A, &llda, w, &worksize, &lwork, &info);
  if(info != 0) {
    return (int)info;
  }

  lwork = (integer)worksize;
  work = (double *)malloc(sizeof(double) * lwork);

  // Main call
  dsyev_(&jobz, &uplo, &ln, A, &llda, w, work, &lwork, &info);

  // Cleanup
  free(work);

  return (int)info;
}

extern int dgemm_(const char *transa, const char *transb,
                  const integer *m, const integer *n, const integer *k,
                  const double *alpha, const double *A, const integer *lda,
                  const double *B, const integer *ldb,
                  const double *beta, double *C, const integer *ldc);

int dgemm_wrap(char transa, char transb, int m, int n, int k, double alpha, const double *A, int lda, const double *B, int ldb, double beta, double *C, int ldc)
{
  integer lm, ln, lk, llda, lldb, lldc;
  int s;

  lm = (integer)m;
  ln = (integer)n;
  lk = (integer)k;
  llda = (integer)lda;
  lldb = (integer)ldb;
  lldc = (integer)ldc;

  // TODO: remove assertions, return -1 instead, propagate error upwards
  assert(m >= 0);
  assert(n >= 0);
  assert(k >= 0);
  if(tolower(transa) == 'n') {
    assert(lda >= MAX(1, m));
  }
  else {
    assert(lda >= MAX(1, k));
  }

  if(tolower(transb) == 'n') {
    assert(ldb >= MAX(1, k));
  }
  else {
    assert(ldb >= MAX(1, n));
  }
  assert(ldc >= MAX(1, m));

  s = dgemm_(&transa, &transb, &lm, &ln, &lk, &alpha, A, &llda, B, &lldb, &beta, C, &lldc);
  //fprintf(stderr, "s = %d\n", s);
  //assert(s == 0);

  return s;
}

void sqrtm_mult(int m, int n, const double *V, const double *lambdasqrt, const double *X, double *prod)
{
  double *VTX;
  int i, j;
  
  VTX = (double *)malloc(n * m * sizeof(double));
  dgemm_wrap('T', 'N', m, n, m, 1, V, m, X, m, 0, VTX, m); 

  for(j = 0; j < m; j++) {
    for(i = 0; i < n; i++) {
      VTX[n*j + i] *= lambdasqrt[i];
    }
  }

  dgemm_wrap('N', 'N', m, n, m, 1, V, m, VTX, m, 0, prod, m);

  free(VTX);
}

void invsqrtm_mult(int m, int n, const double *V, const double *lambdasqrt, const double *X, double *solve)
{
  double *invlambdasqrt;
  int i;

  invlambdasqrt = (double *)malloc(sizeof(double) * n);
  for(i = 0; i < n; i++) {
    invlambdasqrt[i] = 1 / lambdasqrt[i];
  }

  sqrtm_mult(m, n, V, invlambdasqrt, X, solve);

  free(invlambdasqrt);
}

void sqrtm_mult_right(int m, int n, const double *V, const double *lambdasqrt, const double *X, double *prod)
{
  double *XV;
  int i, j;

  XV = (double *)malloc(n * m * sizeof(double));
  dgemm_wrap('N', 'N', m, n, n, 1, X, m, V, n, 0, XV, m); 

  for(j = 0; j < n; j++) {
    for(i = 0; i < m; i++) {
      XV[m*j + i] *= lambdasqrt[j];
    }
  }

  dgemm_wrap('N', 'T', m, n, n, 1, XV, m, V, n, 0, prod, m);

  free(XV);
}

void invsqrtm_mult_right(int m, int n, const double *V, const double *lambdasqrt, const double *X, double *solve)
{
  double *invlambdasqrt;
  int i;

  invlambdasqrt = (double *)malloc(sizeof(double) * n);
  for(i = 0; i < n; i++) {
    invlambdasqrt[i] = 1 / lambdasqrt[i];
  }

  sqrtm_mult_right(m, n, V, invlambdasqrt, X, solve);

  free(invlambdasqrt);
}

void sqrtm_calc(int n, const double *V, const double *lambdasqrt, double *sqrtA)
{
  double *Vlambda;
  int i, j;

  // Right-multiply sqrt of eigenvalues into V
  Vlambda = (double *)malloc(sizeof(double) * n * n);
  for(j = 0; j < n; j++) {
    for(i = 0; i < n; i++) {
      Vlambda[n*j + i] = V[n*j + i] * lambdasqrt[j];
    }
  }

  // sqrtA <- Vlambda * V^T
  dgemm_wrap('N', 'T', n, n, n, 1, Vlambda, n, V, n, 0, sqrtA, n);

  free(Vlambda);
}

void cholsolve(int n, int nrhs, const double *A, double *x)
{
  assert(0);
  // TODO
}

extern int dgesv_(const integer *n, const integer *nrhs, double *A, const integer *lda, integer *ipiv, double *B, const integer *ldb, integer *info);

int lusolve(int n, int nrhs, const double *A, double *x, double *LU, int *piv)
{
  integer ln, lnrhs, info, *ipiv;
  int i;
  double *freeLU;

  ln = (integer)n;
  lnrhs = (integer)nrhs;

  // Set up LU
  if(LU == NULL) {
    freeLU = LU = (double *)malloc(sizeof(double) * n * n);
  }
  else {
    freeLU = NULL;
  }
  memcpy(LU, A, sizeof(double) * n * n);

  // Allocate pivot vector iff int != integer
  if(sizeof(integer) == sizeof(int) && piv != NULL) {
    ipiv = piv;
  }
  else {
    ipiv = (integer *)malloc(sizeof(integer) * n);
  }

  // Main call
  dgesv_(&ln, &lnrhs, LU, &ln, ipiv, x, &ln, &info);

  // Convert permutation
  if(piv != NULL && sizeof(integer) != sizeof(int)) {
    for(i = 0; i < n; i++) {
      piv[i] = (int)ipiv[i];
    }
  }

  // Cleanup
  if(ipiv != piv) {
    free(ipiv);
  }
  free(freeLU);
  
  return (int)info;
}



void rref(int m, int n, double tol, double *A, int *bound, int *rank)
{
  int i, j, k, l, r, nbound;
  double p, swap;

  // Based off rref.m from Matlab

  i = j = nbound = 0;
  while(i < m && j < n) {
    // Find maximum element in column j
    p = -INFINITY;
    k = -1;
    for(l = i; l < m; l++) {
      if(fabs(A[l + j*m]) > p) {
        p = fabs(A[l + j*m]);
        k = l;
      }
    }
    //if(k == -1) {
    //  for(l = i; l < m; l++) {
    //    printf("%lg ", fabs(A[l + j*m]));
    //  }
    //  printf("\ni = %d, j = %d, m = %d\n", i, j, m);
    //}
    assert(k != -1);


    if(p <= tol) {
      // Zero out column
      for(l = i; l < m; l++) {
        A[l + j*m] = 0;
      }
      j++;
    }
    else {
      bound[nbound++] = j;

      for(l = j; l < n; l++) {
        // Swap i-th and k-th rows
        swap = A[i + l*m];
        A[i + l*m] = A[k + l*m];
        A[k + l*m] = swap;
      }

      // Dividing pivot row by pivot element
      p = A[i + j*m];
      for(l = j; l < n; l++) {
        A[i + l*m] /= p;
      }

      // Subtract pivot row out from remaining rows
      for(r = 0; r < m; r++) {
        if(r == i) {
          continue;
        }

        p = A[r + j*m];
        for(l = j; l < n; l++) {
          A[r + l*m] -= p * A[i + l*m];
        }
      }

      i++; j++;
    }
  }

  *rank = nbound;
}
