#include <assert.h>
#include <stdint.h>

#include "matio.h"

void sparsemat_write_to_file(FILE *fp, cs_di *A)
{
  int j, k, m, n, nz, *p, *row;
  double *val;

  m = A->m;
  n = A->n;
  nz = A->nz;

  p = A->p;
  row = A->i;
  val = A->x;

  if(nz == -1) {
    // Compressed-col
    nz = p[n];
    assert(nz >= 0);

    fprintf(fp, "%d %d %d\n", m, n, nz);
    for(j = 0; j < n; j++) {
      for(k = p[j]; k < p[j+1]; k++) {
        fprintf(fp, "%d %d %lg\n", row[k], j, val[k]);
      }
    }
  }
  else {
    // Triplet
    assert(nz >= 0);

    fprintf(fp, "%d %d %d\n", m, n, nz);
    for(k = 0; k < nz; k++) {
      fprintf(fp, "%d %d %lg\n", row[k], p[k], val[k]);
    }
  }
}

void sparsemat_write_to_path(const char *path, cs_di *A)
{
  FILE *fp;

  fp = fopen(path, "w+");
  if(fp == NULL) {
    return;
  }

  sparsemat_write_to_file(fp, A);

  fclose(fp);
}

/////////////////////////// Sparsemat Files (writing) /////////////////////////

static int writeint(FILE *fp, int i);
static int writeuint64(FILE *fp, uint64_t i);
static int writedouble(FILE *fp, double x);

int sparsemat_write_binary_file(FILE *fp, cs_di *A)
{
  cs_di *free;
  int m, n, j, k, nnz;
  double v;
  size_t r;

  // Ensure compressed-col
  free = NULL;
  if(A->nz != -1) {
    free = A = cs_di_compress(A);
  }

  m = A->m;
  n = A->n;

  // Write sizes to file
  if(writeint(fp, A->m)) {
    return 1;
  }
  if(writeint(fp, A->n)) {
    return 1;
  }

  // Write column indices to file
  assert(A->p[0] == 0);
  for(j = 0; j < n + 1; j++) {
    assert(A->p[j] >= 0);
    if(writeint(fp, A->p[j])) {
      return 1;
    }
  }

  // Write entries to file
  nnz = A->p[n];
  assert(nnz <= A->nzmax);
  for(k = 0; k < nnz; k++) {
    assert(A->i[k] >= 0 && A->i[k] < m);
    if(writeint(fp, A->i[k])) {
      return 1;
    }
    if(writedouble(fp, A->x[k])) {
      return 1;
    }
  }

  // Cleanup
  cs_di_spfree(A);

  return 0;
}

int sparsemat_write_binary_path(const char *path, cs_di *A)
{
  FILE *fp;

  fp = fopen(path, "wb+");
  if(fp == NULL) {
    return 1;
  }

  sparsemat_write_binary_file(fp, A);
  fclose(fp);

  return 0;
}

/////////////////////////// Sparsemat Files (reading) /////////////////////////

static uint64_t readuint64(FILE *fp);
static double readdouble(FILE *fp);

struct sparsemat_file
{
  FILE *fp;

  uint64_t m, n;
  uint64_t *p;

  off_t ivoff;
};

sparsemat_file *sparsemat_open_binary_file(const char *path)
{
  FILE *fp;
  sparsemat_file *sf;
  uint64_t j, n;

  sf = (sparsemat_file *)malloc(sizeof(sparsemat_file));
  if(sf == NULL) {
    return NULL;
  }
  
  // Open file
  sf->fp = fp = fopen(path, "rb");
  if(fp == NULL) {
    free(sf);
    return NULL;
  }

  // Read sizes
  sf->m = readuint64(fp);
  sf->n = n = readuint64(fp);

  sf->p = (uint64_t *)malloc(sizeof(uint64_t) * (n+1));
  for(j = 0; j < n+1; j++) {
    sf->p[j] = readuint64(fp);
  }

  sf->ivoff = ftello(fp);
  assert(sf->ivoff == (off_t) (8 * (n + 3)));

  // Check for errors
  if(ferror(fp)) {
    fclose(fp);
    free(sf->p);
    free(sf);
    return NULL;
  }

  return sf;
}

void sparsemat_close_binary_file(sparsemat_file *fp)
{
  if(fp == NULL) {
    return;
  }

  fclose(fp->fp);
  free(fp->p);
  free(fp);
}

void sparsemat_file_dims(sparsemat_file *sf, int dims[2])
{
  dims[0] = (int)sf->m;
  dims[1] = (int)sf->n;
}

int sparsemat_file_nnz(sparsemat_file *sf)
{
  return (int)sf->p[sf->n];
}

int sparsemat_file_colnnz(sparsemat_file *sf, int icol)
{
  uint64_t col;

  if(icol < 0) {
    return -1;
  }

  col = (uint64_t)icol;
  if(col >= sf->n) {
    return -1;
  }

  return (int)(sf->p[col+1] - sf->p[col]);
}

void sparsemat_file_get_col(sparsemat_file *sf, int col, int *i, double *v)
{
  FILE *fp;
  off_t coloffset;
  uint64_t k, nnz;

  fp = sf->fp;

  coloffset = sf->ivoff;
  coloffset += (sizeof(uint64_t) + sizeof(double)) * (off_t)sf->p[col];
  fseeko(fp, coloffset, SEEK_SET);

  nnz = sf->p[col+1] - sf->p[col];
  for(k = 0; k < nnz; k++) {
    i[k] = readuint64(fp);
    v[k] = readdouble(fp);
  }
}

cs_di *sparsemat_file_read_all(sparsemat_file *sf)
{
  uint64_t j, n;
  int p, nnz;
  cs_di *A;

  n = sf->n;
  nnz = sparsemat_file_nnz(sf);
  A = cs_di_spalloc((int)sf->m, (int)n, nnz, 1, 0);
  if(A == NULL) {
    return NULL;
  }

  for(j = 0; j < n; j++) {
    A->p[j] = p = (int)sf->p[j];
    sparsemat_file_get_col(sf, j, A->i + p, A->x + p);
  }
  A->p[n] = nnz;

  return A;
}

cs_di *sparsemat_file_read_all_path(const char *path)
{
  sparsemat_file *sfm;
  cs_di *A;

  sfm = sparsemat_open_binary_file(path);
  if(sfm == NULL) {
    return NULL;
  }

  A = sparsemat_file_read_all(sfm);
  sparsemat_close_binary_file(sfm);

  return A;
}

//////////////////////////// Binary I/O Utility Functions ////////////////////

static int writeint(FILE *fp, int i)
{
  size_t r;
  int64_t i64;

  i64 = (int64_t)i;
  r = fwrite(&i64, sizeof(int64_t), 1, fp);
  if(r != 1) {
    return 1;
  }

  return 0;
}

static int writeuint64(FILE *fp, uint64_t i)
{
  size_t r;

  r = fwrite(&i, sizeof(uint64_t), 1, fp);
  if(r != 1) {
    return 1;
  }

  return 0;
}

static int writedouble(FILE *fp, double x)
{
  size_t r;

  r = fwrite(&x, sizeof(double), 1, fp);
  if(r != 1) {
    return 1;
  }

  return 0;
}

static uint64_t readuint64(FILE *fp)
{
  size_t r;
  uint64_t i;

  r = fread(&i, sizeof(uint64_t), 1, fp);
  if(r != 1) {
    return 0;
  }

  return i;
}

static double readdouble(FILE *fp)
{
  size_t r;
  double x;

  r = fread(&x, sizeof(double), 1, fp);
  if(r != 1) {
    return 0;
  }

  return x;
}
