#include <assert.h>
#include <math.h>
#include <stdlib.h>

#include "randorth.h"
#include "geom.h"

static double frand(double a, double b)
{
  return (rand() / (double)RAND_MAX) * (b - a) + a;
}

static void frand_vector(double a, double b, int n, double *v)
{
  int i;

  for(i = 0; i < n; i++) {
    v[i] = frand(a, b);
  }
}

static double Bprod(const cs_di *B, double *Bv, int n, const double *u, const double *v)
{
  int i;

  if(B == NULL) {
    return geom_dot_product((size_t)n, u, v);
  }

  assert(n == B->m && n == B->n);
  assert(Bv != NULL);

  for(i = 0; i < n; i++) {
    Bv[i] = 0;
  }
  cs_di_gaxpy(B, v, Bv);

  return geom_dot_product((size_t)n, u, Bv);
}


void gram_schmidt(const cs_di *B, int n, int p, double *V, int fwd)
{
  int i, j, k;
  double dotprod, *Bv;
  
  Bv = NULL;
  if(B != NULL) {
    Bv = (double *)malloc(sizeof(double) * n);
    assert(n == B->m && n == B->n);
  }

  if(fwd) {
    for(i = 0; i < p; i++) {
      // Normalize Vi
      dotprod = Bprod(B, Bv, n, V + n*i, V + n*i);
      geom_vector_scale((size_t)n, V + n*i, 1/sqrt(dotprod));

      // Remove projection onto Vi from later terms
      for(j = i+1; j < p; j++) {
        dotprod = Bprod(B, Bv, n, V + n*i, V + n*j);

        for(k = 0; k < n; k++) {
          V[n*j + k] -= dotprod * V[n*i + k];
        }
      }
    }
  }
  else {
    for(i = p - 1; i >= 0; i--) {
      // Normalize Vi
      dotprod = Bprod(B, Bv, n, V + n*i, V + n*i);
      geom_vector_scale((size_t)n, V + n*i, 1/sqrt(dotprod));
      
      // Remove projection onto Vi from earlier terms
      for(j = 0; j < i; j++) {
        dotprod = Bprod(B, Bv, n, V + n*i, V + n*j);

        for(k = 0; k < n; k++) {
          V[n*j + k] -= dotprod * V[n*i + k];
        }
      }
    }
  }

  free(Bv);
}


void randorth(int n, int p, double *V)
{
  int i, j, k;
  double dotprod;

  // Initialize V to random vectors
  frand_vector(0, 1, n*p, V);

  // Orthogonalize
  gram_schmidt(NULL, n, p, V, 1);
}

void randgenorth(const cs_di *B, int p, double *V)
{
  int i, n;
  
  n = B->m;
  assert(n == B->n);

  // Initialize V to random vectors
  frand_vector(0, 1, n*p, V);

  // Orthogonalize
  gram_schmidt(B, n, p, V, 1);
}
