#include <stdlib.h>
#include <memory.h>
#include <assert.h>

#include "sort.h"

#include "rowset.h"

struct rowset
{
  int m, n, k;
  int *ind;
  double *X;
};

rowset *rowset_create(int m, int n, int k, const int *ind, const double *X)
{
  rowset *r;

  r = (rowset *)malloc(sizeof(rowset));
  if(r == NULL) {
    return NULL;
  }

  r->m = m;
  r->n = n;
  r->k = k;

  r->ind = (int *)malloc(sizeof(int) * k);
  if(r->ind == NULL) {
    free(r);
    return NULL;
  }

  memcpy(r->ind, ind, sizeof(int) * k);
  sort_ints((size_t)k, r->ind);

  r->X = (double *)malloc(sizeof(double) * k * n);
  if(r->X == NULL) {
    free(r->ind);
    free(r);
    return NULL;
  }

  memcpy(r->X, X, sizeof(double) * k * n);

  return r;
}

rowset *rowset_clone(const rowset *r)
{
  return rowset_create(r->m, r->n, r->k, r->ind, r->X);
}

void rowset_destroy(rowset *r)
{
  if(r == NULL) {
    return;
  }

  free(r->ind);
  free(r->X);
  free(r);
}

int rowset_nrows(const rowset *r)
{
  return r->k;
}

int rowset_nrows_all(const rowset *r)
{
  return r->m;
}

int rowset_ncols(const rowset *r)
{
  return r->n;
}

int rowset_nelts(const rowset *r)
{
  return r->k * r->n;
}

const int *rowset_inds(const rowset *r)
{
  return r->ind;
}

const double *rowset_data(const rowset *r)
{
  return r->X;
}

const double *rowset_get(const rowset *r, int i)
{
  int *f, j;

  f = bsearch_ints((size_t)r->k, i, r->ind);
  if(f == NULL) {
    return NULL;
  }

  j = f - r->ind;

  return r->X + j*r->n;
}

void rowset_getall_colmajor(const rowset *r, double *X)
{
  int i, j, k, n;

  k = r->k;
  n = r->n;

  for(i = 0; i < k; i++) {
    for(j = 0; j < n; j++) {
      X[i + k*j] = r->X[n*i + j];
    }
  }
}

////////////////////////////////// extract/install ////////////////////

rowset *rowset_extract(int m, int n, int k, const int *ind, const double *X)
{
  rowset *r;
  int i;

  r = (rowset *)malloc(sizeof(rowset));
  if(r == NULL) {
    return NULL;
  }

  r->m = m;
  r->n = n;
  r->k = k;

  r->ind = (int *)malloc(sizeof(int) * k);
  if(r->ind == NULL) {
    free(r);
    return NULL;
  }

  memcpy(r->ind, ind, sizeof(int) * k);
  sort_ints((size_t)k, r->ind);

  r->X = (double *)malloc(sizeof(double) * k * n);
  if(r->X == NULL) {
    free(r->ind);
    free(r);
    return NULL;
  }

  for(i = 0; i < k; i++) {
    assert(ind[i] >= 0 && ind[i] < m);
    memcpy(r->X + n*i, X + n*ind[i], sizeof(double) * n);
  }

  return r;
  
}

void rowset_install(double *X, const rowset *r)
{
  int i, j, k, m, n, *ind;

  m = r->m;
  n = r->n;
  k = r->k;
  ind = r->ind;
  
  for(i = 0; i < k; i++) {
    assert(ind[i] >= 0 && ind[i] < m);

    memcpy(X + n*ind[i], r->X + n*i, sizeof(double) * n);
  }
}
