#include "sample.h"

#include <assert.h>

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <f2c.h>
#include <math.h>

#include "linalg.h"

struct sample
{
  gsl_rng *rng;
};

static void linalg_sqrt(size_t size, double *root, const double *mat);

sample_t *sample_init(unsigned long int seed)
{
  sample_t *s;

  s = (sample_t *)malloc(sizeof(sample_t));

  // Initialize GSL random number generator
  s->rng = gsl_rng_alloc(gsl_rng_default);
  gsl_rng_set(s->rng, seed);

  return s;
}

void sample_end(sample_t *s)
{
  gsl_rng_free(s->rng);
  free(s);
}

double sample_uniform(sample_t *s, double min, double max)
{
  return gsl_ran_flat(s->rng, min, max);
}

size_t sample_discrete(sample_t *s, size_t k, const double *pi)
{
  double sum, x;
  size_t i;

  assert(k > 0);

  sum = 0;
  for(i = 0; i < k; i++) {
    sum += pi[i];
  }

  x = gsl_ran_flat(s->rng, 0, sum);

  for(i = 0; i < k && x >= 0; i++) {
    x -= pi[i];
  }

  return i-1;
}

static void sample_gaussian_sqrt(sample_t *s, size_t ndims, double *p, const double *mu, const double *sqrt_sigma, double *z)
{
  size_t i, j;

  for(i = 0; i < ndims; i++) {
    z[i] = gsl_ran_gaussian(s->rng, 1.0);
  }

  for(i = 0; i < ndims; i++) {
    p[i] = mu[i];
    for(j = 0; j < ndims; j++) {
      p[i] += sqrt_sigma[ndims*i + j] * z[j];
    }
  }
}

void sample_gaussian(sample_t *s, size_t ndims, size_t n, double *p, const double *mu, const double *sigma)
{
  double *sqrt_sigma, *z;
  size_t i;

  // Setup
  z = (double *)malloc(ndims * (ndims + 1) * sizeof(double));
  sqrt_sigma = z + ndims;
  linalg_sqrt(ndims, sqrt_sigma, sigma);

  for(i = 0; i < n; i++) {
    sample_gaussian_sqrt(s, ndims, p + ndims*i, mu, sqrt_sigma, z);
  }

  // Cleanup
  free(z);
}

void sample_MoG(sample_t *s,
                size_t ndims, size_t n, double *p, int *label,
                size_t k, const double *pi, const double *mu,
                const double *sigma)
{
  size_t i, l;
  unsigned int j;

  gsl_ran_discrete_t *label_dist;

  double *sqrt_sigma, *z;

  // Preprocess distributions
  label_dist = gsl_ran_discrete_preproc(k, pi);

  z = (double *)malloc((ndims + k * ndims * ndims) * sizeof(double));
  sqrt_sigma = z + ndims;
  for(i = 0; i < k; i++) {
    linalg_sqrt(ndims, sqrt_sigma + i*ndims*ndims, sigma + i*ndims*ndims);
  }

  // Sample points
  for(i = 0; i < n; i++) {
    label[i] = l = gsl_ran_discrete(s->rng, label_dist);

    sample_gaussian_sqrt(s, ndims, p + ndims*i, mu + ndims*l, sqrt_sigma + ndims*ndims*l, z);
  }

  // Cleanup
  gsl_ran_discrete_free(label_dist);
}


extern int dgesvd_(
  const char *jobu,
  const char *jobv,
  const integer *m, const integer *n, doublereal *A, const integer *lda,
  doublereal *S,
  doublereal *U, const integer *ldu,
  doublereal *VT, const integer *ldvt,
  doublereal *work, const integer *lwork,
  integer *info);

static void linalg_sqrt(size_t size, double *root, const double *mat)
{
  char jobu, jobv;
  integer m, n, lda, ldu, ldvt, lwork, info;
  doublereal worksize, *work;
  doublereal *A, *S, *U;

  size_t i, j, k;

  jobu = 'A';
  jobv = 'N';

  lda = m = (integer)size;
  n = (integer)size;

  ldu = (integer)size;
  ldvt = 1;

  A = (doublereal *)malloc((2 * size * size + size) * sizeof(doublereal));
  U = A + size*size;
  S = U + size*size;

  // Initialize input
  for(i = 0; i < size; i++) {
    for(j = 0; j < size; j++) {
      assert(mat[i*size + j] == mat[i + j*size]);

      A[size*i + j] = mat[size*i + j];
    }
  }

  // Find workspace size
  lwork = -1;
  dgesvd_(&jobu, &jobv, &m, &n, A, &lda, S, U, &ldu, NULL, &ldvt, &worksize, &lwork, &info);
  lwork = (integer)worksize;
  work = (doublereal *)malloc(lwork * sizeof(doublereal));

  // Calculate SVD
  dgesvd_(&jobu, &jobv, &m, &n, A, &lda, S, U, &ldu, NULL, &ldvt, work, &lwork, &info);

  // Calculate square root
  for(i = 0; i < size; i++) {
    S[i] = sqrt(S[i]);
  }

  for(i = 0; i < size; i++) {
    for(j = 0; j < size; j++) {
      root[i*size + j] = 0;
      for(k = 0; k < size; k++) {
        root[i*size + j] += U[i + size*k] * S[k] * U[j + size*k];
      }
    }
  }

  // Cleanup
  free(work);
  free(A);
}
