#include <stdio.h>
#include <assert.h>
#include <pthread.h>

#include "vptree_parallel.h"
#include "vptree_struct.h"

/////////////////////////////// vp-tree Construction ////////////////////////

typedef struct distp {
  double d;
  const void *p;
} distp;

static void sort_distp(int n, distp *dp);

static node *node_create(vptree *vp, node *parent, int n, distp *dp, int nthreads, pthread_t *spawn);
static void node_destroy(vptree *vp, node *nd);
static void node_add(vptree *vp, node *nd, int n, distp *dp, int nthreads);



struct add_thread_params
{
  vptree *vp;
  node *nd;
  int n;
  distp *dp;
  int nthreads;
};

static void *add_thread_routine(void *ptr)
{
  struct add_thread_params *params;

  params = (struct add_thread_params *)ptr;

  printf("started new thread\n");
  node_add(params->vp, params->nd, params->n, params->dp, params->nthreads);

  deallocate(params->vp, params);
  return NULL;
}

static pthread_t node_add_spawn(vptree *vp, node *nd, int n, distp *dp, int nthreads)
{
  struct add_thread_params *params;
  pthread_t pid;

  // Set up params
  params = allocate(vp, sizeof(struct add_thread_params));
  params->vp = vp;
  params->nd = nd;
  params->n = n;
  params->dp = dp;
  params->nthreads = nthreads;

  // Spawn thread
  pthread_create(&pid, NULL, add_thread_routine, params);

  return pid;
}

static node *node_create(vptree *vp, node *parent, int n, distp *dp, int nthreads, pthread_t *spawn)
{
  node *nd;
  int v;
  const void *p, *swap;

  // Null node
  if(n == 0) {
    return NULL;
  }

  nd = allocate(vp, sizeof(node));
  assert(nd != NULL);
  nd->vp = vp;
  nd->parent = parent;

  // Update progress
  //if(callback != NULL) {
  //  (*alli)++;
  //  callback(user_data, *alli, alln);
  //}

  // Select reference node
  // TODO
  v = rand() % n;
  nd->p = p = dp[v].p;

  // Initalize as singleton node
  nd->mu = -1;
  nd->lt = nd->ge = NULL;

  if(n != 1) {
    // Add subnodes
    swap = dp[0].p;
    dp[0].p = dp[v].p;
    dp[v].p = swap;

    if(spawn == NULL) {
      node_add(vp, nd, n-1, dp+1, nthreads);
    }
    else {
      *spawn = node_add_spawn(vp, nd, n-1, dp+1, nthreads);
    }
  }

  return nd;
}

static pthread_t add_subtree(vptree *vp, node *parent, node **subtree_ptr, int n, distp *dp, int nthreads, int spawn)
{
  pthread_t pid;

  if(n == 0) {
    return 0;
  }

  if(*subtree_ptr == NULL) {
    *subtree_ptr = node_create(vp, parent, n, dp, nthreads, spawn ? &pid : NULL);
  }
  else {
    if(spawn) {
      pid = node_add_spawn(vp, *subtree_ptr, n, dp, nthreads);
    }
    else {
      node_add(vp, *subtree_ptr, n, dp, nthreads);
    }
  }

  return pid;
}

static void node_add(vptree *vp, node *nd, int n, distp *dp, int nthreads)
{
  int i, m;
  int ltthreads, gethreads;
  pthread_t spawn;

  spawn = 0;

  // Calculate distances
  for(i = 0; i < n; i++) {
    dp[i].d = distance(vp, nd->p, dp[i].p);

    //if(i % (128*1024) == 0) {
    //  fprintf(stderr, "Distance %d of %d (%.2f%%)\n", i, n, (((float)i)*100)/n);
    //}
  }
  //fprintf(stderr, "Sorting...\n");
  sort_distp(n, dp);

  // Previously a leaf node, find median distance
  if(nd->mu < 0) {
    // Find median distance
    m = n/2;

    if(n % 2 == 0) {
      nd->mu = (dp[m].d + dp[m+1].d)/2;
    }
    else {
      nd->mu = dp[m].d;
    }
  }

  // Find split point
  for(m = 0; m < n && dp[m].d < nd->mu; m++);

  // Add subnodes
  ltthreads = nthreads / 2;
  gethreads = nthreads - ltthreads;

  if(nthreads > 1) {
    spawn = add_subtree(vp, nd, &(nd->lt), m, dp, ltthreads, 1);
  }
  else {
    add_subtree(vp, nd, &(nd->lt), m, dp, 1, 0);
  }
  add_subtree(vp, nd, &(nd->ge), n - m, dp + m, gethreads, 0);

  // Wait for subthreads
  if(spawn != 0) {
    pthread_join(spawn, NULL);
  }
}

void vptree_add_many_parallel(vptree *vp, int n, const void * const *p, int nthreads)
{
  int i;
  distp *dp;

  // Create distance-comparison structure
  dp = allocate(vp, n * sizeof(distp));
  for(i = 0; i < n; i++) {
    dp[i].p = p[i];
  }

  // Add to tree
  if(vp->root == NULL) {
    vp->root = node_create(vp, NULL, n, dp, nthreads, NULL);
  }
  else {
    node_add(vp, vp->root, n, dp, nthreads);
  }

  deallocate(vp, dp);

  vp->n += n;
}

/////////////////////////////// Utility Functions ////////////////////////


static int compare_distp(const void *v1, const void *v2)
{
  double d1, d2;

  d1 = ((const distp *)v1)->d;
  d2 = ((const distp *)v2)->d;

  if(d1 < d2) {
    return -1;
  }
  else if(d1 == d2) {
    return 0;
  }
  else {
    return 1;
  }
}

static void sort_distp(int n, distp *dp)
{
  qsort(dp, (size_t)n, sizeof(distp), compare_distp);
}
